# Restart with tcl: -*- mode: tcl; tab-width: 8; -*- 
#************************************************************************
# E.S.O. - VLT project
#
# "@(#) $Id: audioVerify.tcl,v 1.13 2005/10/28 08:40:22 vltsccm Exp $"
#
# who       when        what
# --------  ----------  ----------------------------------------------
# eallaert  2005-06-02  first version
# 
#
# audioVerify.tcl, audioServer.tcl and audioCtrl.tcl were "inspired" by some
# of the snack demos, in particular rplay.tcl and aserver.tcl.
#************************************************************************
#   NAME
#   audioVerify - verify the audio system
#   
#   SYNOPSIS
#   \fBaudioVerify\fR
#   
#   DESCRIPTION
#   audioVerify will check the audio-server. The list of supported servers is 
#   given by the macro \fISUPPORT\fR, which is set in the Makefile. It can be any
#   list of the words \fISnack\fR, \fIHP\fR and \fINCD\fR. Apart from specifying which
#   audio-servers are supported, it also sets the order in which they will be
#   tested.
#   
#   The following strategy is applied trying to determine the nature of the 
#   various types of sound servers:
#   1. First, the host of the server is given by the AUDIO/DISPLAY 
#      environment variable (see below).
#   2. For type Snack, try to connect to an audioServer. These servers are
#      listening on a particular TCP/IP port, as defined by the macro \fIPORT\fR.
#      If this fails, the audioServer is not (anymore) active.
#   3. For NCD, and if the client process is running on HP-UX, try the 
#      \fIlocal\fR client application for NCD X-terminals. The path to this client 
#      program is given in the Makefile by the variable \fINCDPATH\fR.
#   4. For HP, and if the client process is running on HP-UX, try directly 
#      the \fIlocal\fR snack Tcl-extension, which uses the HP audio library.
#   
#   This determination of the server type takes place the first time a sound
#   file is sent to a particular server, or the first time this procedure is
#   called. If none of the types fit, subsequent calls to audioPlay or this
#   procedure will only look for an audioServer. 
#   
#   Remark that the check for an HP server will take ~30 seconds if 
#   \fIAUDIO/DISPLAY\fR is pointing to an HP box without audio hardware (or with
#   the Aserver not active). However, if audioVerify is used as a procedure,
#   this will only be for the first attempt to play a sound on that server.
#   Subsequent attempts will fail immediately, as only the existence of an
#   active audioServer will be verified - see above.
#   
#   FILES
#   
#   ENVIRONMENT
#   The environment variable \fIAUDIO\fR is used to determine the hostname of the
#   audio-server. It can be set to a hostname or IP address. If \fIAUDIO\fR is not
#   set, the environment variable \fIDISPLAY\fR will be used (i.e. the value up to
#   the colon). If neither is set, it is assumed that the audio-server runs 
#   on the \fIlocalhost\fR.
#   
#   RETURN VALUES
#   In case of success, the audio-server type is returned. It can take the
#   value of \fISnack\fR, \fIHP\fR or \fINCD\fR, depending also on the set of supported
#   devices audioVerify has been configured for via the Makefile.
#
#   CAUTIONS
#   Use of audio on NCD terminals requires that the sound files are on a 
#   directory which the terminal can NFS-mount. On top of that, the NCD
#   software must on the HP-UX box be installed under a particular directory,
#   as defined in the module's Makefile. If this is not the case, there will
#   be no audio played on the NCD terminal.
#
#   Both audioServer and HP's Aserver can be running on a different host than 
#   where the audioVerify client runs. But due to the proprietary nature of HP's
#   audio software library, the use of HP's Aserver can only work if the
#   audioVerify client too is running on HP-UX. For NCD X-terminals, the NCD 
#   client software must be installed on the local machine - and this is only 
#   available for HP-UX; the AUDIO or DISPLAY variable must of course in this
#   case point to the NCD X-terminal (on top of the other requirements listed
#   above).
#  
#   EXAMPLES
#   You have the choice between giving the fully qualified commandname (i.e.
#   including its namespace) or importing it into your current namespace.
#   For that reason, the following 2 examples are equivalent:
#   ...
#   tcl>package require Audio
#   ...
#   tcl>namespace import audio::*
#   tcl>audioVerify
#   Snack
#   tcl>
#
#   ...
#   tcl>package require Audio
#   ...
#   tcl>audio::audioVerify
#   Snack
#   tcl>
#
#   SEE ALSO
#   audioServer(1), audioCtrl(1), audioPlay(1)
#   \fIThe Snack Sound Toolkit\fR (http://www.speech.kth.se/snack/)
#
#------------------------------------------------------------------------
# Remark: this file is used both as a proc for a library and as the source
# for an executable; see the few lines of code at the end of this file.
#
# Ensure the namespace ::audio exists
audio::_initNamespace

proc ::audio::audioVerify {} {
    variable cfg

    set serverHost [_serverHost]

    set port $::audio::cfg(port)
    if {[info exist cfg($serverHost,type)]} {
	if {[catch {::audio::_check$cfg($serverHost,type) $serverHost $port} type]} {
	    error $type
	} else {
	    set cfg($serverHost,type) $type
	}
    } else {
	# must be first time; let us try them all, in right order
	foreach serverType $cfg(supportedTypes) {
	    set cfg($serverHost,type) $serverType
	    if {[catch {::audio::_check$serverType $serverHost $port} type]} {
		unset cfg($serverHost,type)
		continue; # try next type
	    } else {
		return $serverType
	    }		
	}
	# if we get here, none of the types matched
	set cfg($serverHost,type) Snack; # force to snack
	error "no audio server (types: $cfg(supportedTypes)) listening on $serverHost"
    }
}

proc ::audio::_checkSnack {serverHost port} {
    variable cfg

    # The sending of the previous audio-file must be finished
    # before we can send a command.
    audioFlush

    # Check if there is an snack audioServer running on that host,
    # i.e. try to send the "version" command to audioServer (which
    # will try to open binary socket connection to audioServer)
    if {[catch {_sendCmd vers} sock]} {
	close $fd
	error "no Snack audio server listening (anymore) at $serverHost/$port"
    } else {
	close $sock
	return Snack
    }
}

proc ::audio::_checkHP {serverHost port} {
    variable cfg

    if {![string equal $::tcl_platform(os) HP-UX]} {
	# No alternatives for Sun/Linux; actually for Solaris there is
	# /usr/bin/audioplay, but let us conveniently forget about this ;-)
	error "can only send sounds to HP devices from HP-UX"
    }
    switch -- [info command snack::sound] {} {
	set start [clock seconds]
	# we don't use the graphic part of 'snack', so just load 'sound'
	if {[catch {package require sound} msg]} {
	    error "snack not found"
	}
	set end [clock seconds]
	if {[expr {$end - $start}] > 5} {
	    error "HP audio HW problem or Aserver not active"
	}
    }

    # If the gain is not 0, it means the HW is being controlled.
    # If it is 0, try to change it; it will not work without HW
    set currentGain [snack::audio play_gain]
    if {$currentGain == 0} {
	snack::audio play_gain 10
	if {[snack::audio play_gain] == 0} {
	    error "HP audio HW problem or Aserver not active"
	} else {
	    snack::audio play_gain 0
	}
    }
    return HP
}

proc ::audio::_checkNCD {serverHost port} {
    if {![string equal $::tcl_platform(os) HP-UX]} {
	error "can only send sounds to NCD devices from HP-UX"
    }
    set ncdPath $::audio::cfg(ncdPath)
    if {![file exists $ncdPath/xpsh]} {
	error "NCD client program not found"
    }
    set state [catch {exec $ncdPath/xpsh aset q} msg]
    if {$state} {
	error "audio HW/server problem on NCD (check if sound directory can be NFS mounted)"
    } else {
	return NCD
    }
}

# If this file is sourced as a utility, pick up params and play the file.
# Be sure it was not sourced at level 0 by another script (that may also have
# unset the argv0 variable before we get here).
if {[info level] == 0 && [info exists argv0] && [string equal [info script] $argv0]} {
    if {[catch {eval audio::audioVerify $argv} msg]} {
	puts stderr $msg
	exit 1
    } else {
	puts stdout $msg
	exit 0
    }
}
