# Restart with tcl: -*- mode: tcl; tab-width: 8; -*- 
#************************************************************************
# E.S.O. - VLT project
#
# "@(#) $Id: audioVolume.tcl,v 1.13 2005/10/28 08:40:23 vltsccm Exp $"
# 
# who       when        what
# --------  ----------  ----------------------------------------------
# eallaert  2005-02-15  first version
# 
#
# audioPlay.tcl, audioServer.tcl and audioCtrl.tcl were "inspired" by some
# of the snack demos, in particular rplay.tcl and aserver.tcl.
#************************************************************************
#   NAME
#   audioVolume - get or set the audio volume
#   
#   SYNOPSIS
#   \fBaudioVolume\fR \fI?<volume>?\fR
#   
#   DESCRIPTION
#   \fBaudioVolume\fR without further arguments will return the current setting of
#   the audio-gain, on a scale from 0 to 100.
#   If <volume> is given, on a scale from 0 to 100, the audio-gain will be set
#   to <volume>; if the audio-server was muted, muting will be turned off.
#  
#   The strategy trying to determine the type of the sound server is the one
#   implemented in the \fIaudioVerify\fR procedure.
#   
#   FILES
#   
#   ENVIRONMENT
#   The environment variables \fIAUDIO\fR or \fIDISPLAY\fR will be used to determine
#   the hostname of the audio-server. See the manpage of \fIaudioVerify\fR.
#   
#   RETURN VALUES
#   In case of success, the volume is returned.
#
#   CAUTIONS
#   NCD terminals are not really supported by this procedure/utility. Getting
#   the volume will always return 100, setting it will have no effect.
#
#   In the case of HP's Aserver, the volume returned will be 0 if the server
#   is muted, even if the \fIaudioCtrl\fR application shows that the gain is
#   not 0. This is because for HP's Aserver there is no way to distinguish
#   between a gain of 0 and muting. In an attempt to keep these two things
#   apart, setting the <volume> to 0 on such server will actually set it to 1.
#   For Snack-based audio-servers, the audio-gain and muting are two separate
#   issues.
#  
#   EXAMPLES
#   You have the choice between giving the fully qualified commandname (i.e.
#   including its namespace) or importing it into your current namespace.
#   For that reason, the following 2 examples are equivalent:
#   ...
#   tcl>package require Audio
#   ...
#   tcl>namespace import audio::*
#   tcl>audioVolume
#   92
#   tcl>
#
#   ...
#   tcl>package require Audio
#   ...
#   tcl>audio::audioVolume 85
#   85
#   tcl>
#
#   SEE ALSO
#   audioServer(1), audioCtrl(1), audioVerify(n),
#   \fIThe Snack Sound Toolkit\fR (http://www.speech.kth.se/snack/)
#
#------------------------------------------------------------------------
# Remark: this file is used both as a proc for a library and as the source
# for an executable; see the few lines of code at the end of this file.
#
# Ensure the namespace ::audio exists
audio::_initNamespace

proc ::audio::audioVolume { {volume {}} } { 
    variable cfg
	
    set setVol 0
    if {[string length $volume] > 0} {
        if {![string is integer -strict $volume] || $volume < 0 || $volume > 100} {
            error "\"$volume\" must be an integer in the range 0-100"
	}
        set setVol 1
    }
    set serverHost [_serverHost]
    
    set port $::audio::cfg(port)
    if {![info exists cfg($serverHost,type)]} {
        if {[catch {audioVerify} serverType]} {
            error $serverType
	}
    } else {
	set serverType $cfg($serverHost,type)
    }
    
    switch -- $serverType Snack {
        if {$setVol} {
	    set cmd setv
	} else {
	    set cmd getv
	}
        # try to communicate with the Snack server
	if {[catch {_sendCmd $cmd} sock]} {
	    error "no Snack audio server listening (anymore) at $serverHost/$cfg(port)"
	} else {
	    # There is an active Snack server; get/set the volume.
	    if {$setVol} {
		puts $sock $volume
	    } else {
		set volume [lindex [gets $sock] 1]
	    }
	    close $sock
	}
	if {[catch {_sendCmd voce} sock]} {
	    error "no Snack audio server listening (anymore) at $serverHost/$cfg(port)"
	} else {
	    close $sock
	    return $volume
	}
    } HP {
	# Snack must be already loaded, as we know the server type
	snack::mixer volume Play audio::cfg($serverHost,volume)
	if {$setVol} {
	    # cheat if trying to set to 0, as this corresponds to muting
	    ###if {$volume == 0} {set volume 1}
	    set cfg($serverHost,volume) $volume;	# see below
	} else {
	    # there is a bug in Snack and/or Aserver: after a sound has been
	    # played, the gain can no longer be retrieved ?!!!
	    if {![info exists cfg($serverHost,volume)]} {
		snack::mixer update
	    	#####set cfg($serverHost,volume) [snack::audio play_gain]
	    }
	}
	return $cfg($serverHost,volume)
    } NCD {
	# not really a clue how to deal with this ....
	if {![info exists cfg($serverHost,volume)]} {
	    set cfg($serverHost,volume) 100
	}
	if {$setVol} {
	    return [set cfg($serverHost,volume) $volume]
	} else {
	    return $set cfg($serverHost,volume)
	}
    }
}

# If this file is sourced as a utility, pick up params and play the file.
# Be sure it was not sourced at level 0 by another script (that may also have
# unset the argv0 variable before we get here).
if {[info level] == 0 && [info exists argv0] && [string equal [info script] $argv0]} {
    if {[catch {eval audio::audioVolume $argv} msg]} {
	puts stderr $msg
	exit 1
    } else {
	puts stdout $msg
	exit 0
    }
}
