#*******************************************************************************
# E.S.O. - VLT project
#
# "@(#) $Id: bobFindLib.tcl,v 1.145 2005/10/27 09:07:07 vltsccm Exp $"
#
# who       when        what
# --------  ----------  ----------------------------------------------
# eallaert  30/11/96    created
# eallaert  1999-08-03  added bobFindObd
# eallaert  2002-10-25  use seq_findFile for bobFindVlt (SPR 20020574)
#

#************************************************************************
#   NAME
#   bobFindObd, bobFindSeq, bobFindSound, bobFindTsf, bobFindVlt
#	- BOB Find File Library
# 
#   SYNOPSIS
#   bobFindObd <obd> ?<insMode>?
#   bobFindSeq <template> ?<insMode>?
#   bobFindSound <soundFile>
#   bobFindTsf <templateSigFile> ?<insMode>?
#   bobFindVlt <relPathName>
#
#   DESCRIPTION
#   The functions contained herein locate a given file according to the 
#   specification of the ICD VCS-OH and related documents.  Parts of this
#   code have been taken over from or are inspired by vccFindLib.
#
#   bobFindIns: locates a file by its relative path.  The file given as a
#      relative path is searched in the following directories in the order
#      given.
#	1. in the $INS_ROOT/$INS_USER area:  "$INS_ROOT/$INS_USER/<relPathName>"
#             only if the environment variables INS_ROOT and INS_USER are set
#       2. in the $INS_ROOT/SYSTEM area:     "$INS_ROOT/SYSTEM/<relPathName>"
#             only if the environment variable $INS_USER is not set to SYSTEM
#
#   bobFindVlt: locates a file by its relative path.  The file given as a
#      relative path is searched in the same directories as the seq_findFile
#      command does, i.e. (in the order as given here):
#	1. in the current working directory: "./<relPathName>"
#	2. in the current module area:       "../<relPathName>"
#	3. in the $MODROOT:                  "$MODROOT/<relPathName>"
#	4. in the $INTROOT:                  "$INTROOT/<relPathName>"
#	5. in the $VLTROOT:                  "$VLTROOT/<relPathName>"
#
#   bobFindSound: locates a sound file, given its filename, without pathname, 
#      and with or without ".au" extension, under the standard VLT path.
#
#   bobFindSeq: locates a sequencer script belonging to a template, given its 
#      filename, without pathname, and with or without ".seq" extension.  The
#      <insMode> argument indicates the instrument mode, so the INS-
#      subdirectories that will be searched will include both the <insMode> and
#      COMMON branches.  If <insMode> is not specified, only the COMMON branch
#      will be searched. Within that scheme, first the TEMPLATES/SEQ 
#      subdirectory is searched, then (for backward compatibility reasons) the
#      SEQUENCES subdirectory.
#
#   bobFindTsf: locates a template signature file, given its filename, without
#      pathname, and with or without ".tsf" extension.  The <insMode> argument 
#      indicates the instrument mode, so the INS-subdirectories that will be 
#      searched will include both the <insMode> and COMMON subdirectory.  If 
#      <insMode> is not specified, only the COMMON subdirectory will be 
#      searched. Within that scheme, first the TEMPLATES/TSF subdirectory is
#      searched, then (for backward compatibility reasons) the SEQUENCES
#      subdirectory.
#
#   bobFindObd: locates an observation description file, given its filename,
#      without pathname, and with or without ".obd" extension.  The <insMode> 
#      argument indicates the instrument mode, so the INS-subdirectories that
#      will be searched will include both the <insMode> and COMMON subdirectory.
#      If <insMode> is not specified, only the COMMON subdirectory will be 
#      searched. Within that scheme, first the TEMPLATES/OBD subdirectory is
#      searched, then (for backward compatibility reasons) the SEQUENCES
#      subdirectory.
#
#   RETURN VALUES
#   The absolute path of the first found file is returned, or an empty string 
#   if none was found.
#
#   The file must be readable for the calling user, otherwise it will be 
#   ignored even if it exists.
#
#   FILES
#   The specified files are accessed by the procs in read-only mode. They must
#   be readable by the calling user.
#
#   ENVIRONMENT
#   INS_ROOT - path to INS root
#   INS_USER - name of user, corresponding to a subdirectory under $INS_ROOT
#
#   CAUTIONS
#
#   EXAMPLES
#
#   SEE ALSO
#   vccFindLib(n)
#
#------------------------------------------------------------------------
#
#
# bobFindIns <relPath>    => <AbsolutePath> | ""
#
proc bobFindIns { relPath } {
    set areas [set ::bob::searchListIns]
    foreach area $areas {
	set absPath ${area}/$relPath
	if {[file readable $absPath]} {
	    return $absPath;	# found in given area
	}
    }
    return {};			# nothing found
}

#
# bobFindVlt <relPath>    => <AbsolutePath> | ""
#
proc bobFindVlt { relPath } {
    if {[catch {seq_findFile $relPath} absPath]} {
	return {};	# not found
    } else {
	if {[file readable $absPath]} {
	    return $absPath;	# found in given area and readable
	} else {
	    return {};		# not readable
	}
    }
}

#
# bobFindSound <soundFile>    => <AbsolutePath> | ""
#
proc bobFindSound { snd } {
    if {![string equal [file extension $snd] .au]} {
	set snd ${snd}.au
    }
    
    set absPath [bobFindVlt $snd]
    if {[string length $absPath] != 0} {
	return $absPath
    }

    return [bobFindVlt sounds/$snd]
}

#
#    bobFindSeq <template>	<mode> 	==> <AbsolutePath> | ""
#
# Special deal for GT.seq: this should be under $VLTROOT/lib/libbob.tcl !!
proc bobFindSeq { tpl {mode {}} } {
    
    if {![string equal [file extension $tpl] .seq]} {
	set tpl ${tpl}.seq
    }

    # Now look in the INS-tree, under the particular mode, plus COMMON
    foreach mode "$mode COMMON" {
	set absPath [bobFindIns $mode/TEMPLATES/SEQ/$tpl]
	if {[string length $absPath] != 0} {return $absPath}

	# For backward compatibility: also look under SEQUENCES subdir
	set absPath [bobFindIns $mode/SEQUENCES/$tpl]
	if {[string length $absPath] != 0} {return $absPath}
    }
    if {[string equal $tpl GT.seq]} {return [seq_findFile lib/libbob.tcl/GT.seq]}
    return {};	# getting here means we didn't find it
}

#
#    bobFindTsf <templateSigFile> <mode>	=> <AbsolutePath> | ""
#
# Special deal for GT.tsf: this should be under $VLTROOT/config !!
proc bobFindTsf { tsf {mode {}} } {

    if {![string equal [file extension $tsf] .tsf]} {
	set tsf ${tsf}.tsf
    }
    
    # Now look in the INS-tree, under the particular mode, plus COMMON
    foreach mode "$mode COMMON" {
	set absPath [bobFindIns $mode/TEMPLATES/TSF/$tsf]
	if {[string length $absPath] != 0} {return $absPath}
	# For backward compatibility: also look under SEQUENCES subdir
	set absPath [bobFindIns $mode/SEQUENCES/$tsf]
	if {[string length $absPath] != 0} {return $absPath}
    }
    if {[string equal $tsf GT.tsf]} {return [seq_findFile config/GT.tsf]}
    return {};	# getting here means we didn't find it
}

#
#    bobFindObd <obd>	<mode> 	==> <AbsolutePath> | ""
#
proc bobFindObd { obd {mode {}} } {
    
    if {![string equal [file extension $obd] .obd]} {
	set obd ${obd}.obd
    }

    # Now look in the INS-tree, under the particular mode, plus COMMON
    foreach mode "$mode COMMON" {
	set absPath [bobFindIns $mode/TEMPLATES/OBD/$obd]
	if {[string length $absPath] != 0} {return $absPath}

	# For backward compatibility: also look under SEQUENCES subdir
	set absPath [bobFindIns $mode/SEQUENCES/$obd]
	if {[string length $absPath] != 0} {return $absPath}
    }
    return {};	# getting here means we didn't find it
}

# We include some code here which sets up some variables.  As the file will
# be sourced in at some point, this piece of code will then execute.
namespace eval bob {
    global ::env searchListIns
    
    set searchListIns {}
   
    if {[info exists env(INS_ROOT)]} {
	if {[info exists env(INS_USER)] && ![string equal $env(INS_USER) SYSTEM]} {
	    lappend searchListIns $env(INS_ROOT)/$env(INS_USER)
	}
	lappend searchListIns $env(INS_ROOT)/SYSTEM
    }
}

# ___oOo___



