#*******************************************************************************
# E.S.O. - VLT project
#
# "@(#) $Id: bobGuiLib.tcl,v 1.145 2005/10/27 09:07:11 vltsccm Exp $"
#
# who       when        what
# --------  --------    ----------------------------------------------
# eallaert  14/11/1996  created
# mbec      09/04/1997  extended with several GUI functions
# mbec      10/07/1997  header update
# eallaert  1998-03-19  fixed the busy-cursor problem
# eallaert  1998-04-07  changed bobRepeat
# eallaert  1998-09-25  check if file selected is regular
# eallaert  1999-08-03  changed name to bobGuiLib.tcl
# eallaert  1999-11-15  added option -nonewline to seq_msgRecvReply in
#                       bobNextObsBlocks (SPR 990445)
# eallaert  2001-06-01  translate $gvar(tplEvents) into proper offline flag
# eallaert  2001-10-31  allow to retrieve config settings (see setOrGetVar)
# eallaert  2001-11-13  "bobSoundOn" now part of configfile (SPR 20010582)
#                       read-access to config-vars (SPR20010623)
# eallaert  2002-10-10  added bobSnapshot (SPR 20020511)
# eallaert  2003-04-28  reset error stack before starting OB (SPR 20030142)
# eallaert  2004-01-15  added more stuff to bobSnapshot (SPR20030520)
# eallaert  2005-03-08  "bobSoundServer" now part of configfile
#

#************************************************************************
#   NAME
#   bobAbort, bobContinue, bobPause, bobPrint, bobRepeat, bobStart,
#   bobNextObsBlocks, bobReset, bobSaveConfig, bobSnapshot
#       - several procedures used from GUI level to control bob
# 
#   SYNOPSIS
#   bobAbort
#   bobContinue
#   bobPause
#   bobPrint
#   bobRepeat
#   bobStart ?<contFlag>?
#   bobNextObsBlocks <nr> ?<file>?
#   bobReset
#   bobSaveConfig
#   bobSaveConfigAs ?<file>?
#   bobSnapshot
#
#   DESCRIPTION
#   bobAbort: abort execution of running OB
#   bobContinue: equivalent to bobStart <1>
#   bobPause: pause execution of running OB
#   bobPrint: print logs of currently selected OB
#   bobRepeat: request a MUSTREPEAT for selected object
#   bobStart ?<contFlag>?: start execution of selected/paused OB; if
#       <contFlag> is true, then continue after pause instead of start.
#       Default for <contFlag> is false.
#   bobNextObsBlocks <nr> ?<fileName>?: request <nr> OBs from scheduler process.
#       if <nr> is set to the word 'file', then either pop up a file-selection
#       panel (if <fileName> is not set), or try to load <fileName>.
#   bobReset: reset objects and interface
#   bobSaveConfig : save the current config parameters
#   bobSaveConfigAs : save the current config parameters into <file>. If <file>
#       is not specified, bring up a file-selection box first.
#   bobSnapshot : dump bob-related info and logfiles to a timestamped directory.
#       The name of this latter directory is the return value.
#
#   FILES
#
#   ENVIRONMENT
#
#   RETURN VALUES
#
#   CAUTIONS
#
#   EXAMPLES
#
#   SEE ALSO
#
#   BUGS     
#
#------------------------------------------------------------------------
# For internal use:
#   NAME
#   _bobBuildObsBlocks, _bobParseOB 
# 
#   SYNOPSIS
#   _bobBuildObsBlocks <obdSource> <blocks>
#   _bobParseOB <blocks>
#   resetButtons <startButton> <pauseButtonFunction> <pauseButton> \
#		<abortButton> <resetButton>
#
#   DESCRIPTION
#   _bobBuildObsBlocks <obdSource> <blocks>: build the blocks from rawdata
#   _bobParseOB <blocks>: returns a list of OB's raw data
#   resetButtons <startButton> <pauseButtonFunction>\
#		<pauseButton> <abortButton> <resetButton>: configure the GUI 
#               buttons
#
#

#=============================================================================
proc ::bob::bobAbort { {reason {}} } {

    set runningOB [$::ROOT activeChild]
    if { $runningOB == ""} {
	bobDialog "Cannot abort - there is no OB active!" \
		application 5000
	return
    }
    $runningOB abort $reason
}
#============================================================================
proc ::bob::bobContinue { } {
    internal::_bobStart 1
}

#============================================================================
proc ::bob::bobSnapshot { } {
    global seq_ccsStatus seq_ccsProcName errorInfo

    set tclErrorInfo $errorInfo;	# before it gets overwritten
    if {![info exists ::bob::internal::tplErrorInfo]} {
	set ::bob::internal::tplErrorInfo {}
    }
    #########
    # 1. Create the directory /vltdata/tmp/<bobProcName>.<timestamp>
    #########
    set timestamp [seq_isoTime]
    if {[info exists seq_ccsStatus]} {
	# means ccsInit was given before - seq_ccsProcName contains right CCS-name
	set bobName $seq_ccsProcName
    } else {
	# means env not up and running - use proper default name
	set bobName bob
    }
    set dumpDir /vltdata/tmp/$bobName.$timestamp
    if {[catch {file mkdir $dumpDir} msg]} {
	bobDialog "Could not create the directory $dumpDir:\n\
		--------------------------------\n\
		${msg}\n\
		--------------------------------\n\
		Please correct and try again." \
		application
	return
    }

    ########
    # 2. Save the currently running or last executed *.obd file.
    ########
    if {[string length $::ROOT] != 0} {
	set runningOB [$::ROOT activeChild]
    } else {
	set runningOB {}
    }
    if {[string length $runningOB] == 0} {
	bobDialog "There is no OB active, so no OB saved" \
		application 5000
    } else {
	if {[info exists ::gvar(bobObdFilePath)]} {
	    # save these vars, as they can be overwritten by bobSaveAs
	    set bobObdFilePath $::gvar(bobObdFilePath)
	    set bobObdFileName $::bob::internal::info(obdFile)
	}
	::bob::bobSaveAs $dumpDir/active.obd
	if {[info exists bobObdFilePath]} {
	    # Restore the proper values
	    set ::gvar(bobObdFilePath) $bobObdFilePath
	    set ::bob::internal::info(obdFile) $bobObdFileName
	}
    }
    ########
    # 3. Make a copy of /vltdata/tmp/logFile and place it in the above
    #    directory; same procedure for most recent previous logFile.
    ########
    set lastBackup {}
    foreach backup [glob -nocomplain -- /vltdata/tmp/logFile*_*] {
	if {[string length $lastBackup] == 0 || \
		[file atime $backup] > [file atime $lastBackup]} {
	    set lastBackup $backup
	}
    }
    foreach file "/vltdata/tmp/logFile $lastBackup" {
	file copy -force -- $file $dumpDir
    }
	
    ########
    # 4. Make a copy of /vltdata/tmp/bob_<OS>.log into the above directory
    ########
    catch {file copy -force -- $::env(VLTDATA)/tmp/bob_[lindex $::bob::internal::config(bobProcNameOS) 0].log $dumpDir}

    ########
    # 5. Create a text file with a time-stamp, specifying what template 
    # of the OB is being executed, and what exposure is being taken within
    # the template. Also, save the tplLog of the OB.
    # 6. If an error happened in BOB, then append to the text file above:
    #  a) the Tcl error message,
    #  b) the complete CCS error stack
    # 7. Prompt for a motive
    ########
    set PC         undefined
    set runningTPL none
    set tplId      undefined
    set expNr      undefined
    set log        empty
    if {[string length $runningOB] != 0} {
	set PC [expr {[$runningOB getPC] + 1}]
	set runningTPL [$runningOB activeChild]
	if {[string length $runningTPL] == 0} {
	    set PC         "(OB finished)"
	    set runningTPL "(OB finished)"
	    set tplId      "(OB finished)"
	    set expNr      "(OB finished)"
	} else {
	    set tplId [$runningTPL getTplId]
	    catch {set expNr [$runningTPL getExpNr]}
	    set log [$runningOB getLogText]
	}
    } else {
	set PC         undefined
	set runningTPL none
	set tplId      undefined
	set expNr      undefined
	set log        empty
    }
    set errorStack {}
    set stackSize [seq_errGetStackSize]
    for {set i 1} {$i <= $stackSize} {incr i} {
	append errorStack [seq_errGetFromStack $i]\n
    }
    set ans [bobPromptDialog "Please enter motive to take this snapshot" \
	    application 0 info]
    if {[lindex $ans 0] == 1} {
	set motive [lindex $ans 1]
    } else {
	set motive "(Nothing entered as a motive)"
    }
    write_file ${dumpDir}/bobSnapshot \
	    "bobSnapshot - $timestamp" \
	    "*********************************" \
	    "" \
	    "motive for snapshot: $motive" \
	    "" \
	    "running TPL number: $PC" \
	    "running TPL ID: $tplId" \
	    "running exposure number: $expNr" \
	    "" \
	    "last CCS error stack:" \
	    "--------------------" \
	    "$errorStack" \
	    "" \
	    "Tcl errorInfo at last template script problem:" \
	    "---------------------------------------------" \
	    "$::bob::internal::tplErrorInfo" \
	    "" \
	    "Current Tcl errorInfo:" \
	    "---------------------" \
	    "$tclErrorInfo" \
	    "" \
	    "OB template logs:" \
	    "----------------" \
	    "$log"
    
    ########
    # 8) Report via a popup panel the name of the created directory that
    #   contains the above dump (the operator should place this name in
    #   the PPRS).    
    ########
    bobDialog "Snapshot saved in: $dumpDir\n\n\
	    Please record this directory name\n\
	    when submitting a problem report." application 0 info

    ### 9) Execute a user supplied command, registered e.g. in the ~/.bobrc 
    ###   file. It shall be executed from the directory that contains this
    ###   new dump (e.g. to retrieve specific OLDB values, etc. and
    ###   evtl. to transfer all data to a different host).
    
    return $dumpDir
}

#=============================================================================
proc ::bob::bobNextObsBlocks {nr {file {}}} {
    switch -- $nr file {
	#!!#
	##puts "got a request for a file"
	if {[string length $file] == 0} {
	    if {[info exists ::gvar(bobObdFilePath)]} {
		set example $::gvar(bobObdFilePath)/.obd
	    } else {
		set example .obd
	    }
	    #!!#
	    ##puts "calling bobFileSel with example =  $example"
	    set file [bobFileSel {Load Observation Block Description} \
		    $example file]
	    #!!#
	    ##puts "returned - selected file = >$file<"
	    if {[string length $file] == 0} {
		return
	    }		
	} else {
	    if {![string equal [file pathtype $file] absolute]} {
		set keepFile $file
		if {[string length [set file [bobFindObd $file]]] == 0} {
		    bobDialog "Could not locate file\n\n$keepFile\n\n\
			    using the standard search paths." application
		    return
		}
	    }
	}
	# keep filepath for future calls
	set ::gvar(bobObdFilePath) [file dirname $file]
	set ::bob::internal::info(obdFile) $file
	set ::gvar(bobObdFileName) $file
	if {[catch {read_file $file} blocks]} {
	    bobDialog "Could not read file\n\n$file\n\n\
		    Check its access permissions." application
	    return
	}	    
	set obdSource file
    } default {
	# needed in case some of the config vars refer to environment vars:
	global env
	set offline [expr {$::bob::internal::config(tplEvents) ? "false" : "true"}]
	#set offline [expr {$::bob::internal::config(tplEvents) ? "" : "true"}]
	if {[catch {seq_msgSendCommand -nocheck [subst $::bob::internal::config(bobEnvNameOH)] \
		[subst [lindex $::bob::internal::config(bobProcNameOH) 0]] SCRIPT \
		"NextObsBlocks $nr $offline"} cmdId]} {
	    bobDialog "Could not communicate with OH.\n\
		    Check the Environment settings (under Configure menu)\n\
		    and the CCS error-log." application
	    seq_errDisplay -suspend
	    return
	}

	# the -nonewline option only exists from seq 2.40 onwards
	if {[infox appversion] > 2.39} {
	    set nnl -nonewline
	} else {
	    set nnl -all;	# repeating this option won't hurt
	}
	switch -- $::bob::internal::progressBar {} {} default {
	    set ::bob::internal::progress 5
	    $::bob::internal::progressBar -text "Fetching OB from [lindex $::bob::internal::config(bobProcNameOH) 0] ..."
	    wm deiconify $::bob::internal::progressBar
	}
	if {[catch {seq_msgRecvReply -all -ascii $nnl $cmdId errNr} blocks]} {
	    switch -- $::bob::internal::progressBar {} {} default {
		wm withdraw $::bob::internal::progressBar
	    }
	    bobDialog "Could not retrieve observation blocks from OH.\n\
		    There was a problem trying to receive replies from OH.\n\
		    Check the CCS error-log." application
	    seq_errDisplay -suspend
	    return
	}

	if {$errNr != 0} {
	    switch -- $::bob::internal::progressBar {} {} default {
		wm withdraw $::bob::internal::progressBar
	    }
	    bobDialog "Got an error returned by OH.\n\
		    Check the CCS error-log." application
	    seq_errDisplay -suspend
	    return
	}

	if {[string length $blocks] == 0} {
	    switch -- $::bob::internal::progressBar {} {} default {
		wm withdraw $::bob::internal::progressBar
	    }
	    bobDialog "OH is not ready to return observation blocks.\n\
		    Please prepare the schedule (select OBs) first in OH\n\
		    before requesting OBs from within Bob." application
	    return
	}
	set obdSource OH
	set ::bob::internal::info(obdFile) {}
	set ::gvar(bobObdFileName) {}
	##puts "received from schedule: $blocks"

    }
    # Update the canvas label, if the source has changed
    # (Remark that thanks to callbacks, these labels will get modified if the
    # configuration for procNamdOS/OH is modified).
    if {![string equal $obdSource $::bob::internal::info(obdSource)]} {
	#!!#
	##puts "updating the canvas label"
	set ::bob::internal::info(obdSource) $obdSource
	internal::_bobUifSetCanvasLabel both
    }
    internal::_bobBuildObsBlocks $obdSource $blocks 

    #!# LOAD keyword is on 2004-01-15 not yet approved!
    #!# See SPR 20040019
    seq_logFitsAction OBS OBS EXEC LOAD

}

#============================================================================
proc ::bob::bobPause { } {
    
    # disable further pause-button presses (action and toolbar buttons)
    set ::gvar(pauseOBStatus) 0
    $::gvar(vtb) itemconfigure hand_up -state disabled
    
    update
    set runningOB [$::ROOT activeChild]
    if { $runningOB == ""} {
	bobDialog "Cannot pause - there is no OB active!" \
		application 5000
	return
    }
    $runningOB pause
}
#=============================================================================
# The MUSTREPEAT action only operates on an OB, not on individual TPLs
proc ::bob::bobRepeat { {reason {}} } {
    set marked [BobBaseObject::marked]
    if {[string length $marked] == 0} {
	bobDialog "You have to select an OB by clicking on it\n\
		before requesting to repeat it." application 10000
	return
    } elseif {[$marked isa BobTPL]} {
	set marked [$marked up];	# get parent OB of TPL
    }
    $marked mustRepeat $reason
}

#=============================================================================
proc ::bob::bobStart { } {
    internal::_bobStart 0
}

proc ::bob::internal::_bobStart { {cont 0} } {
    
    # if it's a fresh start ...
    if {!$cont} {
	# check if ROOT exist
	switch -- $::ROOT {} {
	    bobDialog "You have to load an Observation Block\
		    before it can be started" application 5000
	    return
	}

	# Test if anything is selected, or by default execute first
	if {[BobBaseObject::markedIsOB]} {
	    set runningOB [BobBaseObject::marked]
	} else {
	    # if no root - please load blocks - thanks to catch
	    if {[catch {$::ROOT resetPC ; set runningOB [$::ROOT getNextChild]}]} {
		set runningOB ""
	    }
	    switch -- $runningOB {} {
		bobDialog "You have to select an Observation Block\
			before it can be started" application 5000
		return
	    }
	}
	
	# tell ROOT who is running
	$::ROOT setActiveChild $runningOB
    } else {
	set runningOB [$::ROOT activeChild]
	# Test if any OB is running
	switch -- $runningOB {} {
	    bobDialog "Cannot continue - there is no OB paused!" \
		    application 5000
	    return
	}
    }
    
    # Start the execution of the templates of the selected OB.
    set loop 1
    while {[string length $runningOB] != 0 && $loop} {
	if {$::bob::internal::config(bobRunAll)} {
	    $::ROOT setPC $runningOB
	    while {[string length $runningOB] != 0 && $loop} {
		seq_errResetStack
		seq_errResetStack -internal
		seq_errResetStack -reply		
		$::ROOT setActiveChild $runningOB
		set execStatus [$runningOB execute]
		# Do not get next OB if anything went wrong or PAUSED
		if {$execStatus==0} {
		    # this OB executed successfully (done or skipped)
		    # I get next
		    set runningOB [$::ROOT getNextChild]
		    if {[string length $runningOB] == 0 && $::bob::internal::config(bobGetNext)} {
			# if OH undefined, and current OBD is from file, repeat file
			if {[string length $::bob::internal::config(bobProcNameOH)] == 0 &&
			    [string equal $::bob::internal::info(obdSource) file]} {
			    ::bob::bobReset
			} else {
			    ::bob::bobNextObsBlocks 1
			}
			$::ROOT resetPC
			set runningOB [$::ROOT getNextChild]
			if {!$::bob::internal::config(bobRunAll)} {
			    set loop 0
			}
		    }
		} else {
		    # not a normal termination (paused, fault or abort)
		    # I leave the loop
		    set loop 0;
		}
	    }
	} else {
	    # execute 1 OB, and then break out (resetting "loop" flag)
	    $::ROOT setActiveChild $runningOB
	    set execStatus [$runningOB execute]
	    # Do not get next OB if anything went wrong or PAUSED
	    if {$execStatus==0} {
		if {$::bob::internal::config(bobGetNext)} {
		    # if OH undefined, and current OBD is from file, repeat file
		    if {[string length $::bob::internal::config(bobProcNameOH)] == 0 &&
			[string equal $::bob::internal::info(obdSource) file]} {
			::bob::bobReset
		    } else {
			::bob::bobNextObsBlocks 1
		    }
		    $::ROOT resetPC
		    set runningOB [$::ROOT getNextChild]
		}
	    }
	    set loop 0
	}
    }
    
    # if OB executed successfully (done or skipped), hilite next automatically
    if {$execStatus == 0} {
	set runningOB [$::ROOT getNextChild]
	# catch because it could be the last one already
	catch {$runningOB mark}
    } elseif {$execStatus != 2} {
	# ie aborted or fault (not paused); automatically take the next OB 
	set runningOB [$::ROOT getNextChild]
	# but don't high-light it !
    }
}

#============================================================================
proc ::bob::bobSaveConfigAs {{file {}}} {
    if {[string length $file] == 0} {
	set example [file join [file dirname \
				    [seq_relToAbsPath $::bob::internal::info(configFile)]] .bobrc*]
	set file [bobFileSel "Save configuration settings" $example]
	if {[string length $file] == 0} {
	    return
	}
    }
    set ::bob::internal::info(configFile) $file
    ::bob::bobSaveConfig
}
    
proc ::bob::bobSaveConfig {} {
    # Relative paths are relative to the HOME dir!
    if {[string equal [file pathtype $::bob::internal::info(configFile)] relative]} {
	set cfgFile ~/$::bob::internal::info(configFile)
    } else {
	set cfgFile $::bob::internal::info(configFile)
    }
   
    write_file $cfgFile  \
	"# $cfgFile - Configuration file for bob" \
	"#"\
	"# DO NOT EDIT BY HAND!!" \
	"#" \
	"# Generated at:" \
	"set timestamp [seq_isoTime]" \
	"# (the variable 'timestamp' will not be seen by the user)" \
	"#" \
	"# Written with:" \
	"set bobVersion $::bob::internal::info(bobVersion)" \
	"# (the variable 'bobVersion' will not be seen by the user)" \
	"#" \
	"# From configuration panel:" \
	"# ------------------------" \
	"::bob::bobEnvNameOH          [list $::bob::internal::config(bobEnvNameOH)]" \
	"::bob::bobEnvNameOHlist      [list $::bob::internal::config(bobEnvNameOHlist)]" \
	"::bob::bobProcNameOH         [list $::bob::internal::config(bobProcNameOH)]" \
	"::bob::bobProcNameOHlist     [list $::bob::internal::config(bobProcNameOHlist)]" \
	"::bob::bobEnvNameOS          [list $::bob::internal::config(bobEnvNameOS)]" \
	"::bob::bobEnvNameOSlist      [list $::bob::internal::config(bobEnvNameOSlist)]" \
	"::bob::bobProcNameOS         [list $::bob::internal::config(bobProcNameOS)]" \
	"::bob::bobProcNameOSlist     [list $::bob::internal::config(bobProcNameOSlist)]" \
	"" \
	"::bob::bobRunAll             $::bob::internal::config(bobRunAll)" \
	"::bob::bobGetNext            $::bob::internal::config(bobGetNext)" \
	"::bob::bobAutoUnfold         $::bob::internal::config(bobAutoUnfold)" \
	"::bob::bobAutoCreateTpl      $::bob::internal::config(bobAutoCreateTpl)" \
	"" \
	"::bob::tplEvents             $::bob::internal::config(tplEvents)" \
	"" \
	"::bob::tplSim                $::bob::internal::config(tplSim)" \
	"::bob::tplVerbose            $::bob::internal::config(tplVerbose)" \
	"::bob::tplInternals          $::bob::internal::config(tplInternals)" \
	"::bob::tplSimList            [list $::bob::internal::config(tplSimList)]" \
	"::bob::tplSpeDelay           $::bob::internal::config(tplSpeDelay)" \
	"::bob::tplDefDelay           $::bob::internal::config(tplDefDelay)" \
	"" \
	"::bob::tplSkipList           [list $::bob::internal::config(tplSkipList)]" \
	"" \
	"::bob::bobUserLevel          $::bob::internal::config(bobUserLevel)" \
	"" \
	"::bob::tplCallBack           [list $::bob::internal::config(tplCallBack)]" \
	"::bob::bobCallBack           [list $::bob::internal::config(bobCallBack)]" \
	"::bob::bobCfgCallBack        [list $::bob::internal::config(bobCfgCallBack)]" \
	""  \
	"::bob::tplLibList            [list $::bob::internal::config(tplLibList)]" \
	""  \
	"::bob::bobPrintCmd           [list $::bob::internal::config(bobPrintCmd)]" \
	""  \
	"::bob::bobPlugin             [list $::bob::internal::config(bobPlugin)]" \
	"::bob::bobPluginFlag         $::bob::internal::config(bobPluginFlag)" \
	"" \
	"# From preferences panel:" \
	"# ----------------------" \
	"::bob::bobTitleFont          [list $::bob::internal::config(bobTitleFont)]" \
	"::bob::bobCanvasFont         [list $::bob::internal::config(bobCanvasFont)]" \
	"::bob::bobLogFont            [list $::bob::internal::config(bobLogFont)]" \
	"::bob::bobDialogFont         [list $::bob::internal::config(bobDialogFont)]" \
	"::bob::bobEditorFont         [list $::bob::internal::config(bobEditorFont)]" \
	"" \
	"::bob::bobSoundOn            $::bob::internal::config(bobSoundOn)" \
	"::bob::bobSoundServer        [list $::bob::internal::config(bobSoundServer)]" \
	"::bob::bobSoundVerify        $::bob::internal::config(bobSoundVerify)" \
	"::bob::bobSoundVolume        $::bob::internal::config(bobSoundVolume)" \
	"::bob::bobSoundSetVolAtInit  $::bob::internal::config(bobSoundSetVolAtInit)" \
	"::bob::bobSoundSetVolAtPlay  $::bob::internal::config(bobSoundSetVolAtPlay)" \
	"::bob::bobErrorSound         [list $::bob::internal::config(bobErrorSound)]" \
	"::bob::bobWarningSound       [list $::bob::internal::config(bobWarningSound)]" \
	"::bob::bobBreakpointSound    [list $::bob::internal::config(bobBreakpointSound)]" \
	"::bob::bobTplSound           [list $::bob::internal::config(bobTplSound)]" \
	"::bob::bobObSound            [list $::bob::internal::config(bobObSound)]"  \
	"" \
	"::bob::bobWaitOptions        [list $::bob::internal::config(bobWaitOptions)]"
}
#============================================================================
proc ::bob::bobReset { } {
    BobBaseObject::resetStatus
    internal::resetButtons
}
#============================================================================
# The action buttons are always present and so always set.
# Toolbar buttons are done only in 0 mode = execution mode
proc ::bob::internal::resetButtons {{startB 1} {pauseBLabel pause} {pauseB 0} {abortB 0} {repeatB 0} {resetB 1} } {
    if {$startB} {
	set ::gvar(execOBStatus) 1
	$::gvar(vtb) itemconfigure telescope -state normal
    } else {
	set ::gvar(execOBStatus) 0
	$::gvar(vtb) itemconfigure telescope -state disabled
    }
    
    # pauseBLabel = unknown .... don't bother
    if {[string equal $pauseBLabel pause]} {
	# caught for the bobWish script where I don't create a dummy 
	catch {$::bcvbob.pauseOB configure -label "Pause" -command {::bob::bobPause}}
	set rememberB hand_up
	set forgetB hand
    } elseif {[string equal $pauseBLabel continue]} {
	# caught for the bobWish script where I don't create a dummy
	catch {$::bcvbob.pauseOB configure -label "Continue" -command {::bob::bobContinue}}
	set rememberB hand
	set forgetB hand_up
    } elseif {[string equal $pauseBLabel auto]} {
	# check the current setting
	catch {$::bcvbob.pauseOB cget -label} Current
	if {[string equal $Current Continue]} {
	    set rememberB hand
	    set forgetB hand_up
	} elseif {[string equal $Current Pause]} {
	    set rememberB hand_up
	    set forgetB hand
	}
    }
    
    if {$pauseB} {
	set ::gvar(pauseOBStatus) 1
	$::gvar(vtb) itemconfigure $rememberB -state normal
	$::gvar(vtb) itemconfigure $forgetB -state disabled
    } else {
	set ::gvar(pauseOBStatus) 0
	$::gvar(vtb) itemconfigure hand -state disabled
	$::gvar(vtb) itemconfigure hand_up -state disabled
    }
    
    if {$abortB} {
	set ::gvar(abortOBStatus) 1
	$::gvar(vtb) itemconfigure abort -state normal
    } else {
	set ::gvar(abortOBStatus) 0
	$::gvar(vtb) itemconfigure abort -state disabled
    }
    
    if {$repeatB} {
	set ::gvar(repeatOBStatus) 1
	$::gvar(vtb) itemconfigure repeat -state normal
    } else {
	set ::gvar(repeatOBStatus) 0
	$::gvar(vtb) itemconfigure repeat -state disabled
    }
    
    # resetButton= -1 .... don't bother
    if {$resetB==1} {
	set ::gvar(resetStatusStatus) 1
    } elseif {$resetB==0} {
	set ::gvar(resetStatusStatus) 0
    }
    
}

#============================================================================
# Test if anything is selected before printing
proc ::bob::bobPrint { } {
    set lsel [BobBaseObject::marked]
    if {[string length $lsel] == 0} {
	bobDialog "Select OB before printing !" \
		application 5000
	return
    }
    foreach sel $lsel {
	$sel printLog
    }
}

#============================================================================
proc ::bob::bobInfo {} {
    set obl [lsort [itcl::find objects -class BobOB]]
    set xyz {}
    foreach ob $obl {
	lappend xyz "$ob [$ob cget -status]" 
    }
    return $xyz
}
#============================================================================

# Create a public command that imports the above defined commands into the
# global namespace.
proc bobMakeUifProcsGlobal {} {
    # Make sure the bobSave* commands exist at this point in time
    auto_load ::bob::bobSave
    namespace eval :: {
	if {[string length [info command bobCallBack]] == 0} {
	    namespace import ::bob::bob* ::bob::tpl*
	    foreach status [array names ::bob::internal::info] {
		# cmds with tpl* or bob* prefix have been imported already
	        if {![string match bob* $status] &&
		    ![string match tpl* $status]} {
		    namespace import ::bob::$status
		}
	    }
	}
    }
}

# declare the "alias function" for setting/getting a config var
proc ::bob::internal::setOrGetConfig {var} {
    set p [info level -1]
    ##set var [namespace tail [lindex $p 0]]
    set nrArgs [llength $p] 
    if {$nrArgs == 1} {
	set ::bob::internal::config($var)
    } elseif {$nrArgs == 2} {
	set ::bob::internal::config($var) [lindex $p 1]
    } else {
	error "wrong # args: should be \"$var ?<value>?\""
    }
}
#=============================================================================
# return a list of block's raw data
proc ::bob::internal::_bobParseOB { blocks } {
    set lob {}
    
    # Analyse blocks, show information in the OB-canvas
    set remain 1
    while {$remain} {

	# Determine if there is a "OBS.ID" at the start of a line,
	# i.e. possibly after a newline and only with whitespace in front,
	# followed by whitespace.
	# We'll check if there is in fact >1 OBs, by looking for multiple
	# occurences.  As we store the resulting substring-indices, we can find
	# out if there is more than 1 OB by looking at these index-pairs
	set pattern "\[\t \]*OBS\\.ID\[\t \]+"
	set pattern1 "^$pattern|\n$pattern"
	set pattern2 "\n$pattern"

	if {[regexp -indices -- ($pattern1) $blocks l1]} {
	    # there is at least 1 OB; get the substring range of the first
	    set idx1 [lindex $l1 0]
	    if {[string equal -length 1 [string index $blocks $idx1] \n]} {
		# the match starts with a newline, we want to skip this newline
		incr idx1
	    }
	    # Check if there are more of these patterns following further
	    if {[regexp -indices -- ($pattern2) \
		    [string range $blocks [expr {$idx1 +1}] end] l2]} {
		# yes, there are
		set idx2 [expr {[lindex $l2 0] + $idx1}]
	    } else {
		# set the substring range of the only remaining OB
		set idx2 end
		set remain 0
	    }
	} else {
	    bobDialog "The PAF file with the set of OBs has an\n\
		    illegal format: there is no OBS.ID" application
	    break;      # break out of the "while" loop
	}
	
	lappend lob [string range $blocks $idx1 $idx2]
	
	if {$remain} {
	    set blocks [string range $blocks $idx2 end]
	}
    }
    return $lob
}
#=============================================================================
proc ::bob::internal::_bobBuildObsBlocks {obdSource blocks} {
    
    # Determine if there is a PAF.HDR.START at the start of a line,
    # i.e. possibly after a newline and only with whitespace in front,
    # followed by whitespace and a newline or semi-colon.
    # There should be one, and only one, PAF header per set of OBs
    # We'll check if there is in fact >1 PAF headers, by looking for multiple
    # occurences.

    #!!#
    ##puts "building OB ..."

    set pattern "\[\t \]*PAF\\.HDR\\.START\[\t \]*\[;\n\]"
    set pattern1 "^$pattern|\n$pattern"
    set pattern2 "\n$pattern"
    if {[regexp -indices -- ($pattern1).*($pattern2) $blocks l1 l2 l3]} {
	# there is >1 PAF header
	bobDialog "The PAF file with the set of OBs has an illegal\n\
		format: it contains more than one PAF.HDR.START" application
        return

    } elseif {[regexp -indices -- ($pattern1) $blocks l1]} {
	# get the substring range of the first OB
	set OBidx [lindex $l1 0]
	if {[string equal -length 1 [string index $blocks $OBidx] \n]} {
	    # if the match starts with a newline, we want to skip this newline
	    incr OBidx
	}
	
    } else {
	bobDialog "The PAF file with the set of OBs has an\n\
		illegal format: it has no PAF.HDR.START" application
	return
    }

    # create the root point

    # ROOT has to be defined here (it is initialized at start-up)
    if {![string equal $::ROOT ""]} {
        delete object $::ROOT
    }
    # create root object
    # trick on icon set to not fold for GUI but will unfold anyway...
    
    set ::ROOT [namespace eval :: "BobVisualPCTree #auto $::gvar(cvOB) \
	    -label {Next observation blocks:} -icon root -fold_state nofold"]
    # The object name does not include the namespace, while for our
    # purposes we need to include that => leading double colon
    BobVisualTree::setRoot ::$::ROOT

    catch {$::ROOT busyCursor}
    #parse the OB and create them now
    set blockList [_bobParseOB $blocks]
    set nrBlocks [llength $blockList]
    for {set i 0} {$i < $nrBlocks} {incr i} {
	if {[string length $::bob::internal::progressBar] != 0 && \
		[string equal $obdSource OH]} {
	    $::bob::internal::progressBar configure -text "Analyzing OB # [expr {$i+1}]"
	}
	#!!#
	##puts "creating object of class OB ..."
	set obj [namespace eval :: "BobOB #auto {[lindex $blockList $i]} \
		$obdSource $::gvar(cvOB) $::gvar(stTpl)"]
	#!!#
	##puts "done creating object $obj"

	set obsId [$obj getObsKey ID]
	set obsName [$obj getObsKey NAME]
	#!# FILENAME keyword is on 2004-01-15 not yet approved!
	#!# See SPR 20040019
	if {[string equal $::bob::internal::info(obdSource) file]} {
	    seq_logFitsParRecord -string OBS OBS "" FILENAME $::bob::internal::info(obdFile)
	} else {
	    seq_logFitsParRecord -string OBS OBS "" ID $obsId
	}
	seq_logFitsParRecord -string OBS OBS "" NAME $obsName

	$::ROOT add ::$obj
	# only analyze OB at creation time if it is the first one, or if
	# configured to do it automatically
	if { $i == 0  || $::bob::internal::config(bobAutoCreateTpl)==1 } {
	    $obj populate
	}
    }
    $::ROOT refresh;	 # refresh display
    $::ROOT normalCursor
    resetButtons

    #!!#
    ##puts "done building OB"
}

#
# ___oOo___



