#*******************************************************************************
# E.S.O. - VLT project
#
# "@(#) $Id: bobLogger.tcl,v 1.145 2005/10/27 09:07:07 vltsccm Exp $"
#
# who       when        what
# --------  ----------  ----------------------------------------------
# eallaert  2005-06-10  don't insert log to textbox if it is not for this object
# eallaert  2000-10-10  differentiated between getLogText and getLogTextWithTags
# eallaert  2000-07-03  added logging procs/methods
#                       (openLogFile, logToDisk, closeLogFile)
# eallaert  1998-03-13  added method appendLog; cleanup
# mbec      05/05/97    header update
# mbec      25/04/1997  created

#************************************************************************
#   NAME
#   Logger - handles the logging of template in a scrolledtext widget
# 
#   SYNOPSIS
#   Logger <name>|#auto <widget> ?<args>?
# 
#   DESCRIPTION
#   This class handles all basic operations for logging purposes. It needs
#   a scrolled text widget as argument.
#
#   METHODS
#   constructor <sctxt> : a scrolled-text widget (see [incr Tcl])
#   destructor: destroy the object and the associated widget; see [incr Tcl]
#   configure: see [incr Tcl]
#
#   appendLog <logInfo>: append <logInfo> (as returned by getLogTextWithTags)
#        to the current log
#   clearLog: clear the window's content
#   deleteLog: clear the window's content and reset/unset log-related variables
#   displayLog: get the log and display it with the proper tags
#   getLogText: returns the log text without tag-info
#   getLogTextWithTags: returns the log text and tag-info
#   logLog <text> ?<color> <underline> <background>?: log text with specified
#	       style atttributes
#   logToDisk <header>: log the current object's log-text to disk, preceeded
#             by and underlined <header>.
#   refreshLog: refresh the display
#   printLog <header>: send the log text to printer spool.
#
#   PUBLIC PROCEDURES
#   openLogFile <fileName>: open <fileName> as logfile for "logToDisk"
#   closeLogFile
#
#   FILES
#
#   ENVIRONMENT
#
#   RETURN VALUES
#
#   CAUTIONS
#
#   EXAMPLES
#
#   SEE ALSO
#
#   BUGS     
#
#------------------------------------------------------------------------
#

#=============================================================================
#=============================================================================
itcl::class BobLogger {
    constructor {stwin args} {}
    destructor {}
    
    protected variable scrolledText {};

    public method appendLog {args} {};	# append logInfo as returned by getLog
    public method clearLog {} {};	# only clears the window
    public method deleteLog {} {};	# clears the whole records

    public method displayLog {args} {}
    public method getLogText {} {}
    public method getLogTextWithTags {} {}
    public method logLog {args} {}
    public method logToDisk {args} {}
    public method refreshLog {args} {}
    public method printLog {args} {}

    private variable clln 1;		# current log line number
    private variable logText {};	# log text for this template
    private variable logTags;		# associated log tags

    private common currentLog {};	# object with log currently displayed
    private common logFile {};		# name of the log-file
    private common logFileId {};	# fileId for the disc log-file

    public proc closeLogFile {args} {}
    public proc openLogFile {args} {}
}
#=============================================================================
#=============================================================================


# ------------------------------------------------------------------
#                        CONSTRUCTOR
# ------------------------------------------------------------------
itcl::body BobLogger::constructor {stwin args} {
    set scrolledText $stwin
    eval configure $args
}
# ------------------------------------------------------------------
#                        DESTRUCTOR
# ------------------------------------------------------------------
itcl::body BobLogger::destructor {} {
    deleteLog
}
# ------------------------------------------------------------------
#                        PUBLIC METHOD appendLog
# ------------------------------------------------------------------
itcl::body BobLogger::appendLog {logInfo} {
    set text [lindex $logInfo 0]
    if {[string length $text] == 0} return
    append logText ${text}\n

    set offset [expr {$clln - 1}]
    foreach attr [lrange $logInfo 1 end] {
	foreach {tag ranges} $attr {
	    foreach range $ranges {
		foreach {start end} $range {}
		set start [expr {$start + $offset}]
		set end   [expr {$end +   $offset}]
		lappend logTags($tag) "$start $end"
	    }
	}
    }
    set nrLines [llength [split $text \n]]
    set clln [expr {$clln + $nrLines}];   # new last line number
}

# ------------------------------------------------------------------
#                        PUBLIC METHOD clearLog
# ------------------------------------------------------------------
itcl::body BobLogger::clearLog {} {
    $scrolledText clear
}

# ------------------------------------------------------------------
#                        PUBLIC METHOD deleteLog
# ------------------------------------------------------------------
itcl::body BobLogger::deleteLog {} {
    clearLog
    foreach name [array names logTags] {
	unset logTags($name)
    }
    set logText {}
    set clln 1
}

# ------------------------------------------------------------------
#                        PUBLIC METHOD displayLog
# ------------------------------------------------------------------
itcl::body BobLogger::displayLog {} {
    clearLog
    $scrolledText insert end $logText
    
    # adjust the tags (foreground colours, underline, background)
    foreach tagName [array names logTags] {
	foreach range $logTags($tagName) {
	    eval $scrolledText tag add $tagName $range
	}
	if {[string equal $tagName underline]} {
	    $scrolledText tag configure underline -underline 1
	} elseif {[string first fg $tagName] == 0} {
	    $scrolledText tag configure $tagName \
		    -foreground [string range $tagName 2 end]
	} elseif {[string first bg $tagName] == 0} {
	    $scrolledText tag configure $tagName \
		    -background [string range $tagName 2 end]
	}
    }
    # the most interesting piece of info is most likely at the end
    $scrolledText see end
    set currentLog $this;	# keep track of whose logs are shown
    return
}
# ------------------------------------------------------------------
#                        PUBLIC METHOD getLogText
# ------------------------------------------------------------------
itcl::body BobLogger::getLogText {} {
    return $logText
}

# ------------------------------------------------------------------
#                        PUBLIC METHOD getLogTextWithTags
# ------------------------------------------------------------------
itcl::body BobLogger::getLogTextWithTags {} {
    set reply [list $logText]
    foreach name [array names logTags] {
	lappend reply "$name [list $logTags($name)]"
    }
    return $reply
}

# ------------------------------------------------------------------
#                        PUBLIC METHOD logLog
# ------------------------------------------------------------------
itcl::body BobLogger::logLog {text {colour {}} {underline 0} {backgd {}}} {
    set nrLines [llength [split $text \n]]
    if {$nrLines == 0} {incr nrLines};	# text is empty -> blank line
    set lln [expr {$clln + $nrLines}];	# last line number
    
    # Only display this if the log corresponding to this template is shown.
    if {[string equal $currentLog $this]} {
	$scrolledText insert end ${text}\n
	if {[string length $colour] != 0} {
	    $scrolledText tag add fg$colour ${clln}.0 ${lln}.0
	    if {![info exists logTags(fg$colour)]} {
		$scrolledText tag configure fg$colour -foreground $colour
	    }
	}
	if {$underline} {
	    $scrolledText tag add underline ${clln}.0 ${lln}.0
	    if {![info exists logTags(underline)]} {
		$scrolledText tag configure underline -underline 1
	    }
	}
	if {[string length $backgd] != 0} {
	    $scrolledText tag add bg$backgd ${clln}.0 ${lln}.0
	    if {![info exists logTags(bg$backgd)]} {
		$scrolledText tag configure bg$backgd -background $backgd
	    }
	}
	$scrolledText see end
    }

    # Now save it in the different vars, independent of the textbox's contents.
    append logText ${text}\n
    if {[string length $colour] != 0} {
	lappend logTags(fg$colour) "${clln}.0 ${lln}.0"
    }
    if {$underline} {
	lappend logTags(underline) "${clln}.0 ${lln}.0"
    }
    if {[string length $backgd] != 0} {
	lappend logTags(bg$backgd) "${clln}.0 ${lln}.0"
    }
    set clln $lln
}
# ------------------------------------------------------------------
#                        PUBLIC METHOD logToDisk
# ------------------------------------------------------------------
itcl::body BobLogger::logToDisk { header } {
    for {set i 0} {$i < [string length $header]} {incr i} {
	append underline -
    }
    puts $logFileId "${underline}\n${header}\n${underline}\n"
    puts $logFileId $logText
    flush $logFileId
}

# ------------------------------------------------------------------
#                        PUBLIC METHOD printLog
# ------------------------------------------------------------------
itcl::body BobLogger::printLog { header } {
    # needed in case (bobPrintCmd) refers to environment vars:
    global env
    set f1 [open |[subst $::bob::internal::config(bobPrintCmd)] w]
    puts $f1 $header
    puts $f1 $logText
    close $f1
}
# ------------------------------------------------------------------
#                        PUBLIC METHOD refreshLog
# ------------------------------------------------------------------
itcl::body BobLogger::refreshLog {} {
    foreach name [array names logTags] {
   	set logTags($name) $logTags($name)
    }
}
# ------------------------------------------------------------------
#                        PUBLIC PROCEDURE openLogFile
# ------------------------------------------------------------------
# No need for action if requesting to use the logfile which is already open.
# Otherwise, close current logfile, check if new logfile exists (if so,
# move it to ${fileName}.old), and create new logfile.
itcl::body BobLogger::openLogFile { fileName } {
    if {[string equal $fileName $logFile]} {
	return
    }
    if {[string length $logFileId] != 0} {
	close $logFileId
    }
    set logFile $fileName
    if {[file isfile $logFile]} {
	file rename -force -- $logFile $logFile.old
    }
    set logFileId [open $logFile w]
}

# ------------------------------------------------------------------
#                        PUBLIC PROCEDURE closeLogFile
# ------------------------------------------------------------------
itcl::body BobLogger::closeLogFile { } {
    if {[string length $logFileId] != 0} {
	close $logFileId
    }
    set logFile {}
    set logFileId {}
}

#
# ___oOo___




