#*******************************************************************************
# E.S.O. - VLT project
#
# "@(#) $Id: bobMobsBase.tcl,v 1.145 2005/10/27 09:07:11 vltsccm Exp $"
#
# who       when        what
# --------  ----------  ----------------------------------------------
# imunoz    30/09/97    created
# imunoz    30/10/97    added -arrayvar
# imunoz    21/11/97    man page added
# eallaert  1998-05-15  added INTROOT to auto_path (for pkg TkTable)
# eallaert  1998-06-15  completely reworked
# eallaert  1998-09-14  added fixed column for loop counter (SPR 980466)
# eallaert  1999-07-22  now also using oslx for NOCCS
# eallaert  2001-04-23  show *all* categories except SEQ and TPL
# eallaert  2004-10-22  replaced obsolete Iwidget config-option (-items)
#                       general code clean-up
#

#************************************************************************
#   NAME
#   BobMobsBase - Multiple Observation Software base class
#
#
#   SYNOPSIS
#   BobMobsBase - widget to create/edit mobs files.
#      
#   DESCRIPTION
#   BobMobsBase provides a table based widget with two scrollbars to 
#   edit/create mobs files.
#
#   mobs files edit/display SETUP keywords from .DET, .INS and
#   .TEL files. Therefore the first 3 columns of the widget are reserved
#   to type file names of the type mentioned above.
#
#   Keywords defined in the loaded setup files can be selectively 
#   displayed into table's columns, using the method configColumns. It 
#   presents a listbox where multiple selection is allowed. Table columns
#   are added removed as needed, table geometry is kept constant.
#
#   Finally it provides methods for loading and writing mobs files 
#   out of the contents of the widget, the generated file is in
#   PAF format.
#
#   The handling of SETUP keywords and files is done through seq_oslx.
#
#   PUBLIC VARIABLES
#   None. 
#
#   PUBLIC METHODS
#   addRow : add a line to the bottom of the table
#   configColumns: create a toplevel listbox to select attributes to display
#         in the table widget.
#
#   getTableData: generates and OBD string in PAF format.
#
#   load <obd>: load a table from and OBD string <obd>. It is assumed that <obd>
#         is indeed MOBS-compliant, i.e. that the 2nd till last template are
#         generic ones.
#   tableBrowse <row col> : validate value of cell <row col>
#
#   PUBLIC OPTIONS
#   -cols      number of columns
#   -rows      number of rows
#   -arrayvar  name of global array variable where values are stored by this
#              widget (i.e. the value of each cell is an element in the array)
#   -helpvariable (name: helpVariable, class: HelpVariable)
#       The name of the global Tcl-variable which is used to display help text
#       when the cursor is over the table.
#   -helptext (name: helpText, class: Help)
#       The short help text for the scrolled listbox
#
#   FILES
#   A test INS_ROOT is provides under test directory
#  
#   ENVIRONMENT
#   INS_ROOT file structure is required. 
#
#   RETURN VALUES
#
#   CAUTIONS
#   A working INS_ROOT is needed, INS_USER must be defined.
#
#   EXAMPLES
#   Creates a BobMobsBase object, updates array gvar.
#
#   % BobMobsBase .m -arrayvar gvar 
#
#   Load file eg.mobs inside the widget, arrayVar uodated 
#   accordingly. 
#
#   % .m load eg.mobs
#
#   Show all keywords readed and allow te user to select 
#   the ones to be displayed/edited.
#
#   %.m configColumns
#
#   SEE ALSO
#
#
#------------------------------------------------------------------------
#

global auto_path INTROOT seq_ccsType
if {![string equal $INTROOT undefined]} {
    set auto_path "$INTROOT/lib $auto_path"
}
# package requirements
package require Tktable
package require SeqOslx
# oslx setup 
catch {seq_oslxCmd DICTION -load DicId ALL}
catch {seq_oslxCmd GEN -loadSetDicAll ALL}

#
# Usual options. These are the configuration options that shall be kept
# in the composite list of configuration options if this widget is made
# part of a megawidget - to the extent that these options are not yet
# part of that composite list due to other widgets that constitute
# the megawidget.
# This set of config options is normally added to the composite list of 
# the megawidget either via the "inherit" command or via the "itk_component"
# method, as "itk_component add <widget> <creationScript>" is equivalent to
# "itk_component add <widget> <creationScript> usual".
#
usual BobMobsBase {
    keep -cols -rows -helpvariable -helptext -arrayvar
}

#
# Use option database to override default resources of base classes.
#
option add *BobMobsBase.relief raised widgetDefault
option add *BobMobsBase.borderWidth 2 widgetDefault
option add *BobMobsBase.rows 10 widgetDefault
option add *BobMobsBase.cols 10 widgetDefault

itcl::class BobMobsBase {
    inherit itk::Widget

    constructor {args} { }

    public method addRow {} {}
    public method configColumns {} {}
    public method getTableData {} {}
    public method load {name} {}
    public method tableBrowse  {idx}     {}

    private method getTplParams {args} {}
    private method setColumns {} {}
    private method setHelpBind {args}
    private method setLineNrs   {}     {}
    private method updateRow    {row} {}
    private method getKeywordList {} {}
    private method writePafHeader {} {}

    private variable userColNames {};	# only for user-defined columns
    private variable rdFormat {%-30s "%s"};	# to format raw data
    private variable keyValues;	# array(rowNr,keyName) - value

    # The order of next list must correspond to *_col variables below !!
    private common fixedKeyNames {SEQ.NEXPO SEQ.SETUP.TIMEOUT}
    ::foreach cat {SEQ INS DET TEL} {
	::lappend fixedKeyNames ${cat}.SETUP.FILE
    }
    private common loop_col 1;		# col number for number of repetitions
    private common timeout_col 2;	# col number for timeout value
    private common refSetup_col 3;	# col number for ref detup file
    private common ins_col 4;		# col number for .ins files
    private common det_col 5;		# col number for .det files
    private common tel_col 6;		# col number for .tel files
    private common firstUserCol 7;	# first user-defined column
    private common minRows 10;	# show always at least this many rows

    itk_option define -helpvariable helpVariable HelpVariable {} {setHelpBind}
    itk_option define -helptext helpText Help {} {setHelpBind}
    itk_option define -arrayvar arrayVar Variable {} {}
}
# ------------------------------------------------------------------
#                        CONSTRUCTOR
# ------------------------------------------------------------------
itcl::body BobMobsBase::constructor {args} {
    global env

    if {[catch {set env(INS_ROOT)}]} {
	error "INS_ROOT environment variable must be set"
    }
    if {[catch {set env(INS_USER)}]} {
	error "INS_USER environment variable must be set"
    }
    
    # Revive the following options of hull:
    itk_option add hull.width hull.height hull.borderwidth hull.relief 

    itk_component add table {
	table $itk_interior.t \
	    -browsecmd [code $this tableBrowse %s] \
	    -yscrollcommand [code $itk_interior.sy set] \
	    -xscrollcommand [code $itk_interior.sx set]
    } {
	# as TkTable is not an Iwidget, we have to list here the
	# config options we want to keep, or declare that for the
	# Table class somewhere else and use the "usual" command.
	# For list of these config-options of "table" widget, see TkTable
	# Just put all of them into the composite list of config options
	keep -anchor -autoclear -background -bordercursor -borderwidth \
	    -browsecommand -cache -colorigin -cols -colseparator \
	    -colstretchmode -coltagcommand -colwidth -command -cursor \
	    -drawmode -exportselection -flashmode -flashtime -font -foreground \
	    -height -highlightbackground -highlightcolor -highlightthickness \
	    -insertbackground -insertborderwidth -insertofftime -insertontime\
	    -insertwidth -invertselected -ipadx -ipady -justify -maxheight \
	    -maxwidth -multiline -padx -pady -relief -resizeborders -rowheight \
	    -roworigin -rows -rowseparator -rowstretchmode -rowtagcommand \
	    -selectioncommand -selectmode -selecttitles -selecttype \
	    -sparsearray -state -takefocus -titlecols -titlerows \
	    -usecommand -validate -validatecommand -width \
	    -wrap -xscrollcommand -yscrollcommand
	# The next line is actually not necessary, as there is no
	# "usual" cmd defined for Table widgets. But just to make it
	# very clear that we take care of this -variable config-option
	# in our own way (see -arrayvar)
	ignore -variable
	#rename -width -colwidth colwidth ColWidth
    }

    itk_component add sy {
	scrollbar $itk_interior.sy \
	    -command [code $itk_component(table) yview]
    }
    itk_component add sx {
	scrollbar $itk_interior.sx -orient horizontal\
	    -command [code $itk_component(table) xview] 
    }

    grid $itk_component(table) $itk_component(sy) -sticky news -padx 2 -pady 2
    grid $itk_component(sx) -sticky ew -padx 2 -pady 2
    grid columnconfig . 0 -weight 1
    grid rowconfig . 1 -weight 1

    eval itk_initialize $args		
    
    switch -- $itk_option(-arrayvar) {} {
	configure -arrayvar ${this}.arrayVar; # in case option was not given
    }

    # Do this after creation of table (due to bug in TkTable?)
    $itk_component(table) configure -titlerows 1 -titlecols 1 \
	    -colwidth 15 -rowstretch unset -colstretch last
    setLineNrs
}

# ------------------------------------------------------------------
#                        PUBLIC VAR -arrayvar
# ------------------------------------------------------------------
itcl::configbody BobMobsBase::arrayvar {
	set arrayVar $itk_option(-arrayvar)
	# convert into fully qualified name (incl all namespaces)
	set arrayVar "[namespace qualifiers $arrayVar]::[namespace tail $arrayVar]"
	$itk_component(table) configure -variable $arrayVar
	# set the titles of the fixed columns
	set ${arrayVar}(0,$timeout_col) "timeout"
	set ${arrayVar}(0,$loop_col) "loops"
	set ${arrayVar}(0,$refSetup_col) "REF Setup File"
	set ${arrayVar}(0,$ins_col) "INS Setup File"
	set ${arrayVar}(0,$det_col) "DET Setup File"
	set ${arrayVar}(0,$tel_col) "TEL Setup File"
}
# ------------------------------------------------------------------
#                        PUBLIC METHOD addRow
# ------------------------------------------------------------------
# add a line to the table
itcl::body BobMobsBase::addRow {} {
    configure -rows [incr itk_option(-rows)]
    setLineNrs
}
# ------------------------------------------------------------------
#                        PUBLIC METHOD configColumns
# ------------------------------------------------------------------
# create listbox to select columns (i.e. keywords) to display in table widget
itcl::body BobMobsBase::configColumns {} {
    set keywords [getKeywordList]
    namespace eval ::bob {variable selectionDone}
    set keepGrab [grab current]
    panImportPanel bobMobsColConfig
    grab set .bobMobsColConfig

    ##$::gvar(bobSlb) configure -items $keywords
    eval $::gvar(bobSlb) insert end $keywords
    foreach name $userColNames {
	set idx [lsearch -exact $keywords $name]
	if {$idx >= 0} {
	    $::gvar(bobSlb) selection set $idx
	}
    }

    vwait ::bob::selectionDone

    set selectionDone [set ::bob::selectionDone]
    if {$selectionDone == 1} {
	set userColNames [$::gvar(bobSlb) getcurselection]
	setColumns
    }
    catch {grab set $keepGrab}
    panClosePanel no .bobMobsColConfig
}

# ------------------------------------------------------------------
#                        PUBLIC METHOD getTableData
# ------------------------------------------------------------------
# Get the data in PAF-format for the entire table; this is actually done
# by calling getTplParams for every row.
itcl::body BobMobsBase::getTableData {} {
    for {set i 1; set j 0} {$i <= $itk_option(-rows)} {incr i} {
	set tpl [getTplParams $i]
	if {[string length $tpl] != 0} {
	    append table "# Generic Template [incr j]" \n \
		    "# Edited via BOB's table editor" \n \
		    [format $rdFormat TPL.ID GT] \n \
		    [format $rdFormat TPL.NAME "Generic Template"] \n \n \
		    ${tpl} \n
	}
    }
    return $table
}
# ------------------------------------------------------------------
#                        PUBLIC METHOD load
# ------------------------------------------------------------------
# Load a table with data from an OBD (which is presumably in the right format)
itcl::body BobMobsBase::load {obd} {
    set row -1
    # get rid of continuation marks (i.e. concatenate)
    set pattern "\\\\\n"
    regsub -all -- $pattern $obd " " obd
    # insert a space between quote and semi-colon
    regsub -all -- \"\; $obd "\" \;" obd

    set waitForTplId 1
    set userColNames {}
    foreach line [split $obd \n] {
	set pattern "\[\t \]*\[A-Z\]\[A-Z\]\[A-Z\]\[0-9\]*\\.\[^\n\]*"
	if {![regexp -- $pattern $line]} {continue}

	lassign $line key value
	set cat [lindex [split $key .] 0]
	if {[string equal [string index $value 0] \"]} {
	    set value [lindex $value 0]; # remove quotes
	}

	# check if new row, marked by TPL.ID; skip no-INS-DET-TEL-keywords
	switch -- $key TPL.ID {
	    if {[incr row] == 0} {
		# very first template could be an acquisition template; skip it
		switch -- $value GT {
		    set waitForTplId 0
		} default {
		    continue
		} 
	    } else {
		set waitForTplId 0
		continue
	    }
	} default {
	    if {$waitForTplId} continue
	}	    

	# postpone cell-update until all column positions are known
	# (cf setColumns below)
	# SEQ/TPL is for "internal" use, all other cats should be sent to OS
	switch -- $cat SEQ - TPL {
	} default {
	    if {[lsearch -exact $fixedKeyNames $key] < 0} {
		lappend userColNames $key
	    }
	    set keyValues($row,$key) $value
	}
    }

    set userColNames [lrmdups $userColNames]
    setColumns
    # nr of rows is one higher than last row nr; show a min nr of rows
    configure -rows [expr {([incr row] < $minRows) ? $minRows : $row}]
    setLineNrs
}
# ------------------------------------------------------------------
#                        PRIVATE METHOD getKeywordList
# ------------------------------------------------------------------
# update internal variable <keywords> which holds available column names
# I.e. it is the union of all the keywords in the different reference 
# setup-files.
itcl::body BobMobsBase::getKeywordList {} {
    set arrayVar $itk_option(-arrayvar)
    # convert into fully qualified name (incl all namespaces)
    set arrayVar "[namespace qualifiers $arrayVar]::[namespace tail $arrayVar]"
    for {set row 1} {$row < $itk_option(-rows)} {incr row} {
	if {[info exists ${arrayVar}($row,$refSetup_col)] && \
		![lempty [set ${arrayVar}($row,$refSetup_col)]]} {
	    if {[seq_catch {seq_oslxCmd OSETUP -load ${this}.$row.SEQ \
		    [set ${arrayVar}($row,$refSetup_col)]} temp] || \
		    [seq_catch {seq_oslxCmd OSETUP -getSetupBuf \
		    ${this}.$row.SEQ} temp]} {
		error $temp
	    }
	    seq_catch {seq_oslxCmd OSETUP -clear ${this}.$row.SEQ}    
	    set data [split $temp ,]
	    lvarpop data
	    append keys $data " "
	}
    }

    set keywords {}
    if {[info exists keys]} {
	foreach key $keys {
	    lappend keywords [lindex $key 0]
	}
    }
    return [lrmdups $keywords]
}
# ------------------------------------------------------------------
#                        PRIVATE METHOD getTplParams
# ------------------------------------------------------------------
# Return the parameters (in PAF-format) of one table-row <r>
itcl::body BobMobsBase::getTplParams {r} {
    set arrayVar $itk_option(-arrayvar)
    # convert into fully qualified name (incl all namespaces)
    set arrayVar "[namespace qualifiers $arrayVar]::[namespace tail $arrayVar]"

    set tpl {}
    
    # get the fixed columns; if none is set, proceed anyway
    set i 1
    foreach key $fixedKeyNames {
	if {![catch {set ${arrayVar}($r,$i)} value] && ![lempty $value]} {
	    append tpl [format $rdFormat $key $value]\n
	}
	incr i
    }
    ###switch -- $tpl {} return
    
    for {set i $firstUserCol} {$i < $itk_option(-cols)} {incr i} {
	set key [set ${arrayVar}(0,$i)]
	if {![catch {set ${arrayVar}($r,$i)} tData]} {
	    append tpl [format $rdFormat $key $tData]\n
	}     
    }
    return $tpl
}

# ------------------------------------------------------------------
#                        PRIVATE METHOD setColumns
# ------------------------------------------------------------------
#   setColumns: set the number and labels of configurable columns.
#         <userColNames> is the list of labels, starting from the 1st 
#         configurable column.
# Method to set the amount and labels of configurable columns; the
# cell-contents also gets updated.
itcl::body BobMobsBase::setColumns {} {
    set arrayVar $itk_option(-arrayvar)
    # convert into fully qualified name (incl all namespaces)
    set arrayVar "[namespace qualifiers $arrayVar]::[namespace tail $arrayVar]"
    
    # insert user columns, after the fixed ones
    configure -cols [expr {[llength $userColNames] + $firstUserCol}]

    # Set column titles
    set i $firstUserCol
    foreach label $userColNames {
	set ${arrayVar}(0,$i) $label
	$itk_component(table) width $i [expr {[string length $label] + 4}]
	incr i
    }
    
    # Update display
    for {set i 1} {$i < $itk_option(-rows)} {incr i} {
	updateRow $i
    }
}
# ------------------------------------------------------------------
#                        PRIVATE METHOD setHelpBind
# ------------------------------------------------------------------
itcl::body BobMobsBase::setHelpBind {} {
    if {[string length $itk_option(-helpvariable)] == 0} return
    bind $itk_component(table) <Enter> \
	    "set $itk_option(-helpvariable) [list $itk_option(-helptext)]"
    bind $itk_component(table) <Leave> \
	    "set $itk_option(-helpvariable) {}"
}
# ------------------------------------------------------------------
#                        PRIVATE METHOD setLineNrs
# ------------------------------------------------------------------
# Set the line numbers for every row as a label in cell 0.
# The column width for these line numbers depends on the maximum number.
itcl::body BobMobsBase::setLineNrs {} {
    set arrayVar $itk_option(-arrayvar)
    # convert into fully qualified name (incl all namespaces)
    set arrayVar "[namespace qualifiers $arrayVar]::[namespace tail $arrayVar]"

    for {set i 1} {$i<$itk_option(-rows)} {incr i} {
	set ${arrayVar}($i,0) $i
    }
    $itk_component(table) width 0 [expr {[string length [incr i -1]] +2}]
}
# ------------------------------------------------------------------
#                        PRIVATE METHOD tableBrowse
# ------------------------------------------------------------------
# The next method gets executed anytime the active cell changes , i.e. when
# mouse button 1 is clicked on a different cell (see also -browsecommand 
# option of table in constructor). The argument is the index of the last
# active cell - which could have been modified.
itcl::body BobMobsBase::tableBrowse {idx} {
    set arrayVar $itk_option(-arrayvar)
    # convert into fully qualified name (incl all namespaces)
    set arrayVar "[namespace qualifiers $arrayVar]::[namespace tail $arrayVar]"
    if {[scan $idx %d,%d row col] != 2 || $row == 0 || $col == 0} return
    set val [$itk_component(table) get $idx]

    if {$col < $firstUserCol} {
	set key [lindex $fixedKeyNames [expr {$col -1}]]
	set keyValues($row,$key) $val
    } else {
	set key [set ${arrayVar}(0,$col)];	# this cell MUST exist
	switch -- $val {} {
	    catch {unset ${arrayVar}($row,$col)}
	    catch {unset keyValues($row,$key)}
	} default {
	    set keyValues($row,$key) $val
	}
    }
}

# ------------------------------------------------------------------
#                        PRIVATE METHOD updateRow
# ------------------------------------------------------------------
# Update the values on row <r>, for all keywords incl the setup files
itcl::body BobMobsBase::updateRow {r} {
    set arrayVar $itk_option(-arrayvar)
    # convert into fully qualified name (incl all namespaces)
    set arrayVar "[namespace qualifiers $arrayVar]::[namespace tail $arrayVar]"

    set i 1;	# number of column we're dealing with
    foreach key "$fixedKeyNames $userColNames" {
	if {[catch {set ${arrayVar}($r,$i) $keyValues($r,$key)}]} {
	    catch {unset ${arrayVar}($r,$i)}
	}
	incr i
    }
}

#
# ___oOo___




