#*******************************************************************************
# E.S.O. - VLT project
#
# "@(#) $Id: bobSave.tcl,v 1.145 2005/10/27 09:07:07 vltsccm Exp $"
#
# who       when        what
# --------  ----------  ----------------------------------------------
# eallaert  1997-02-18  created
# eallaert  2004-01-07  created bobSaveAs for programmatic access.
#

#************************************************************************
#   NAME
#   bobSave, bobSaveAs - save the observation block description of a 
#         selected OB
# 
#   SYNOPSIS
#   bobSave
#   bobSaveAs ?<file>?
# 
#   DESCRIPTION
#   These procs will retrieve the OBD of the selected OB, and prepend a proper
#   header to it before writing it to an .obd file. There are two forms of
#   basically the same function:
#   
#   bobSave
#      saves the current OBD into a file. If the OBD was originally loaded
#      from a local file, it will be saved into that same file; otherwise
#      a file-selection window will pop up to ask for a filename.
#
#   bobSaveAs ?<file>?
#      saves the current OBD into <file>. Remark that if <file> is given but
#      does not contain an absolute path, the saving will be done with 
#      respect to the current working directory, which can be manipulated 
#      by any template. 
#      If <file> is not specified, bobSaveAs will pop up a file-selection 
#      window to select a filename.
#
#   PUBLIC VARIABLES
#   
#   FILES
#
#   ENVIRONMENT
#
#   RETURN VALUES
#
#   CAUTIONS
#
#   EXAMPLES
#
#   SEE ALSO
#
#
#------------------------------------------------------------------------
#

#=============================================================================

# _bobSave: internal proc, used by all externally accessible ones.
#   <forced_popup> : optional flag, indicating whether a file selection
#        box should pop up, requesting confirmation of user. Default: 0
#        * -1: use bob::internal::info(obdFile) as name of the file to save
#        *  0: use bob::internal::info(obdFile) only if source of OBD is local file
#        *  1: bring up a file selection box up in any case.
#
proc ::bob::internal::_bobSave { {forced_popup 0} } {

    switch -- $::ROOT {} {
	bobDialog "You have to load an OB first before you can save it." \
		application 10000
	return
    }

    # Test if anything is selected; if only one OB, save that one by default
    if {[BobBaseObject::markedIsOB]} {
	set OB [BobBaseObject::marked]
    } else {
	if {[catch {$::ROOT contents} OB]} {
	    set OB {}
	}
	if {[llength $OB] > 1} {
	    set OB [$::ROOT activeChild];	# select the one which is running
	}
	switch -- $OB {} {
	    bobDialog "You have to select an OB first by clicking on it\n\
		    before it can be saved." application 10000
	    return
	}
    }

    if {[info exists ::gvar(bobObdFilePath)]} {
	set example $::gvar(bobObdFilePath)/.obd
    } else {
	set example .obd
    }

    if { [string equal [$OB getObdSource] file] && $forced_popup == 0} {
    	set file $::bob::internal::info(obdFile)
    	if { ! $::bob::internal::config(bobUserLevel) } {
	    if {[bobDialog "saving current OBD in $file" \
		    application 0 info $::bob::internal::config(bobUserLevel) 1] != 1} {
		return
	    }
	}
    } else {
	if {$forced_popup == -1} {
	    set file $::bob::internal::info(obdFile)
	} else {
	    set file [bobFileSel "Save Observation Block Description" $example]
	    if {[string length $file] == 0} {
		return
	    } else {
		# keep the selected filename, for future save-actions
		set ::bob::internal::info(obdFile) $file
	    }
	}
    	# keep also filepath for future calls
    	set ::gvar(bobObdFilePath) [file dirname $file]

    	set ext [file extension $file]
    	if {![string equal $ext .obd]} {
	    set file ${file}.obd
    	}
    }
    #ISO time string: e.g. 1997-11-05T23:34:45.67
    set daytim [clock format [clock seconds] -format %Y-%m-%dT%H:%M:%S -gmt 1]
    write_file $file \
	    {#} \
	    {# OBD-file written by BOB}\
	    {# It includes the description of one observation block}\
	    {#}\
	    {}\
	    {# Standard parameter file header}\
	    {}\
	    {PAF.HDR.START   ;               # Marks start of header}\
	    {PAF.TYPE                "OB Description";  # Type of parfile}\
	    {PAF.ID                  "";     # Unused}\
	    {PAF.NAME                "";     # Unused}\
	    {PAF.DESC                "";     # Unused}\
	    {PAF.CRTE.NAME        "BOB";     # Broker for OBs}\
	    "PAF.CRTE.DAYTIM     \"$daytim\"; # Date+time of creation"\
	    {PAF.LCHG.NAME           "";     # Unused}\
	    "PAF.LCHG.DAYTIM     \"$daytim\"; # Date+time of last change"\
	    {PAF.CHCK.NAME           "";     # Unused}\
	    {PAF.CHCK.DAYTIM         "";     # Unused}\
	    {PAF.CHCK.CHECKSUM       "";     # Unused}\
	    {PAF.HDR.END             ;       # Marks end of header}\
	    {}\
	    {# Observation Block description follows}\
	    [$OB getRawData]
}

#=============================================================================

proc ::bob::bobSaveAs { {file {} } } {

    if {[string length $file] == 0} {
	::bob::internal::_bobSave 1
    } else {
	# keep some gvar variables, to restore them in case of 
	# file-access problems
	if {![info exists ::gvar(bobObdFilePath)]} {
	    set ::gvar(bobObdFilePath) {}
	    set ::bob::internal::info(obdFile) {}
	}
	set bobObdFilePath $::gvar(bobObdFilePath)
	set bobObdFileName $::bob::internal::info(obdFile)
	set dir  [file dirname $file]
	set name [file tail $file]
	set ::bob::internal::info(obdFile) [file join [seq_relToAbsPath $dir] $name]
	
	if {[catch {::bob::internal::_bobSave -1} msg]} {
	    set ::gvar(bobObdFilePath) $bobObdFilePath
	    set ::bob::internal::info(obdFile) $bobObdFileName
	    error $msg
	}
    }
}

proc ::bob::bobSave {} {
    ::bob::internal::_bobSave 0
}




