#*******************************************************************************
# E.S.O. - VLT project
#
# "@(#) $Id: bobVarious.tcl,v 1.145 2005/10/27 09:07:08 vltsccm Exp $"
#
# who       when        what
# --------  ----------  ----------------------------------------------
# eallaert  30/11/96    created
# eallaert  2000-07-03  added setting of log-file to bobPropagateVars
# eallaert  2002-10-10  added update of canvas label if config changes
# eallaert  2003-04-28  delay effect of bobProcNameOH change (SPR 20030145)
#

#************************************************************************
#   NAME
#   _bobAppendAutoPath, traceBobVars, _bobLoadPlugin - BOB Various Library
# 
#   SYNOPSIS
#   _bobAppendAutoPath <libs>
#   traceBobVars
#   _bobLoadPlugin <plugIn>
#
#   DESCRIPTION
#   _bobAppendAutoPath will append the proper path of each library in <libs> 
#   to the auto_path global variable. <libs> is a list of library names,
#   excluding the "lib" prefix and the ".tcl" suffix.
#
#   traceBobVars will undertake all necessary actions when particular global 
#   configuration variables are changed (in particular the ones which directly
#   affect some settings which are otherwise not taken up during whatever loop).
# 
#   _bobLoadPlugin will load <plugIn>, first by attempting a "panImportPanel", 
#   and if this fails followed by a plain "source". <plugin> needs to be under
#   bin subdirectory of either ., .., INTROOT or VLTROOT (searched in that 
#   order). If the search fails, _bobLoadPlugin will pop up an error dialog box.
#   The return is always succesful, with an empty value.
#
#   These procedures are all defined in the namespace ::bob::internal, clearly
#   indicating for which use they are meant.
#
#   RETURN VALUES
#
#   FILES
#
#   ENVIRONMENT
#
#   CAUTIONS
#
#   EXAMPLES
#
#   SEE ALSO
#
#------------------------------------------------------------------------
#
#
# Make sure that the namespaces we will be dealing with in here do exist
namespace eval ::bob {}
namespace eval ::bob::internal {}

# Needed for backward compatibility - was used in .bobrc
proc bobPropagateVars {} {}

proc ::bob::internal::_bobAppendAutoPath {libs} {

    set failures {}
    foreach lib $libs {
	set path [bobFindVlt lib/lib${lib}.tcl]
	if {[string equal $path {}]} {
	    lappend failures lib${lib}.tcl
	} elseif {[lsearch -exact $::auto_path $path] == -1} {
	    lappend ::auto_path $path
	}
    }
    set f [llength $failures]
    if {$f == 1} {
	error "Tcl library \"$failures\" not found"
    } elseif {$f > 1} {
	error "Tcl libraries \"$failures\" not found"
    }
}

proc ::bob::internal::_bobLoadPlugin {plugin} {
    if {[catch {panImportPanel $plugin}]} {
	if {[catch {seq_findFile bin/$plugin} path]} {
	    bobDialog "Could not locate the plug-in \"$plugin\".\
		    \n\nIt must be under the bin/ subdirectory of either\
		    \n\., .., \$INTROOT or \$VLTROOT (searched in that order).\
		    \n\Continuing without loading this plug-in." \
		    application 10000
	} else {
	    uplevel #0 source $path
	}
    }
    return
}

proc ::bob::internal::traceBobVars {array el op} {
    switch -- $el bobProcNameOH {
	_bobUifSetCanvasLabel OH
    } bobProcNameOS {
	BobLogger::openLogFile $::env(VLTDATA)/tmp/bob_[lindex $::bob::internal::config(bobProcNameOS) 0].log
	_bobUifSetCanvasLabel OS
    } bobSoundServer {
	# Deal with the audio-server; audioPlay takes the env var AUDIO
	if {[catch {subst $::bob::internal::config(bobSoundServer)} audioServer]} {
	    bobDialog "Could not set the sound-server:\n$audioServer" application
	} else {
	    if {[string length [string trim $audioServer]] > 0} {
		set ::env(AUDIO) $audioServer
	    } elseif {[info exists ::bob::internal::origAUDIO]} {
		# set it back to the original setting of the AUDIO env var
		set ::env(AUDIO) $::bob::internal::origAUDIO
	    } else {
		# we've started up without an AUDIO env var
		unset -nocomplain ::env(AUDIO)
	    }
	}
    } tplLibList {
	set ::auto_path $::gvar(origAutoPath)
	if {[catch {::bob::internal::_bobAppendAutoPath $::bob::internal::config(tplLibList)} msg]} {
	    bobDialog "Could not set the auto_path according to the config-info:\n\
		    $msg" application
	}
    }
}
	    
# Remark that bobCfgCallBack needs 4 arguments additional arguments:
# - the name of the configuration variable
# - a flag indicating whether this is the initialization phase (1=yes)
# - the old configuration value (empty in case of initialization)
# - the current config value
proc ::bob::internal::traceConfigChange {array el op} {
    set init 0
    if {![string equal $array ::bob::internal::config] } {
	error "traceConfigChange called for array $array ?!?!"
    } elseif {[string equal $op init]} {
	# This is not a Tcl call-back, but a call during the init phase
	set init 1
    } elseif {![string equal $op write]} {
	error "traceConfigChange called for operation $op ?!?!"
    }
    
    if {[info exists ::bob::internal::prevConfig($el)] && 
	[string equal $::bob::internal::prevConfig($el) $::bob::internal::config($el)]} {
	return
    }

    if {[info exists ::bob::internal::prevConfig($el)]} { 
	set previous $::bob::internal::prevConfig($el)
    } else {
	set previous {}
    }
    # set prevConfig immediately, to be ready for subsequent calls
    set ::bob::internal::prevConfig($el) $::bob::internal::config($el)

    if {[string length $::bob::internal::config(bobCfgCallBack)] > 0} {
	set command [lindex $::bob::internal::config(bobCfgCallBack) 0]
	if {[string length [info command $command]] > 0 || [auto_load $command]} {
	    uplevel \#0 $::bob::internal::config(bobCfgCallBack) $el $init \
		[list $previous] [list $::bob::internal::config($el)]
	}
    }
}	

proc ::bob::internal::gvarAbused {array el op} {
    switch -- $op read {
	if {![info exists ::bob::internal::abusedGvars(read)] || \
		([info exists ::bob::internal::abusedGvars(read)] &&
		 [lsearch -exact $::bob::internal::abusedGvars(read) gvar($el)] < 0)} {
	    lappend ::bob::internal::abusedGvars(read) gvar($el)
	}
    } write {
	if {[string equal $el bobObdFileName]} {
	    # We assume that wherever this is done, there is also
	    # a cmd 'set ::bob::internal::info(obdFile) <value>'
	    # This is OK for bob; template should not write into this var anyway.
	} else {
	    ::bob::$el $::gvar($el)
	}
	if {![info exists ::bob::internal::abusedGvars(write)] || \
		([info exists ::bob::internal::abusedGvars(write)] &&
		 [lsearch -exact $::bob::internal::abusedGvars(write) gvar($el)] < 0)} {
	    lappend ::bob::internal::abusedGvars(write) gvar($el)
	}
    } unset {
	# This is really kinky ...
	if {[info exists ::bob::internal::abusedGvars(unset)] || \
		([info exists ::bob::internal::abusedGvars(unset)] &&
		 [lsearch -exact $::bob::internal::abusedGvars(unset) gvar($el)] < 0)} {
	    lappend ::bob::internal::abusedGvars(unset) gvar($el)
	}
	error "caught unset command on gvar($el)"
    }
}
