#*******************************************************************************
# E.S.O. - VLT project
#
# "@(#) $Id: irtdDartUtils.tcl,v 1.22 2005/04/18 14:33:36 vltsccm Exp $"
#
# who       when      what
# --------  --------  ----------------------------------------------
# pbiereic  15/02/00  created
# jstegmei  26/08/03  added configurable port number and nx, ny parameter
# pbiereic  28/10/03  removed deprecated -irtdPort option. The server port
#                     number of irtdDart is communicated to irtd via the
#                     Rtd remote interface. irtdDart assigns the port
#                     number dynamically.
# mkiekebu 2004-10-25 removed check for irtdPort during the first second.
#                     It only check if it can connect to the irtdDart
#                     otherwise it terminates.
# pbiereic  27/10/04  terminate irtdDart if connection failed
# pbiereic  18/04/05  irtdDartConnect() passes the pid and remote port to irtdDart
#                     to make sure that irtdDart connects to it's parent RTD

#************************************************************************
#   NAME
#   irtdDartUtils.tcl - internal utility procs for the IRACE data
#                        reception task ("Dart") interface
# 
#   SYNOPSIS
# 
#   DESCRIPTION
#   The following proc's are contained in this file
#
#   - irtdDartConnect: start and connect to the data reception task irtdDart
#   - irtdSendDart:    send a message to the data server irtdDart
#   - irtdDart:        irtdDart uses the remote interface and calls this routine
#
#   FILES
#
#   ENVIRONMENT
#
#   RETURN VALUES
#
#   CAUTIONS
#   none
#
#   EXAMPLES
#
#   SEE ALSO
#   RTD documentation
#
#   BUGS     
#
#------------------------------------------------------------------------

# send a message to the data server irtdDart

proc irtdSendDart {msg} {
    global ::irtd

    if {! [info exists irtd(dartSock)]} {
	puts "irtdSendDart: $msg"
	return
    }
    puts $irtd(dartSock) $msg
    flush $irtd(dartSock)
    return 0
}

# irtdDart uses the remote interface and calls this routine

proc irtdDart {args} {
    global ::irtd

    # rtdbFeedback $args
    if {"$args" == ""} { return }
    set dcsctl $irtd(Irace)

    set cmd [lindex $args 0]
    set parms [lreplace $args 0 0]
    
    switch -exact $cmd {
        TYPES {
            $dcsctl set_data $parms
        }
        ON {
            $dcsctl onOff 1
        }
        OFF {
            $dcsctl onOff 0
        }
        TRANS {
            $dcsctl trans $parms
        }
        RATE {
            $dcsctl rate $parms
        }
        NEWFR {
            $dcsctl newfr $parms
        }
        TYPENAMES {
            set irtd(frameTypes) $parms
        }
        CONNECTING {
            $dcsctl connecting
        }
        CONNECTED {
            $dcsctl connected
        }
    }
}

# configure menu buttons for data menu

proc irtdData_menu {frames menu w meth} {
    global ::irtd
    global ::$menu

    set data [$menu.mb cget -text]
    if {"[$menu.mb.m index 0]" == "0"} {
        $menu.mb config -text ""
        catch {$menu.mb.m delete 0 10}
    }
    foreach el $frames {
        $menu add -label $el -command [code $w $meth $el]
    }
    set found 0
    set i 0
    foreach el $frames {
        if {"$data" == "$el"} {
            $menu.mb.m invoke $i
            set found 1
            break
        }
        incr i
    }
    if {! $found} {
        $menu.mb.m invoke 0
    }
}

# return frame type

proc irtdGetFrameType {frame} {
    global ::irtd
    if {! [info exists irtd(frameTypes)]} {
        set irtd(frameTypes) "DIT INT"
    }    
    set idx [lsearch $irtd(frameTypes) $frame]
    if {$idx < 0} {
        return 2
    }
    return [expr {round(pow(2, $idx))}]
}

# receive server port number from data reception task irtdDart

proc irtdSrvPort { irtdPort args } {
    global ::irtd
    set irtd(irtdPort) $irtdPort
}

# start and connect to the data reception task irtdDart

proc irtdDartConnect { {waitDart 5} } {
    global ::irtd ::rtdb ::env ::argv ::argc

    set nx_max 0
    set ny_max 0

    # pass pid and remote port to irtdDart
    set cmdArgs "-rtdPid [pid] -rtdPort [[$rtdb(RtdImage) get_image] remote]"

    set idx [lsearch -exact $argv "-nx"]
    if { $idx != -1 } {
	set nx_max [lindex $argv [expr {$idx +1}]]
    }

    set idx [lsearch -exact $argv "-ny"]
    if { $idx != -1 } {
	set ny_max [lindex $argv [expr {$idx +1}]]
    }

    set idx [lsearch -exact $argv "-camera"]
    if { $idx != -1 } {
	set cameraName [lindex $argv [expr {$idx +1}]]
	set cmdArgs "$cmdArgs -camera $cameraName"
    }

    foreach el "SDMA_HOST SDMA_DATA HOST" {
	rtdbCheckEnv $el
    }

    # check if port is already in use
    if { $nx_max > 0 } {
	set cmdArgs "$cmdArgs -nx $nx_max"
    }
    if { $ny_max > 0 } {
	set cmdArgs "$cmdArgs -ny $ny_max"
    }

    # start data reception task
    if {$rtdb(VERBOSE) == 1} {
	append cmdArgs " -v"
    }
    rtdbFeedback "Starting: irtdDart $cmdArgs"
    set err [catch {eval exec "irtdDart $cmdArgs &"} irtdDartPid]
    if { $err } {
        error_dialog "Unable to start process irtdDart: $irtdDartPid"
        rtdbTerminate
    }

    # connect to data reception task. Wait first until we got the port number
    # from irtdDart via the Rtd remote interface. irtdDart will call the
    # proc irtdSrvPort which sets the port number into the global array irtd.

    set nretries [expr {$waitDart * 10}]  ; # retry count for timeout
    set failed 1
    for {set i 0} {$i < $nretries} {incr i} {
        if { [info exists irtd(irtdPort)] } {
	    if { ! [catch {server_connect $env(HOST) $irtd(irtdPort)} irtd(dartSock)] } {
		set failed 0
		break
	    }
        }
	after 100
	update
    }

    if { $failed } {
	error_dialog "Unable to connect to process irtdDart"
	catch {kill $irtdDartPid}
	rtdbTerminate
    }
}
