#*******************************************************************************
# E.S.O. - VLT project
#
# "@(#) $Id: irtdRtd.tcl,v 1.22 2005/04/18 14:33:33 vltsccm Exp $"
#
# who       when      what
# --------  --------  ----------------------------------------------
# pbiereic  15/02/00  created
# pbiereic  20/10/00  adapted for NOV2000
# pbiereic  19/07/01  request full frame if hw-window is on
# pbiereic  11/11/03  added methods: connecting {} and connected {}
# pbiereic  16/04/05  added command option -irtdRoot to specify the IRACE dbRoot
# pbiereic  18/04/05  added command option -irtdWait (VLTSW20050155)
#
#************************************************************************
#   NAME
#   irtdRtd.tcl - tcl plugin for the IRACE-DCS
# 
#   SYNOPSIS
#   irtdRtd <path> ?options?
#
#   STANDARD OPTIONS
#        -image image
#                     Rtd image widget = [$rtdb(Rtd) get_image]
#
#        -shorthelpwin
#                     Short help window = $rtdb(Rtd) 
#
#   PUBLIC OPTIONS
#   itk_option define -irtdRoot irtdRoot IrtdRoot {<alias>iracq}
#   - database root point for irtd
#
#   itk_option define -waitDart waitDart WaitDart {5}
#   - max. wait time for process irtdDart to respond (sec)
#
#   DESCRIPTION
#   irtdRtd consists of methods called by the control buttons and menu's
#   defined in the IRACE-DCS config files for RTDB. Some methods are called
#   via the RTD remote interface by the data reception task to communicate
#   status events and other display data like the frames which are available
#   for display in the data server, start/end of data transmission etc.
#
#   irtdRtd generally handles database events from the IRACE-DCS e.g. windowed
#   readout.
#
#   PARENT CLASS
#        util::FrameWidget
#
#   PUBLIC PROCEDURES
#        irtdSelDet {number {id ""}}
#             Select detector (e.g 0 for SW 1 for LW arm). Used for fixed pattern.
#
#        irtdCtrl {cmd}
#             Evaluates a command which calls a method of this class. Used by
#             the control widgets.
#
#        irtdRtdMenu {cmd {flgRtd 0}}
#             Procedure which is called by the menu items. When flgRtd is set to 1
#             then we call the Rtd main widget
#
#   FILES
#
#   ENVIRONMENT
#   CCS
#
#   RETURN VALUES
#
#   CAUTIONS
#
#   EXAMPLES
#
#   SEE ALSO
#   RTD documentation
#
#   BUGS     
#
#------------------------------------------------------------------------

itk::usual irtdRtd {}

class ::irtdRtd {
    inherit util::FrameWidget

    constructor {args} {
	global ::rtdb
	eval itk_initialize $args

	# keep local copy of Rtd widget names
	set RtdWdg_   $rtdb(Rtd)
	set panelWdg_ [$RtdWdg_ get_panel]
	set zoomWdg_  [$RtdWdg_ get_zoom]
    }

    # method init is called by the TopLevelWidget after all options have been evaluated

    method init {} {
	global ::rtdb ::irtd
	# name of control widgets
	foreach el "start fpon stfp data fast" {
	    if { ! [winfo exists $panelWdg_.$el] } {
		rtdbExit_dialog "Control widget '$el' was not created. Check config files."
	    }
	    set $el\Wdg_ $panelWdg_.$el
	}
	set ccsInit_ $rtdb(CCSINIT) ; # CCS initialized flag

	configState
	
	# trace status of a global variable which is set by the rtdimage
	# subcommand 'biasimage'
	global ::$image_
        trace variable ${image_}(BIAS) w [code $w_ updateFpsts]

	onOff 0
	trans 0
	# we need to be informed about new image events...
	$RtdWdg_ add_notify       [code $this notify_cmd]
	$RtdWdg_ add_preCmd       [code $this preCommand 0]
	$RtdWdg_ add_postCmd      [code $this postCommand 0]
	$RtdWdg_ add_rapidPreCmd  [code $this preCommand 1]
	$RtdWdg_ add_rapidPostCmd [code $this postCommand 1]

        irtdDartConnect $itk_option(-waitDart) ; # connect to data reception task

	attachToDbEvents
	windowedRead
    }

    # run the rate display window

    method rate_display {} {
	set rateWdg_ $RtdWdg_.rate
	if {! [winfo exists $rateWdg_]} {
	    irtdRateDisp $rateWdg_
	}
	rtdbPopup $rateWdg_
    }

    # run the count-down display window

    method count_display {} {
	set countWdg_ $RtdWdg_.count
	if {! [winfo exists $countWdg_]} {
	    irtdCountDisp $countWdg_
	}
	rtdbPopup $countWdg_
    }

    # toggle rapid frame display

    method fast {} {
	global ::$fastWdg_
	if {! [set $fastWdg_]} {
	    catch {$itk_component(rapid) remove_RF}
	    return
	}
	if {! [rtdbIsVideoImage]} {
	    update
	    after 100
	    set $fastWdg_ 0
	    return
	}
	set rapidWdg_ $RtdWdg_.rapid
	if {! [winfo exists $rapidWdg_]} {
	    itk_component add rapid {
		irtdRapid $rapidWdg_ \
			-target_image [$RtdWdg_ get_image] \
			-shorthelpwin $itk_option(-shorthelpwin) \
			-zoomwin      [$RtdWdg_ get_zoom] \
			-command "$this fastOff" \
			-center 1 \
			-with_control 0
	    }
	} else {
	    rtdbPopup $itk_component(rapid)
	}
	after idle [code $itk_component(rapid) set_data $frames_]
	$itk_component(rapid) raise_dwin
    }


    # Switch off the "Rapid frame..." checkbutton

    method fastOff {} {
	global ::$fastWdg_
	set $fastWdg_ 0
    }

    # Start/Stop main frame

    method start {} {
	global ::irtd
	set irtd(cmain) 0

	if {[set_start]} {
	    irtdSendDart "MAIN 0 1 $mainData_"
	    $startWdg_ config -text Stop
	    waiting_msg 1
	} else {
	    irtdSendDart "MAIN 0 0 $mainData_"
	    $startWdg_ config -text Start -fg black
	    waiting_msg 0
	    if {[set_start -1] != 0} {
		set_start 0
	    }
	    $image_ update
	}
    }

    # request frame type

    method start_data {frame} {
	waiting_msg 1
	set mainData_ [irtdGetFrameType "$frame"]
	irtdSendDart "MAIN 0 [set_start -1] $mainData_"
    }

    # method to set/get the start button

    method set_start {{ss -1}} {
	global ::$startWdg_
	if {$ss == -1} {
	    return [set $startWdg_]
	}
	set $startWdg_ $ss
    }

    # indicate "waiting for data" on start button

    method waiting_msg {waiting} {
	global ::$startWdg_
	if {[set_start] && $waiting} {
	    $startWdg_ config -fg blue
	} else {
	    $startWdg_ config -fg black
	}
    }

    # Store fixed pattern

    method stfp {} {
	catch {$image_ biasimage copy $det_}
    }

    # Toggle fixed pattern subtraction

    method fpon {} {
	global ::$fponWdg_

	if {[set $fponWdg_]} {
	    catch {$image_ biasimage on}
	} else {
	    $image_ biasimage off
	}
    }

    # display fixed pattern frame

    method displayFP {} {
	catch {
	    $image_ biasimage select $det_
	    $image_ biasimage display
	}
    }

    # negate real-time image (needed for pick object)

    method rtiNegate {} {
	set negate_ [expr {! $negate_}]
	irtdSendDart "NEGATE $negate_"
    }

    # method to attach to DCS, OS and ICS database events.

    method attachToDbEvents {} {
	foreach el "STARTX STARTY NX NY hwWindow" {
	    if {[catch {dbAttach ":exposure.$el" W windowedRead}]} { return }
	}
    }

    method dbAttach { attr filter eventCb {point ""}} {
	if {! $ccsInit_} { return }
	seq_evtAttach "[cget -irtdRoot]$attr" $filter "$this $eventCb $point %D"
    }

    method dbRead { attr } {
	if {! $ccsInit_} { return 0 }
	return [seq_dbReadSymbolic "[cget -irtdRoot]$attr"]
    }

    # method to set the data menu items

    method set_data {{frames 2}} {
	set frames_ $frames
	irtdData_menu $frames $dataWdg_ $this start_data
	catch {$itk_component(rapid) set_data $frames_}
    }

    # handle database events

    method handleDbEvents {} {
	windowedRead
    }

    # callback for windowed readout

    method windowedRead {args} {
	set hwsw [dbRead ":exposure.hwWindow"]
	lassign [detWin] x0 y0 w h
	
	# Only send window parameters if there are differences
	set win_ [list $x0 $y0 $w $h $hwsw]
	if { $hwsw } {
	    # Request all data when hardware window is set
	    set win_ [list 0 0 0 0 1]
	}
	loop i 0 5 {
	    if {[lindex $win_ $i] != [lindex $winParms_ $i]} {
		irtdSendDart "WIN_MAIN [rtdbIsVideoImage] $x0 $y0 $w $h $hwsw"
		set winParms_ $win_
		return
	    }
	}
    }

    # return coordinates of detector window

    method detWin {} {
	set x0 [dbRead ":exposure.STARTX"]
	set y0 [dbRead ":exposure.STARTY"]
	set w  [dbRead ":exposure.NX"]
	set h  [dbRead ":exposure.NY"]

	set nx [dbRead ":detector.NX"]
	set ny [dbRead ":detector.NY"]
	if {[rtdbIsVideoImage]} {
	    if {[info exists x0_]} {
		lassign [rtdbDistImageToCanvas [expr {$x0_ - $x0}] [expr {$y0_ - $y0}]] dx dy
		after idle "$this reDraw $dx $dy"
	    }
	}
	set x0_ $x0
	set y0_ $y0
	waiting_msg 1
	return "$x0 $y0 $w $h"
    }

    # redraw graphics when image window was changed

    method reDraw {dx dy} {
	foreach el "objects grip" {
	    lappend tags [$canvas_ find withtag $el]
	}
	foreach el [join $tags] {
	    $canvas_ move $el $dx $dy
	}
    }

    method preCommand {rapid} {
	if {$rapid} {
	    catch {$itk_component(rapid) rapidPreCommand}
	    return
	}
	set clk_ [clock clicks]
    }


    # This method is called whenever a new image has been received from the camera
    # Update the widgets that need to display new values

    method postCommand {rapid} {
	global ::irtd
	if {$rapid} {
	    catch {$itk_component(rapid) rapidPostCommand}
	    if {[info exists irtd(crf)]} {incr irtd(crf)}
	    catch {$RtdWdg_.count reset crf}
	    return
	}
	if {[info exists irtd(cmain)]} {incr irtd(cmain)}
	lassign [image_size $canvas_] nx ny
	catch {
	    $rateWdg_ drate [format {%.1f} \
				 [expr {double($nx * $ny / 1024.) / (double([clock clicks] - \
										$clk_) * 1.0E-06)}]]
	}
	# data arrived
	waiting_msg 0
	catch {$RtdWdg_.count reset cmain}
    }

    # get size of target image

    method image_size { cv } {
	lassign [$cv bbox all] null null dwidth dheight
	set cwidth [set width [winfo width $cv]]
	set cheight [set height [winfo height $cv]]
	
	if {$dwidth < $cwidth} {
	    set width $dwidth
	}
	if {$dheight < $cheight} {
	    set height $dheight
	}
	return "$width $height"
    }

    # configure button states.

    method configState {} {
	foreach el "start data fast" {
	    if {$dartStopped_} {
		[set $el\Wdg_] config -state disabled
	    } else {
		[set $el\Wdg_] config -state normal
	    }
	}
    }

    method newfr { frameId } {
	if { $frameId == 2 } {
	    set timeLastFrame2_ [clock seconds]
	} elseif { $frameId == 4 } {
	    set timeLastFrame4_ [clock seconds]
	}
    }

    # this method is called when the data server goes to on/off

    method onOff {on} {
	$RtdWdg_ camera_ready $on
	set dartStopped_ [expr {! $on}]
	configState
    }

    # this method is called when the data trasmission starts

    method trans {on} {
	$RtdWdg_ set_flash $on
    }

    # this method is called when trying to connect to the data server

    method connecting {} {
	catch {$canvas_ delete connect}
	lassign [rtdbScreenToCanvas 250 20 0 0] cx0 cy0 cx1 cy1
	set fontH "-*-courier-Bold-r-*-*-*-240-*-*-*-*-iso8859-*"
	catch {$canvas_ create text $cx0 $cy0 \
		   -text "Trying to connect to IRACE..." \
		   -fill blue -font $fontH -tags connect}
    }

    # this method is called when we are connected to the data server

    method connected {} {
	catch {$canvas_ delete connect}
    }

    method rate {parms} {
	lassign $parms rate size
	catch {$rateWdg_ rate $rate $size}
    }

    # method to define a centered window interactively.

    method defCenteredWin { } {
	if {[check_image]} {
	    rtdbReply "-1 No image is currently loaded."
	    return
	}
	set win $w_.defcwin
	if {! [popup $win]} {
	    itk_component add defcwin {
		utilReUseWidget util::rtdbDefCrect $win \
			-target_image $this \
			-title "Window parameters" \
			-shorthelpwin $itk_option(-shorthelpwin)
	    }
	}
	$itk_component(defcwin) raise_dwin
	return $itk_component(defcwin)
    }

    # notify_cmd is called after each new image

    method notify_cmd {args} {
	global ::$fastWdg_
	if {! [rtdbIsVideoImage] && [set $fastWdg_]} {
	    catch {$itk_component(rapid) remove_RF}
	}
	configState
    }

    # use fixed pattern for the current instrument arm

    method selDet {det id} {
	set det_ $det
	$image_ biasimage select $det_
	if {"$id" == ""} {
	    $stfpWdg_ config -text "Store fix. pattern"
	} else {
	    $stfpWdg_ config -text "Store fix. pattern ($id)"
	}
	return 0
    }

    # update status of the fpon widget

    method updateFpsts {{args}} {
	global ::$fponWdg_
	if {[$image_ biasimage status] == -1} {
            set $fponWdg_ 0
            $fponWdg_ config -state disabled
        } else {
            set $fponWdg_ [$image_ biasimage status]
            $fponWdg_ config -state normal
        }
	# re-display image
 	$image_ update
    }

    # image transformations called from the menu bar
    method flipX {} {
	[cget -image] flip X [expr {! [$image_ flip X]}]
    }
    method flipY {} {
	[cget -image] flip Y [expr {! [$image_ flip Y]}]
    }
    method rotate {} {
	[cget -image] rotate [expr {! [$image_ rotate]}]
    }

    # init bindings for "cut region"

    method cutregion {} {
	if {! [info exists cutWdg_]} {
	    set cutWdg_ $panelWdg_.cutregion
	    bind $canvas_ <Enter> +[code $this cutMakeRegion %x %y]
	    bind $canvas_ <ButtonRelease-1> +[code $this cutDone]
	    bind $canvas_ <ButtonPress-1> +[code $this cutInitResizeRegion 0 %x %y]
	    bind $canvas_ <Double-Button-2> +[code $this cutInitResizeRegion 1 %x %y]
	    bind $canvas_ <Motion> +[code $this cutMoveRegion %x %y]
	    bind $canvas_ <B1-Motion> +[code $this cutResizeRegion %x %y]
	    bind $canvas_ <Leave> +[code $this cutCancel %s]
	}
	$cutWdg_ config -relief sunken -command [code $this cutDone]
	if {[info exists xcutc_]} {
	    cutMakeRegion $xcutc_ $ycutc_
	}
	update
	return
    }

    # create a region for the cut window

    method cutMakeRegion {x y} {
	if {"[$cutWdg_ cget -relief]" == "raised"} { return }
	set cutregion_ [$canvas_ create rectangle 0 0 0 0 \
		-fill pink -stipple pat12 -width 3 -tags cutregion]
	cutMoveRegion $x $y
	return
    }

    # init the region for the cut window

    method cutInitResizeRegion {new x y} {
	if {! $cutregion_} { return }
	if {$new} {
	    set cutSizeX_ 50
	    set cutSizeY_ 50
	}
	set xcutr_ $x
	set ycutr_ $y
	cutMoveRegion $xcutm_ $ycutm_
    }

    # resize the cut window

    method cutResizeRegion {x y} {
	if {! $cutregion_} { return }
	lassign [cutResizeGet $x $xcutr_ $cutSizeX_] cutSizeX_
	lassign [cutResizeGet $ycutr_ $y $cutSizeY_] cutSizeY_
	set xcutr_ $x
	set ycutr_ $y
	cutMoveRegion $xcutm_ $ycutm_
    }

    # get size of cut window

    method cutResizeGet {pos posSv val} {
	if {$pos == $posSv} { return $val }
	set dp [expr {$pos > $posSv ? 5 : -5}]
	return [rtdbCheckRange [expr {$val + $dp}] 15.0 100.0]
    }

    # move cut window

    method cutMoveRegion {x y} {
	if {! $cutregion_} { return }
	if {[rtdbIsImageCleared]} {
	    cutCancel
	    return
	} 
	set xcutm_ $x
	set ycutm_ $y
	lassign [rtdbScreenToImage $x $y 0 0] x y d d
	lassign [rtdbCheckRange $x 0.0 [$image_ width]] x
	lassign [rtdbCheckRange $y 0.0 [$image_ height]] y
	set x0 [max [expr {$x - $cutSizeX_ / 2.0}] 0.0]
	set x1 [min [expr {$x + $cutSizeX_ / 2.0}] [$image_ width]]
	set y0 [max [expr {$y - $cutSizeY_ / 2.0}] 0.0]
	set y1 [min [expr {$y + $cutSizeY_ / 2.0}] [$image_ height]]

	lassign [rtdbImageToCanvas $x0 $y0 $x1 $y1] cx0 cy0 cx1 cy1
	$canvas_ coords cutregion $cx0 $cy0 $cx1 $cy1
	return "$x0 $y0 $x1 $y1"
    }

    # auto set cut levels within the cut region

    method cutDone { } {
	if {! $cutregion_ || ! [info exists xcutm_]} {
	    cutCancel
	    return
	}
	set xcutc_ $xcutm_
	set ycutc_ $ycutm_

	lassign [cutMoveRegion $xcutm_ $ycutm_] x0 y0 x1 y1
	set nx [expr {$x1 - $x0}]
	set ny [expr {$y1 - $y0}]

	cutCancel

	# get value list from region
        set vals ""
        set ar [$image_ get [expr {round($x0 + $nx / 2)}] [expr {round($y0 + $ny / 2)}] \
		    image [expr {round($nx)}] [expr {round($ny)}]]

        lassign [irtdMedianCut $ar [expr {$nx * $ny}]] low high
	$image_ cut $low $high
	$panelWdg_ updateValues
	return
    }

    # cancel cut window operation

    method cutCancel {{evtStateField 0}} {
	if {$evtStateField != 0} { return }
	$canvas_ delete cutregion
	set cutregion_ 0
	$cutWdg_ config -relief raised -command [code $this cutregion]
    }

    # -- public vars --

    # main RtdImage widget (set by caller)
    itk_option define -image image Image {} {
	set image_  [$itk_option(-image) get_image]
	set canvas_ [$itk_option(-image) get_canvas]
    }
    
    # specify TopLevelWidget to display short help messages
    itk_option define -shorthelpwin shortHelpWin ShortHelpWin {}

    # database root point for irtd
    itk_option define -irtdRoot irtdRoot IrtdRoot {<alias>iracq}

    # max. wait time for process irtdDart to respond (sec)
    itk_option define -waitDart waitDart WaitDart {5}

    # -- protected vars -- 

    # internal rtdimage widget for main image
    protected variable image_

    # canvas for main image
    protected variable canvas_

    # Rtd widgets
    protected variable RtdWdg_
    protected variable panelWdg_
    protected variable zoomWdg_

    # widget names of some control widgets
    protected variable startWdg_
    protected variable fponWdg_
    protected variable stfpWdg_
    protected variable dataWdg_
    protected variable fastWdg_

    # widget names of some test widgets
    protected variable rateWdg_
    protected variable countWdg_
    protected variable rapidWdg_
    protected variable cutWdg_

    # current data for main window
    protected variable mainData_ 2

    # data reception task stopped, flag
    protected variable dartStopped_ 1

    # ccs initialized flag
    protected variable ccsInit_

    # Current selected detector (0=SW, 1=LW)
    protected variable det_ 0

    # clock value
    protected variable clk_

    # cut region id
    protected variable cutregion_ 0

    # size of cut window
    protected variable cutSizeX_ 50
    protected variable cutSizeY_ 50

    # view of cut image
    protected variable cutImage_

    # Rtd image negated, flag
    protected variable negate_ 0

    # Current frames available
    protected variable frames_ 0

    # Screen coords of last move of cut window
    protected variable xcutm_
    protected variable ycutm_

    # Screen coords of last resize of cut window
    protected variable xcutr_
    protected variable ycutr_

    # Screen coords of last cut
    protected variable xcutc_
    protected variable ycutc_

    # Start chip coordinates
    protected variable x0_
    protected variable y0_

    # Window parameters
    protected variable winParms_ {-1 -1 -1 -1 -1}
}

# ------------------------------ init the plugin -------------------------

# define a procedure which is called by the menu items
# if flgRtd is set to 1 then we call the Rtd main widget

proc irtdRtdMenu {cmd {flgRtd 0}} {
    global ::irtd ::rtdb
    if {$flgRtd} {
	eval $rtdb(Rtd) $cmd
    } else {
	eval $irtd(Irace) $cmd
    }
}

# define a procedure which is called by the control widgets

proc irtdCtrl {args} {
    #puts "irtdCtrl: $args"
    global ::irtd
    eval $irtd(Irace) [join $args]
}

# procedure called via SCRIPT command ---

# select detector (e.g 0 for SW 1 for LW arm). Used for fixed pattern

proc irtdSelDet {det {id ""}} {
    global ::irtd
    set res [$irtd(Irace) selDet $det $id]
    return $res
}

# start acquisition

proc irtdStart {} {
    global ::irtd
    $irtd(Irace) set_start 1
    $irtd(Irace) start
}

# stop acquisition

proc irtdStop {} {
    global ::irtd
    $irtd(Irace) set_start 0
    $irtd(Irace) start
}

# create an instance of the irtdRtd class
# Note that we are running here in a global space

# load packages
set ext [info sharedlibextension]
set lib [cconcat libirtdc $ext]

foreach envdir "INTROOT VLTROOT" {
    if {! [info exists env($envdir)]} { continue }
    if {[file exists $env($envdir)/lib/$lib]} {
	load $env($envdir)/lib/$lib
	break
    }
}
if {[catch {package require Irtdc} msg]} {
    rtdbExit_dialog $msg
}

# check if argv list contains valid options for this plugin
set opts {}
set optlist "-irtdRoot -waitDart"; # list of valid class options
set i 0
set n 0
foreach el $argv {
    incr n
    if { ! [expr {[incr i] % 2}] } { continue }
    if { [lsearch -exact $optlist $el] != -1 } {
	append opts "$el [lindex $argv $n] "
    }
}

# create an instance of class irtdRtd
set irtd(Irace) $rtdb(Rtd).irace
eval irtdRtd $irtd(Irace) \
    -image [$rtdb(Rtd) get_image] \
    -shorthelpwin $rtdb(Rtd) \
    $opts

#
# ___oOo___

