#*******************************************************************************
# E.S.O. - VLT project
#
# "@(#) $Id: osrtdM2XY.tcl,v 1.20 2004/01/23 15:21:49 vltsccm Exp $"
#
# who       when      what
# --------  --------  ----------------------------------------------
# pbiereic  12/09/00  created

#************************************************************************
#   NAME
#   osrtdM2XY.tcl - draw a line for setting chopping parameters
# 
#   SYNOPSIS
#   osrtdM2XY <path> ?options?
# 
#   DESCRIPTION
#   osrtdM2XY draws a line in the canvas and returns the line parameters
#   (x0, y0), (x1, y1) in image coordinates. The returned parameters
#   can be convert by OS to set the chop throw and amplitude of M2.
#
#   FILES
#
#   ENVIRONMENT
#
#   RETURN VALUES
#
#   CAUTIONS
#
#   EXAMPLES
#
#   SEE ALSO
#   RTD documentation
#
#   BUGS     
#
#------------------------------------------------------------------------

itk::usual osrtdM2XY {}

class ::osrtdM2XY {
    inherit util::TopLevelWidget

    #  constructor: create a new instance of this class

    constructor {args} {
	eval itk_initialize $args
	wm protocol $w_ WM_DELETE_WINDOW "$this cancel"
	# make the control panel
	make_control
	make_help
	wm resizable $w_ 0 0
	set lineId_ -1
    }

    # create the control window

    method make_control {} {
        pack [frame $w_.status -borderwidth 2 -relief groove] \
            -side top -fill none -expand 0

	blt::table $w_.status
	foreach el {startx starty endx endy} {
	    itk_component add $el {
		LabelEntry $w_.status.$el \
		    -text "[string toupper $el]: " \
		    -command [code $this setValues] \
		    -validate numeric \
                    -anchor e \
		    -relief sunken \
		    -valuewidth 8 \
		    -labelwidth 8
	    }
	}

	blt::table $w_.status \
		$itk_component(startx)     1,0 -padx 10 \
		$itk_component(starty)     2,0 \
		$itk_component(endx)       1,1 \
		$itk_component(endy)       2,1

        pack [frame $w_.buttons -borderwidth 2 -relief groove] \
            -side top -fill x -expand 1

	blt::table $w_.buttons
	foreach el {ok clear cancel} {
	    itk_component add $el {
		button $w_.buttons.$el \
		    -text "[string toupper $el]" \
		    -pady 3 \
		    -width 8 \
		    -command [code $this $el]	    
	    }
	}

	blt::table $w_.buttons \
		$itk_component(ok)         2,0 \
		$itk_component(cancel)     2,1
    }

    # short help text

    method make_help {} {
	add_short_help $itk_component(cancel) \
	    {CANCEL: {bitmap b1} = Cancel setting M2 chopping}
	add_short_help $itk_component(ok) \
	    {OK: {bitmap b1} = Set M2 chopping}
	foreach el {startx starty endx endy} {
	    add_short_help $itk_component($el) \
		"[string toupper $el]: {bitmap b1} = Image/detector coordinate"
	}
    }

    # send offset parameters to OS

    method ok {} {
	set id $lineId_
	lassign [update_status] x0 y0 x1 y1
	if {$x0 == $y0 && $x1 == $y1} { return }
	rtdbReply "$x0 $y0 $x1 $y1"
	cancel 0
    }

    # lower window and delete line

    method cancel {{with_ack 1}} {
	if {$with_ack} {
	    rtdbReply -1
	}
	wm withdraw $w_
	deselect_line
	$draw_ set_drawing_mode anyselect	
    }

    # method which is called when values are entered

    method setValues {args} {
	foreach el {startx starty endx endy} {
	    set $el [$itk_component($el) get]
	    if {"[set $el]" == "" || [set $el] <=0 } {
		error_dialog "Illegal value in \"[string toupper $el]\""
		return
	    }
	}
	lassign [check_coords $startx $starty $endx $endy] x0 y0 x1 y1
	if {$lineId_ == -1} {
            lassign [rtdbCanvasToChip $x0 $y0 $x1 $y1 $target_image_] x0 y0 x1 y1
	    $draw_ create_object $x0 $y0
	    $draw_ update_line $x1 $y1
	    $draw_ check_done $x1 $y1
	} else {
            lassign [rtdbChipToCanvas $x0 $y0 $x1 $y1 $target_image_] x0 x1 y0 y1
	    $target_canvas_ coords $lineId_ $x0 $x1 $y0 $y1
	    $draw_ adjust_line_selection $lineId_
	    update_status
	}
    }

    # Draw a line via the draw widget

    method make_line { x0 y0 x1 y1 } {
	lassign [rtdbChipToCanvas $x0 $y0 $x1 $y1 $target_image_] x0 y0 x1 y1
	rtdbPopup $w_
	clearAll
	set lineId_ [create_line $x0 $y0 $x1 $y1]
	# add bindings for moving and resizing the rapid frame
	$draw_ add_object_bindings $lineId_
	$draw_ select_object $lineId_
	$draw_ add_notify_cmd $lineId_ "$w_ notify_cmd"
	update_status
    }

    method create_line {x0 y0 x1 y1} {
	set opts "-width    2 \
		-fill       green \
		-stipple    pat0 \
		-arrow      both \
		-arrowshape {8 10 3} \
		-tags {objects chop}"
	return [eval $target_canvas_ create line $x0 $y0 $x1 $y1 $opts]
    }

    method clearAll {} {
	deselect_line
	$target_canvas_ delete chop
	foreach el {startx endx starty endy} {
	    $itk_component($el) config -value ""
	}
    }

    # deselect line

    method deselect_line {} {
	if {$lineId_ == -1} {
	    return
	}
        $draw_ remove_notify_cmd $lineId_
	$draw_ deselect_object $lineId_
	set sequence [$target_canvas_ bind $lineId_]
	foreach el $sequence {
	    $target_canvas_ bind $lineId_ $el ""
	}
	set lineId_ -1
	$draw_ deselect_objects
    }


    # This method is called when the line was drawn, moved, resized or deleted. 

    method notify_cmd {op args} {
	# puts "Chop: $op $args"

	if {"$op" == "move" || "$op" == "resize"} {
	    update_status
	} elseif {"$op" == "delete"} {
	    if {$lineId_ != -1} {
		cancel
	    }
	} elseif {"$op" == "scale"} {
	    update_status
	} elseif {"$op" == "new_image"} {
	    if {[winfo viewable $w_]} {
		clear
	    } else {
		cancel
	    }
	} else {
	    update_status
	}
    }
    
    # check coordinates. Enter with x0,y0,x1,y1. Returns updated x0,y0,x1,y1.

    method check_coords {x0 y0 x1 y1} {
        set iw [$target_image_ width]
        set ih [$target_image_ height]
        set x0 [rtdbCheckRange $x0 1 $iw]
        set x1 [rtdbCheckRange $x1 1 $iw]
        set y0 [rtdbCheckRange $y0 1 $ih]
        set y1 [rtdbCheckRange $y1 1 $ih]
        return "$x0 $y0 $x1 $y1"
    }

    # update status display and return window parameters

    method update_status {} {
        if {$lineId_ == -1} {
            return
        }
        lassign [line_image_coord] x0 y0 x1 y1
        # check that frame is within max. size and in main image
        lassign [check_coords $x0 $y0 $x1 $y1] x0 y0 x1 y1
        lassign [format "%6.2f %6.2f %6.2f %6.2f" $x0 $y0 $x1 $y1] x0 y0 x1 y1

        $itk_component(startx) config -value $x0
        $itk_component(endx)   config -value $x1
        $itk_component(starty) config -value $y0
        $itk_component(endy)   config -value $y1

        return "$x0 $y0 $x1 $y1"
    }

    # image coordinates of the line

    method line_image_coord {} {
	lassign [$target_canvas_ coords $lineId_] x0 y0 x1 y1
	# convert to pixel coordinates
	lassign [rtdbCanvasToChip $x0 $y0 $x1 $y1 $target_image_] x0 y0 x1 y1
	return "$x0 $y0 $x1 $y1"
    }
    
    # -- public vars --
    
    # target rtdimage
    itk_option define -target_image target_image Target_image {} {
	# get internal widget names for target image
	set target_image_ [$itk_option(-target_image) get_image]
	set target_canvas_ [$itk_option(-target_image) get_canvas]
	set draw_ [$itk_option(-target_image) component draw]
    }

    # fonts used
    #itk_option define -labelFont labelFont LabelFont -Adobe-helvetica-bold-r-normal--12*
    #itk_option define -valueFont valueFont ValueFont -Adobe-helvetica-bold-r-normal--12*

    # set the state to normal/disabled to enable/disable widgets
    itk_option define -state state State {disabled} {}

    # set the width for  displaying labels and values
    #itk_option define -labelwidth labelWidth LabelWidth 5
    #itk_option define -valuewidth valueWidth ValueWidth 10

    # -- protected vars --

    # canvas id of the line object used to position the tel. offset
    protected variable lineId_
    
    # target internal rtdimage
    protected variable target_image_ 

    # canvas widget for main image and region object marking frame.
    protected variable target_canvas_ 

    # CanvasDraw object, for setting up move, resize operations on embedded image
    protected variable draw_

    # internal rtdimage for Chop frame
    protected variable image_

    # canvas window containing Chop frame image, 
    # different than target_canvas for popup frames
    protected variable canvas_
}
