#*******************************************************************************
# E.S.O. - VLT project
#
# "@(#) $Id: rtdbCtrl.tcl,v 1.24 2005/06/19 19:36:28 vltsccm Exp $"
#
# who       when      what
# --------  --------  ----------------------------------------------
# pbiereic  15/02/00  created
# pbiereic  12/12/00  use the 'pack' geometry manager instead of 'blt::table'
# pbiereic  27/10/03  changed 'body rtd::RtdImagePan::pan' and 'body rtd::RtdImage::make_spectrum'
#

#************************************************************************
#   NAME
#   rtdbCtrl.tcl - itcl widget combining the RtdImage itcl widget with
#                  a control panel, zoom and panning windows.
# 
#   SYNOPSIS
#   rtdbCtrl <path> ?options?
# 
#   DESCRIPTION
#   rtdbCtrl builds the control panel which consists of a zoom window, a pan
#   window, a cursor coordinates display, an image control panel, a color ramp etc.
#   The control panel is packed vertically on the left side of the window in
#   order to provide more space for the image display window.
#   rtdbCtrl inherits RtdImageCtrl and adds/redefines methods to:
#   - evaluate user defined camera pre- and post- commands
#   - define a window interactively (e.g. for detector windowed readout)
#   - compute statistics on a window
#   - start/stop cursor motions events (via <B2>)
#
#
#   PARENT CLASS
#        ::rtd::RtdImageCtrl
#
#   PUBLIC METHODS
#   The public methods are called via rtdbRtd(n) and are used internally
#   by the IRTD software. They are not supposed to be used by external plugins.
#
#   FILES
#
#   ENVIRONMENT
#
#   RETURN VALUES
#
#   CAUTIONS
#
#   EXAMPLES
#
#   SEE ALSO
#   RTD documentation, RtdImageCtrl(n)
#
#   BUGS     
#
#------------------------------------------------------------------------

itk::usual rtdbCtrl {}

class ::rtdbCtrl {
    inherit rtd::RtdImageCtrl

    constructor {args} {
	global ::rtdb env
	set rtdb(RtdImage) $w_
	eval itk_initialize $args
   }

    # the following method is called from the base class (TopLevelWidget) after all
    # the options have been evaluated

    method init {} {
	global ::rtdb

	$w_ config -canvasbackground #303030

 	rtd::RtdImageCtrl::init

	set panel $itk_component(panel)
        # Info panel, RtdImagePanel(n) object used to display image controls
        itk_component add info {
            eval ::rtdbPanel $panel.info -image $this -state disabled \
		    -showobject 0 -showtrans 0 -labelwidth 5 -valuewidth 12 \
		    [defaultOpts min_scale max_scale shorthelpwin]
	}

	# remove packings to avoid conflicts between the 2 geometry managers 'pack'
	# and 'blt::table'
	set list "zoom pan info"
	foreach compo $list {
	    set $compo $itk_component($compo)
	    set wdg [set $compo]
	    catch {pack forget $wdg}
	    catch {blt::table forget $wdg}
	    if {"$compo" != "info"} {
		$wdg config -borderwidth 2 -relief raised -background [cget -background]
	    }
	}

	foreach el [winfo children $zoom] {
	    catch {$el config -padx 0}
	    pack $el -ipadx 0 -padx 1 -expand 1 -fill x
	}
	pack $zoom.f -padx 2 -pady 2 -fill x
	pack [$pan component image] -side top -anchor c

	# remove scroll bars from images
	foreach obj "[[rtdbItclInfo] objects -class ::rtd::RtdImage]" {
	    catch {pack forget [$obj component hscrollf] [$obj component vscrollf]}
	    $obj config -borderwidth 0 -canvasbackground #202020
	}

	$panel config -borderwidth 1 -relief sunken

	# change packing order of the panel subwindows
	set indices [rtdbIdx $list CTRLWIN]
	foreach idx $indices {
	    set el [lindex $list [expr {$idx - 1}]]
	    if { "$el" != {} } {
		pack $itk_component($el) -fill none -anchor c -expand 0 -side top -padx 2 -pady 0
	    }
	}
	if {! [lempty $indices]} {
	    pack $panel -padx 0 -anchor nw -side left -fill both -before $w_.imagef
	}

        # flash background color of object label for real-time events
        config -cameraPreCmd "$this preCmd" -cameraPostCmd "$this postCmd"

        $canvas_ bind $image_ <Lock-Motion> "[code $info stopXY 2]"
	$canvas_ bind $image_ <ButtonRelease-2> "[code $this motion_event]"
	after idle [code $this hello_in_canvas]
    }

   # make the panel info subwindow
    
    public method make_panel_info {panel} {
    }

    # return options string with default values

    method defaultOpts {args} {
	# pass on the default options
	foreach el $args {
	    append opt "-$el \{$itk_option(-$el)\} "
	}
	return $opt
    }
    
    # redefine method make_toolbox: make the graphics toolbox and menu

    body rtd::RtdImage::make_toolbox {} {
	foreach el "shorthelpwin withtoolbox show_object_menu regioncommand" {
	    append opts "-$el \{$itk_option(-$el)\} "
	}
        itk_component add draw {
            eval ::rtdbDraw $w_.draw \
		    -canvas $canvas_ \
		    -defaultcursor {$itk_option(-cursor)} \
		    -transient 1 -withdraw 1 -center 0 $opts
        }
        set_drawing_area

        # clicking on the image or image background deselects other objects
        $canvas_ bind $image_ <1> [code $itk_component(draw) deselect_objects]
    }

    # pass these methods on to the panel widget unchanged
    # (this just generates methods on the fly...)

    ::foreach i {add_control camera_ready set_flash get_panel add_updateValues \
	    add_transCmd} {
	method $i {args} [::format {return [eval "$itk_component(info) %s ${args}"]} $i]
    }

    # redefine method make_spectrum: Create a graph to display the image data values along a line
    
    body rtd::RtdImage::make_spectrum {line_id x0 y0 x1 y1} {
	if {[winfo exists $w_.spectrum]} {
	    $w_.spectrum quit
	}
	eval ::rtdbSpectrum $w_.spectrum \
	    -x0 [expr {int($x0)}] \
	    -y0 [expr {int($y0)}] \
	    -x1 [expr {int($x1)}] \
	    -y1 [expr {int($y1)}] \
		-image $this \
		-transient 1 \
		-shorthelpwin $itk_option(-shorthelpwin) \
		-line_id $line_id
    }

    body rtd::RtdImagePan::pan {x1 y1 x2 y2 changed} {
	global ::rtdb
	draw_compass;  # make sure the compass is displayed
	if { [info exists rtdb(pan_x1)] } {
	    if { $changed && $x1 == $rtdb(pan_x1) && $y1 == $rtdb(pan_y1) && \
		     $x2 == $rtdb(pan_x2) && $y2 == $rtdb(pan_y2) && \
		     $panImageWidth_ > 1 } {
		return
	    }
	}
	set rtdb(pan_x1) $x1
	set rtdb(pan_y1) $y1
	set rtdb(pan_x2) $x2
	set rtdb(pan_y2) $y2
	set rtdb(pan_changed) $changed

	if {$changed} {
	    init_panning
	} else {
	    if {$x1 == 0 && $y1 == 0 && abs($x2-$panImageWidth_)<3 && abs($y2-$panImageHeight_)<3} {
		$itk_option(-target_image) center
	    }
	    $canvas_ coords $marker_ [incr x1 -2] [incr y1 -2] [incr x2 2] [incr y2 2]
	    $canvas_ coords $panner_ $x1 $y1 $x2 $y2
	}
    }

    # method preCmd is evaluated before the image is displayed

    public method preCmd {args} {
	global ::rtdb
	disableAsyncEvent
	wm title $rtdb(Rtd) "$rtdb(APPNAME) Real-Time Display $rtdb(APPVERS) \
		[clock format [clock seconds] -format %T]"
	config -file ""

	evalCmds $preCmd_
	$itk_component(info) flash 1
    }

    # evaluate user commands

    public method evalCmds {args} {
	if {"$args" == ""} { return }
	foreach cmd [lindex $args 0] {
	    if {[catch {eval $cmd} msg]} {
		rtdbFeedback $msg
	    }
	}
    }

    # method postCmd is evaluated after the image was displayed

    public method postCmd {} {
	global ::rtdb
	$itk_component(info) flash 0
	evalCmds $postCmd_
	if {$wait_next_image_} {
	    set wait_next_image_ 0
	    rtdbReply 0
	}
	enableAsyncEvent
    }

    # add command which is evaluated before a real-time image display is started

    method add_preCmd { cmd } {
	lappend preCmd_ $cmd
    }

    # add command which is evaluated after a real-time image display

    method add_postCmd { cmd } {
	lappend postCmd_ $cmd
    }

    method add_rapidPreCmd { cmd } {
	lappend rapidPreCmd_ $cmd
    }

    method add_rapidPostCmd { cmd } {
	lappend rapidPostCmd_ $cmd
    }

    # add command which is evaluated when the image was scaled

    method add_scaleCmd { cmd } {
	lappend scaleCmd_ $cmd
    }

    # add command which is evaluated when the image X or Y pointer coordinate changes

    method traceXY { cmd } {
        global ::$image_
	trace variable ${image_}(X) w "[code $this movedXY]"
	trace variable ${image_}(Y) w "[code $this movedXY]"
	set traceXYCmd_ $cmd
    }

    # disable asynchrounous file events from seqWish

    method disableAsyncEvent {} {
	global ::rtdb
	if {! $rtdb(CCSINIT)} { return }
	set rtdb(asyncCallback) [fileevent $rtdb(CCS_FILE) readable]
	fileevent $rtdb(CCS_FILE) readable ""
    }

    # enable asynchrounous file events from seqWish

    method enableAsyncEvent {} {
	global ::rtdb
	if {! $rtdb(CCSINIT)} { return }
	fileevent $rtdb(CCS_FILE) readable $rtdb(asyncCallback)
    }

    # set flag when waiting for next image

    method wait_next_image {} {
	set wait_next_image_ 1
    }

    # return name of the pan window

    public method get_pan {args} {
	return $itk_component(pan)
    }

    # return name of the zoom window

    public method get_zoom {args} {
	return $itk_component(zoom)
    }

    # toggle the visibility of the control panel 
    
    public method hide_control_panel {variable} {
        set panel $itk_component(panel)
        global ::$variable ::$panel.zoom.dozoom

        if {[set $variable]} {
            # hide the panel, turn off the zoom window, keep the canvas width
            set zoom_state_ [set $panel.zoom.dozoom]
            set $panel.zoom.dozoom 0
            set w [winfo width $canvas_]
            pack forget $itk_component(panel)
            $canvas_ config -width $w
        } else {
            # show the panel, restore the zoom window
            set $panel.zoom.dozoom $zoom_state_
	    pack $itk_component(panel) -anchor nw -side left -fill x -before $w_.imagef
        }
    }

    # clear the current image display and remove any windows that
    # access it. Redefined method.

    public method clear {} {
        if {[info exists itk_component(statistics)]} {
            $itk_component(statistics) lower_dwin
        } 
	rtd::RtdImageCtrl::clear
    }

    # this method is called by the image code whenever a new image is loaded.
    # (for real-time updates, see camera command)

    method new_image_cmd {} {
	global ::rtdb

	rtd::RtdImageCtrl::new_image_cmd
	if {! [$image_ isclear]} {
	    catch {$canvas_ delete hello}
	} else {
	    wm title $rtdb(Rtd) "$rtdb(APPNAME) Real-Time Display $rtdb(APPVERS)"
	}
	evalCmds $notify_
    }

    # add command which is evaluated after a new image event

    method add_notify { cmd } {
	lappend notify_ $cmd
    }

    # remove command from the list for the new image event

    method remove_notify { cmd } {
	lappend notify_ $cmd
	regsub -all $cmd $notify_ "" notify_
    }

    # open and load a new FITS image file via file name dialog
    
    public method open {{dir "."} {pattern "*.*fit*"}} {
	global ::rtdb
	if {! [info exists rtdb(DATADIR)]} {
	    set dir [rtdbGetDataDir]
	    set rtdb(DATADIR) $dir
	}
        set file [filename_dialog $dir $pattern $w_]
        if {"$file" != ""} {
            if {[file isfile $file]} {
                after 0 "$this load_file $file"
            } else {
                error_dialog "There is no file named '$file'" $w_
            }
        }
    }

    # call pick widget with the following argument options:
    # NOW  x0 y0 nx : pick object now at (x0, y0). nx is the size of the pick window
    # WAIT x0 y0 nx : wait until next image is displayed and pick object then at (x0, y0).
    # no args : wait until user has picked an object

    public method pick {args} {
	if {[check_image 1]} {
	    rtdbReply -1
	    return
	}
        set what [lindex $args 0]
	if {"$args" == ""} {
	    set parms ""
	} else {
	    set parms [lreplace $args 0 0]
	}
	set wait 0
	if {"$what" == "WAIT"} {
	    set wait 1
	}
	if {! [info exists itk_component(pick)]} {
	    pick_dialog "" 0 ; # popup pick window
	}
	$itk_component(pick) set_result_value ; # init values
	$itk_component(pick) pick_object $parms $wait ; # pick the object
    }

    # display a dialog for selecting objects in the image and displaying information
    # about the selected area of the image. Redefined method.

    public method pick_dialog {{command ""} {pick 1}} {
        if {[check_image 1]} { return }

	if {! [rtdbPopup $w_.pick]} {
	    itk_component add pick {
		eval ::rtdbPick $w_.pick -target_image $this \
			-command {$command} -labelwidth 14 -valuewidth 10 \
			-orient $itk_option(-pickobjectorient) \
			-transient 1 -withdraw 1 -center 0 \
			[defaultOpts verbose debug shorthelpwin]
	    }
	    $itk_component(pick) center_window
	    update idletasks ; # method init (after job)
	    wm resizable $w_.pick 0 0
	}
	if {$pick} {
	    $w_.pick pick_object
	}
    }


    # load a file and display its name on the title bar

    public method load_file {file} {
	global ::rtdb
	if {![catch {config -shm_data 0 -shm_header 0 -file $file} msg]} {
	    if {"[cget -file]" != ""} {
		wm title $rtdb(Rtd) $file
		return 1
	    }
	}
	return 0
   }
    
    # display hello text in canvas
    
    method hello_in_canvas {} {
	if {! [$image_ isclear]} { return }
        global ::rtdb
	set fontH "-*-courier-Bold-r-*-*-*-240-*-*-*-*-iso8859-*"
	$canvas_ create text 0 -200 \
		-text "$rtdb(APPNAME) Real-Time Display $rtdb(APPVERS)\n$rtdb(APPDESC)" \
		-fill white -font $fontH -justify center -tags hello
    }

    # check if image is loaded

    method check_image {{warning 0}} {
	if {[$image_ isclear]} {
	    if {$warning} {
		warning_dialog " No image is currently loaded" $w_
	    }
	    return 1
	}
	return 0
    }

    # method to define a readout window interactively

    method defwin {} {
        if {[check_image 1]} {
            rtdbReply -1
	    return
        }
        set win $w_.defwin
        if {! [rtdbPopup $win]} {
            itk_component add defwin {
                eval ::rtdbDefwin $win -target_image $this -with_control 1 \
			-title {"Window parameters"} [defaultOpts shorthelpwin]
            }
	    update
        }
        $itk_component(defwin) raise_dwin
    }

    method save_file {file} {
	global ::env
        if {[check_image 1]} {
            rtdbReply -1
	    return
        }
	if {[string first "/" $file] == -1 && [info exists env(INS_ROOT)]} {
	    set file "$env(INS_ROOT)/SYSTEM/DETDATA/$file"
	}
	if {[catch {set ret [save_as_file $file 1]}]} {
	    rtdbReply -1
	} else {
	    rtdbReply "0 $ret"
	}
    }

    # save file

    method save_as_file {file {increment 0}} {
	if {"$file" == "" || [check_image]} { return }
	if { $increment } {
	    set i 0
	    set ext  [file extension $file]
	    set fnam [file rootname $file]
	    set fincr $fnam\_0
	    while { [file isfile $fincr$ext] } {
		incr i
		set fincr $fnam\_$i
	    }
	    set file $fincr$ext
	}
	if {[file isfile $file]} {
	    if {![confirm_dialog " $file exists - Do you want to overwrite it ?" $w_]} {
		return
	    }
	    if {[file isdir $file]} {
		error_dialog " $file is a directory"
		return -1
	    }
	}
	$image_ dump $file
	if { $increment } {
	    chmod 0444 $file
	}
	return $file
    }

    # Select an area in the image and display statistics

    method statistics {} {
	if {[check_image 1]} {
	    return
	}
	set win $w_.statistics
	if {! [rtdbPopup $win]} {
	    itk_component add statistics {
		eval ::rtdbStat $win -target_image $this  \
			-title "Statistics" -buttons {"default cancel"} \
			-fillcolor orange -outlinecolor green4 \
			[defaultOpts shorthelpwin]
	    }
	}
	update
	$itk_component(statistics) raise_dwin
   }

    # this method is called when the mouse pointer is moved within the image

    method movedXY {args} {
	if {[$image_ isclear]} { return }
        global ::$image_
	set x [set ${image_}(X)]
	set y [set ${image_}(Y)]
	if {$x < 1 || $x > [$image_ width] || $y < 1 || $y > [$image_ height]} { return }
	if {"$traceXYCmd_" == ""} { return }
	if {[catch {eval $traceXYCmd_ $x $y} msg]} {
	    trace variable ${image_}(X) w {}
	    trace variable ${image_}(Y) w {}
	    warning_dialog "X/Y trace disabled !!\n$msg"
	}
    }

    # redefined 'scale' method

    method scale {x y} {
	RtdImage::scale $x $y
	$itk_component(info) transCmd
	evalCmds $scaleCmd_
    }

    # method to start/stop motion event in canvas

    method motion_event {args} {
	set motion_ [expr {! $motion_}]
	set imgObjects [[rtdbItclInfo] objects -class rtd::RtdImage]

	foreach el "$w_ $imgObjects" {
	    catch {[$el get_image] motionevent $motion_}
	}
	$image_ warp 0 1       ;# make a small move
	$image_ warp 0 -1

	$itk_component(info) stopXY $motion_
	$image_ update
	update idletasks
    }

    # redefine the following methods to process rapid frame events

    public method camera_pre_command {frameid} {
	if {$frameid == 0} {
	    rtd::RtdImage::camera_pre_command $frameid
	    return
	}
	evalCmds $rapidPreCmd_
    }

    public method camera_post_command {frameid} {
 	if {$frameid == 0} {
	    rtd::RtdImage::camera_post_command $frameid
	    return
	}
	evalCmds $rapidPostCmd_
    }

    # -- public vars --

    # -- protected vars --

    # list of commands to be evaluated for various events
    protected variable notify_ ""
    protected variable preCmd_ ""
    protected variable postCmd_ ""
    protected variable rapidPreCmd_ ""
    protected variable rapidPostCmd_ ""
    protected variable scaleCmd_ ""
    protected variable traceXYCmd_ ""

    # saved zoom button state
    protected variable zoom_state_ 0

    # accept motion events, flag
    protected variable motion_ 1

    # flag which is set when waiting for next real-time image
    protected variable wait_next_image_ 0
}

# ___oOo___
