#*******************************************************************************
# E.S.O. - VLT project
#
# "@(#) $Id: rtdbDefwin.tcl,v 1.24 2005/06/19 19:36:29 vltsccm Exp $"
#
# who       when      what
# --------  --------  ----------------------------------------------
# pbiereic  15/02/00  created
# pbiereic  05/11/03  workaround bug in tcl8.4.3
#

#************************************************************************
#   NAME
#   rtdbDefwin.tcl - itk class to define a window for windowed readout
# 
#   SYNOPSIS
#   rtdbDefwin <path> ?options?
# 
#   DESCRIPTION
#   rtdbDefwin allows to define interactively a window for windowed readout.
#   The size and position of the window are specified with a rectangle in
#   the main window.
#   It return the lower left and upper right chip coordinates of the rectangle
#
#   FILES
#
#   ENVIRONMENT
#
#   RETURN VALUES
#
#   CAUTIONS
#   none
#
#   EXAMPLES
#
#   SEE ALSO
#   RTD documentation
#
#   BUGS     
#
#------------------------------------------------------------------------

itk::usual rtdbDefwin {}

class ::rtdbDefwin {
    inherit util::TopLevelWidget

    #  constructor: create a new instance of this class

    constructor {args} {
	eval itk_initialize $args
    }

    method init {args} {
	global ::rtdb
	# make the control panel
	if {$with_control_} {
	    make_control
	    make_help
	    set regionId_ -1
	    wm resizable $w_ 0 0
	}
	wm protocol $w_ WM_DELETE_WINDOW "$this cancel"
	$rtdb(Rtd) add_transCmd [code after idle "$this notify_cmd move"]
	$rtdb(Rtd) add_scaleCmd [code after idle "$this notify_cmd move"]
    }

    # create the defwin frame window

    method make_control {} {
        itk_component add status {
	    frame $w_.status \
		-borderwidth 2 -relief groove
	}
        pack $itk_component(status) -side top -fill none -expand 0

	foreach el {startx starty nx ny endx endy} {
	    itk_component add $el {
		LabelEntry      $w_.status.$el \
		    -text       "[string toupper $el]: " \
		    -command    [code $this setValues] \
                    -anchor     w \
		    -relief     sunken \
		    -validate   numeric \
		    -valuewidth 8 \
		    -labelwidth 8
	    }
	}
	foreach el {endx endy} {
	    $itk_component($el) config -relief groove -command ""
	}

	blt::table $w_.status \
		$itk_component(startx)     1,0 -padx 10 \
		$itk_component(starty)     1,1 \
		$itk_component(endx)       2,0 \
		$itk_component(nx)         3,0 \
		$itk_component(endy)       2,1 \
		$itk_component(ny)         3,1 

        itk_component add buttons {
	    frame $w_.buttons \
		-borderwidth 2 -relief groove
	}
        pack $itk_component(buttons) -side top -fill x -expand 1

	set i 0
	foreach el $itk_option(-buttons) {
	    itk_component add $el {
		button $w_.buttons.$el \
		    -text    "[string toupper $el]" \
		    -pady    3 \
		    -width   8 \
		    -command [code $this $el]	    
	    }
	    blt::table $w_.buttons $itk_component($el) 1,$i
	    incr i
	}

    }

    # short help text

    method make_help {} {
	if {[info exists itk_component(cancel)]} {
	    add_short_help $itk_component(cancel) \
		{CANCEL: {bitmap b1} = Cancel defining windowed readout}
	}
	if {[info exists itk_component(default)]} {
	    add_short_help $itk_component(default) \
		{DEFAULT: {bitmap b1} = Set the window to the center of the image.}
	}
	if {[info exists itk_component(ok)]} {
	    add_short_help $itk_component(ok) \
		{OK: {bitmap b1} = Set the window parameters}
	}
	foreach el {nx ny} {
	    if {[info exists itk_component($el)]} {
		add_short_help $itk_component($el) \
		    "[string toupper $el]: {bitmap b1} = Number of pixels"
	    }
	}
	foreach el {startx starty endx endy} {
	    if {[info exists itk_component($el)]} {
		add_short_help $itk_component($el) \
		    "[string toupper $el]: {bitmap b1} = Image/detector coordinate"
	    }
	}
    }

    # method which is called when values are entered

    method setValues {args} {
	if {[$target_image_ isclear]} {
	    return
	}
	foreach el {startx starty nx ny} {
	    set $el [$itk_component($el) get]
	    if {"[set $el]" == "" || [set $el] <=0 } {
		error_dialog "Illegal value in \"[string toupper $el]\""
		return
	    }
	}
	set endx [expr {$startx + $nx -1}]
	set endy [expr {$starty + $ny -1}]
	remove_region
	lassign "$startx $starty $endx $endy" x0_ y0_ x1_ y1_
	set bug "$x0_ $y0_ $x1_ $y1_"  ; # workaround bug in tcl8.4.3
	raise_dwin
    }

    # cancel window

    method cancel {} {
	update
	lower_dwin
	rtdbReply -1
    }

    # set default window parameters

    method default {} {
	update
	catch {remove_region}
	set x0_ -1
	raise_dwin
    }

    # send window parameters to DCS

    method ok {} {
	update
	lassign [update_status] x0 y0 x1 y1
	lower_dwin
	rtdbReply "$x0 $y0 [expr {$x1-$x0+1}] [expr {$y1-$y0+1}]"
    }

    # create a dummy rect to get events for image

    method make_rect {} {
	lassign [$target_canvas_ bbox $regionId_] x0 y0 x1 y1
        set xoffset [expr {int($x0)}]
        set yoffset [expr {int($y0)}]
        set width   [expr {int($x1-$x0+1)}]
        set height  [expr {int($y1-$y0+1)}]

	set rectId_ [$target_canvas_ create rectangle \
		$xoffset $yoffset \
			 [expr {$xoffset+$width}] [expr {$yoffset+$height}] \
		-tags    $w_\defwinrect \
		-fill    black \
		-outline $itk_option(-outlinecolor) \
		-width   2 \
		-stipple pat7]

	$target_canvas_ itemconfigure $w_\defwin -width 2 -stipple pat1

	# add bindings for moving and resizing the defwin frame
	$draw_ add_object_bindings $rectId_ $regionId_
    }


    # This method is called whenever the associated rectangular of the defwin frame
    # is moved, resized or deleted. 

    method notify_cmd {op args} {
	if {[$target_image_ isclear]} {
	    return
	}
	update
	if {"$op" == "new_image"} {

	} elseif {"$op" == "resize_start"} {
	    return

	} elseif {"$op" == "move" || "$op" == "resize"} {
	    if {$args == ""} {
		move_dwin
	    } else {
		move_dwin $args
	    }

	} elseif {"$op" == "delete"} {
	    cancel

	} elseif {"$op" == "scale" && $regionId_ != -1} {
	    lassign [$target_canvas_ bbox $regionId_] x0 y0 x1 y1
	    $target_canvas_ coords $rectId_ $x0 $y0 $x1 $y1

	} elseif {"$op" == "new_image"} {
	    set firstRf 1
	    notify_cmd move    

	} else {
	    # rotate, flip
	}

	if {"$itk_option(-notify)" != "" && $regionId_ != -1} {
	    lassign [update_status] x0 y0 x1 y1
	    eval $itk_option(-notify) $op $x0 $y0 [expr {$x1-$x0+1}] [expr {$y1-$y0+1}]
	}
    }


    # check coordinates of RF. Enter with x0,y0,x1,y1. Returns updated x0,y0,x1,y1.

    method check_coords_dwin {x0 y0 x1 y1 {ps "sw"}} {
	set iw [$target_image_ width]
	set ih [$target_image_ height]
	set x0 [rtdbCheckRange $x0 1 $iw]
	set x1 [rtdbCheckRange $x1 1 $iw]
	set y0 [rtdbCheckRange $y0 1 $ih]
	set y1 [rtdbCheckRange $y1 1 $ih]
	return "$x0 $y0 $x1 $y1"
    }

    # move the defwin frame 

    method move_dwin {{ps "sw"}} {
	if {$regionId_ == -1} { return }
	# get image coordinates of region
	lassign [region_image_coord] x0 y0 x1 y1
	# save the region's coordinates
	lassign "$x0 $y0 $x1 $y1" x0_ y0_ x1_ y1_
	set bug "$x0_ $y0_ $x1_ $y1_"  ; # workaround bug in tcl8.4.3

	# check defwin frame is within max. size and in main image
	lassign [check_coords_dwin $x0 $y0 $x1 $y1 $ps] x0 y0 x1 y1

	# move rectangle and region
	lassign "$y1 $y0" y0 y1
	lassign [rtdbImageToCanvas $x0 $y0 $x1 $y1 $target_image_] sx0 sy0 sx1 sy1
	# coordinates are approx. 3 bins smaller than boundary box
	set cx0 [shrink_coord $sx0 $sx1]
	set cx1 [shrink_coord $sx1 $sx0]
	set cy0 [shrink_coord $sy0 $sy1]
	set cy1 [shrink_coord $sy1 $sy0]
	$target_canvas_ coords $regionId_ $cx0 $cy0 $cx1 $cy0 $cx1 $cy1 $cx0 $cy1 $cx0 $cy0

	lassign [$target_canvas_ bbox $regionId_] cx0 cy0 cx1 cy1
	$target_canvas_ coords $w_\defwinrect $cx0 $cy0 $cx1 $cy1
	$draw_ adjust_object_selection $regionId_

	update_status

    }

    # update status display and return window parameters

    method update_status {} {
	lassign [region_image_coord] x0 y0 x1 y1
	lassign [rtdbRoundCoords $x0 $y0 $x1 $y1] x0 y0 x1 y1
	# check that frame is within max. size and in main image
	lassign [check_coords_dwin $x0 $y0 $x1 $y1] x0 y0 x1 y1
	lassign [rtdbImageToChip $x0 $y0 $x1 $y1 $target_image_] x0 y0 x1 y1

	if {$with_control_} {
	    $itk_component(nx) config -value [expr {$x1 - $x0 +1}]
	    $itk_component(ny) config -value [expr {$y1 - $y0 +1}]
	    
	    $itk_component(startx) config -value $x0
	    $itk_component(endx)   config -value $x1
	    $itk_component(starty) config -value $y0
	    $itk_component(endy)   config -value $y1
	}
	return "$x0 $y0 $x1 $y1"
    }

    # image coordinates of region

    method region_image_coord {} {
	lassign [$target_canvas_ bbox $regionId_] x0 y0 x1 y1
	# convert to pixel coordinates
	lassign [rtdbCanvasToImage $x0 $y0 $x1 $y1 $target_image_] x0 y0 x1 y1
	if {$y0 > $y1} {
	    lassign "$y1 $y0" y0 y1
	}
	if {$x0 > $x1} {
	    lassign "$x1 $x0" x0 x1
	}
	return "$x0 $y0 $x1 $y1"
    }

    # lower defwin frame and bounding box

    method lower_dwin {} {
	catch {wm withdraw $w_}
	if {[$target_canvas_ find withtag $w_\defwin] == ""} {
	    return
	}
	remove_region
    }

    # remove region

    method remove_region {} {
	if {$regionId_ == -1} { return }
        $draw_ remove_notify_cmd $regionId_
	$target_canvas_ delete $w_\defwin $w_\defwinrect grip.$regionId_
	set regionId_ -1
    }
    
    # raise defwin frame and bounding box

    method raise_dwin {} {
	rtdbPopup $w_
	catch {remove_region}
	if {$x0_ == -1} {
	    make_region
	    make_rect
	    lassign [region_image_coord] x0_ y0_ x1_ y1_
	    set bug "$x0_ $y0_ $x1_ $y1_"  ; # workaround bug in tcl8.4.3

	} else {
	    lassign [rtdbImageToCanvas $x0_ $y0_ $x1_ $y1_ $target_image_] x0 y0 x1 y1

	    # coordinates are approx. 3 bins smaller than boundary box
	    set x0 [shrink_coord $x0 $x1]
	    set x1 [shrink_coord $x1 $x0]
	    set y0 [shrink_coord $y0 $y1]
	    set y1 [shrink_coord $y1 $y0]

	    set regionId_ [$target_canvas_ \
		    create line $x0 $y0 $x1 $y0 $x1 $y1 $x0 $y1 $x0 $y0 \
		    -width 0 \
		    -fill $itk_option(-fillcolor) \
		    -tags $w_\defwin]
	    make_rect
	}

	# setup callbacks for moving and resizing
	$draw_ add_notify_cmd $regionId_ [code $this notify_cmd]
	move_dwin
    }

    # shrink distance between points by 3

    method shrink_coord {x0 x1} {
	if {$x0 < $x1} {
	    set x0 [expr {$x0 + 3}]
	} else {
	    set x0 [expr {$x0 - 3}]
	}
	return "$x0"
    }

    # create a region in the target image

    method make_region {} {
        set w [$target_image_ width]
        set h [$target_image_ height]
        set ffwidth $itk_option(-fframeWidth)
        # create a square bounding box in the center of the image
        lassign [rtdbCenterBox $w $h $ffwidth] x0 y0 x1 y1
	lassign [rtdbImageToCanvas $x0 $y0 $x1 $y1 $target_image_] x0 y0 x1 y1

	set regionId_ [$target_canvas_ \
		create line $x0 $y0 $x1 $y0 $x1 $y1 $x0 $y1 $x0 $y0 \
		-width 0 \
		-fill $itk_option(-fillcolor) \
		-tags $w_\defwin]
    }
    
    # -- public vars --
    
    # target rtdimage
    itk_option define -target_image target_image Target_image {} {
	# get internal widget names for target image
	set target_image_  [$itk_option(-target_image) get_image]
	set target_canvas_ [$itk_option(-target_image) get_canvas]
	set draw_          [$itk_option(-target_image) component draw]
    }

    # set the state to normal/disabled to enable/disable widgets
    itk_option define -state state State {disabled} {}

    # command which is evaluated whenever the rectangle was moved
    itk_option define -notify notify Notify {}

    # the buttons to be displayed
    itk_option define -buttons bottons Buttons {ok default cancel}

    # the colors for the rectange
    itk_option define -fillcolor fillcolor Fillcolor {blue}
    itk_option define -outlinecolor outlinecolor Outlinecolor {yellow}

    # set the width for  displaying labels and values
    itk_option define -labelwidth labelWidth LabelWidth 5
    itk_option define -valuewidth valueWidth ValueWidth 10

    # default width of fast frame in #of pixels
    itk_option define -fframeWidth fframeWidth FframeWidth 100
    
    # specify TopLevelWidget to display short help messages
    itk_option define -shorthelpwin shortHelpWin ShortHelpWin {}
    
    # option for making the control window
    itk_option define -with_control with_control With_control {1} {
	set with_control_ [cget -with_control]
    }

    # -- protected vars --

    # option for making the control window, flag
    protected variable with_control_

    # canvas id of the (region) object used to position and move the image 
    # in the canvas
    protected variable regionId_
    
    # target internal rtdimage
    protected variable target_image_ 

    # canvas widget for main image and region object marking frame.
    protected variable target_canvas_ 

    # CanvasDraw object, for setting up move, resize operations on embedded image
    protected variable draw_
    
    # canvas id of rectangle used to get events for moving/resizing image
    protected variable rectId_

    # bounding box of region
    protected variable x0_ -1
    protected variable y0_ -1
    protected variable x1_ -1
    protected variable y1_ -1
}

# ___oOo___
