#*******************************************************************************
# E.S.O. - VLT project
#
# "@(#) $Id: rtdbRtd.tcl,v 1.24 2005/06/19 19:36:29 vltsccm Exp $"
#
# who       when      what
# --------  --------  ----------------------------------------------
# pbiereic  15/02/00  created
# pbiereic  12/09/00  enable remote interfaces after all widgets were
#                     created and initialized
#

#************************************************************************
#   NAME
#   rtdbRtd.tcl - Toplevel widget for the base RTD application.
# 
#   SYNOPSIS
#   rtdbRtd <path> ?options?
# 
#   DESCRIPTION
#
#   This class defines the top level window for the base RTD.
#   It inherits the Rtd toplevel widget, redefines and adds the application
#   specific methods which are used to
#   - load the plugins
#   - add menu items
#   - configure menu the bar according to user definitions.
#
#   PARENT CLASS
#        ::rtd::Rtd
#
#   PUBLIC METHODS
#   
#        add_transCmd {cmd}
#              Add a command which is evaluated whenever the image transformations
#              change (rotate, flip)
#
#        add_scaleCmd {cmd}
#              Add a command which is evaluated when the image was scaled
#
#        add_rapidPreCmd {cmd}
#              Add a command which is evaluated before a rapid frame is displayed
#
#        add_rapidPostCmd {cmd}
#              Add a command which is evaluated after a rapid frame was displayed
#
#        add_updateValues {cmd}
#              Add a command which is evaluated when cut levels were changed
#
#        add_notify {cmd}
#              Add a command which is evaluated whenever a new image was 
#              displayed which has different size or data type than the previous 
#              image. The command will also be evaluated when the image was 
#              cleared.
#
#        get_panel
#              Return the pathName of the panel widget
#
#        get_image
#              Return the pathName of the Rtd image widget
#
#        get_canvas
#              Return the pathName of the canvas of the main Rtd image widget
#
#        get_pan
#              Return the pathName of the pan window
#
#        traceXY
#              Define a command which is evaluated when the mouse pointer is
#              moved in the main window. The image pointer coordinates x, y
#              are passed as parameters.
#              traceXY can be used, e.g. to display the wavelength
#
#   FILES
#
#   ENVIRONMENT
#
#   RETURN VALUES
#
#   CAUTIONS
#
#   EXAMPLES
#
#   SEE ALSO
#   RTD documentation, Rtd(n), TopLevelWidget(n)
#
#   BUGS     
#
#------------------------------------------------------------------------

itk::usual rtdbRtd {}

class ::rtdbRtd {
    inherit ::rtd::Rtd

    constructor {args} {
	global ::rtdb
	set rtdb(Rtd) $w_

	set args [eval check_args $args]
	set rtdb(MAKING_MENU) 1
	eval itk_initialize $args
    }

    destructor {
        rtdbTerminate
    }

    # workaround for tcl's panic code with "Tcl_AppendStringsToObj called with
    # shared object" when an option argument is not valid

    method check_args {args} {
	set arglist ""
	foreach el [$this configure] {
	    append arglist "[lindex $el 0] "
	}
	set argvs ""
	set i 0
	set n 1
	foreach el $args {
	    if {[set i [expr {! $i}]]} {
		if {[lsearch -exact $arglist $el] != -1} {
		    append argvs "$el [lindex $args $n] "
		} else {
		    rtdbFeedback "Skipping option '$el [lindex $args $n]' which maybe used by a plugin"
		}
	    }
	    incr n
	}
	return $argvs
    }

    # this  method is called when all the other objects were
    # created and initialized (see TopLevelWidget)

    method init {} {
	global ::rtdb

	rtd::Rtd::init ; # init parent class

	option add *Entry.disabledForeground black 
	option add *Entry.disabledBackground #B2B2B2
	option add *Checkbutton.selectColor green

 	after 0 "$this more_init"
   }

   # load plugins and optionally a file when all widgets were created

    method more_init {} {
 	global ::rtdb ::env

	add_menubar_now
	raise $w_
	set name [winfo name .]

	# load plugins when all widgets were created
	update
	if {[info exists rtdb(LOADFILE)]} {
	    $itk_component(image) load_file $rtdb(LOADFILE)
	}
	if {[info exists rtdb($name\PLUGIN)]} {
	    set env([string toupper $name\_PLUGIN]) $rtdb($name\PLUGIN)
	    check_plugins $w_
	    call_plugin_procs $w_
	    unset rtdb($name\PLUGIN)
	}
	if {[rtdbCamera] && $rtdb(ATTACH)} {
	    attach_camera
	    camera_ready -1
	}

	# the last action is to enable the CCS interface 
	if {$rtdb(CCSINIT)} {
	    seq_ccsAsyncInput open
	}

	# we are ready to receive commands
	# create a file similar to $HOME/.rtd-remote for rtdbStart
	catch {
	    write_file "$env(VLTDATA)/tmp/$rtdb(APPNAME)-$env(USER)" \
		    "[pid] $env(HOST) [[$image_ get_image] remote]"
	}
	set rtdb(STATE) READY
    }

    # pass these methods on to the image widget unchanged
    # (this just generates methods on the fly...)

    ::foreach i {add_preCmd add_postCmd add_notify remove_notify camera_ready docmd \
	    get_canvas set_flash add_control get_panel perftest get_zoom defwin statistics \
	    pick pick_dialog add_rapidPreCmd add_rapidPostCmd add_updateValues get_pan \
	    add_scaleCmd add_transCmd traceXY wait_next_image save_file} {
	method $i {args} [::format {return [eval "$itk_component(image) %s $args"]} $i]
    }
   
    # redefine the TopLevelWidget method add_menuitem: we only record
    # the arguments for later evaluation if this menu item has been configured

    body util::TopLevelWidget::add_menuitem {menu type label msg args} {
	global ::rtdb

	if {[info exists rtdb(MAKING_MENU)]} {
	    lappend rtdb(MENU_ITEMS) "{$menu} {$type} {$label} {$msg} {$args}"
	} else {
	    eval [concat [list $menu add $type -label $label] $args]
	    setup_menuitem $menu $label $msg $args
	}
    }

    # add to the menu item

    method make_menuitem {menu type label msg args} {
	global ::rtdb
	lappend rtdb(MENU_ITEMS) "$menu $type"
    }

    # make the RTD menu according to the definition in the config file

    method make_menu {key n label} {
	global ::rtdb

	set wdg $itk_component(menubar).$label.m
	if { ! [winfo exists $wdg] } { return }
	set menu_$label\_ $wdg

	# remove the widgets for this menu - they will be added later when needed
	$wdg delete 0 50

	set list [rtdbSetList $n]
	set indices [rtdbIdx $list $key]
	foreach idx $indices {
	    if {$idx == 0} {
		$wdg add separator
		continue
	    }
	    if {"$label" == "Toolbox..." && $idx == 2} {
		[$image_ component draw] add_menuitems $wdg
		continue
	    }
	    set cmd [lindex $rtdb(MENU_ITEMS) [expr {$idx - 1}]]
	    if {"$cmd" == {}} { continue }
	    set i 0
	    foreach el "menu type label msg args" {
		set $el [lindex $cmd $i]
		incr i
	    }
	    eval [concat [list $menu add $type -label $label] $args]
	    setup_menuitem $menu $label $msg $args
	}
	unset rtdb(MENU_ITEMS)
   }

    # return name of the rtd image widget

    public method get_image {} {
	return $w_.image
    }

    # the menu bar is the very last thing to do...

    protected method add_menubar {} {
 	TopLevelWidget::add_menubar
        pack $itk_component(menubar) \
            -side top -fill x -ipady 2
   }

    method add_menubar_now {} {
	global ::rtdb

	# add menus and menubuttons
	set list [rtdbSetList 4]
	set indices [rtdbIdx $list MENU]
	foreach idx $indices {
	    if {$idx == 1} {
		::rtd::Rtd::add_file_menu
		make_menu MENUITM1 10 file
	    }
	    if {$idx == 2} {
		::rtd::Rtd::add_view_menu
		make_menu MENUITM2 10 view
	    }
	    if {$idx == 3} {
		::rtd::Rtd::add_graphics_menu
		catch {eval destroy [winfo children $itk_component(menubar).graphics.m]}
		make_menu MENUITM3 10 graphics
	    }
	    if {$idx == 4} {
		::rtd::Rtd::add_realtime_menu
		make_menu MENUITM4 10 real-time
	    }
	}

	# add user defined menus
	if {[info exists rtdb(ADD_MENUS)]} {
	    set onval 0
	    foreach mnu $rtdb(ADD_MENUS) {
		set menu [string tolower [lindex $mnu 1]]
		if {"$menu" == ""} { continue }
		set m $itk_component(menubar).$menu
		if {! [winfo exists $m]} {
		    add_menubutton [lindex $mnu 1]
		}
		set m $m.m
		set type [lindex $mnu 2]
		if {"$type" == "separator"} {
		    $m add separator
		    continue
		}
		set label [lindex $mnu 3]
		set cmd   "-command {[lindex $mnu 4]}"
		if {"$type" == "checkbutton"} {
		    set cmd "$cmd -onvalue [incr onval] -offvalue 0 -variable $m.var"
		}
		set msg   [lindex $mnu 5]
		rtdbFeedback "Adding menu item '$label' \t to \t $m"
		eval [concat [list $m add $type -label $label] $cmd]
		setup_menuitem $m $label $msg $cmd
	    }
	}
	unset rtdb(MAKING_MENU)
	add_help_menu
    }

    # create the rtd image widget

    protected method make_rtdimage {} {
        set image_ $w_.image
	# pass on the default options
	foreach opt "zoom_width zoom_height pan_width pan_height" {
	    config -$opt 141
	}
	foreach el "file usexshm usexsync verbose \
		min_colors max_colors drag_scroll scrollbars subsample \
		use_zoom_view with_zoom_window \
		dozoom with_pan_window zoom_factor zoom_width zoom_height \
		pan_width pan_height colorramp_height color_scale \
		default_cmap default_itt with_colorramp port debug \
		float_panel panel_layout with_warp min_scale max_scale \
		pickobjectorient xscale yscale" {
	    append opts "-$el {$itk_option(-$el)} "
	}
	if { ! [catch {cget -sampmethod}] } {
	    append opts "-sampmethod {$itk_option(-sampmethod)} "
	}
        # RtdImageCtrl(n) widget containing image and control panel.
        itk_component add image {
            eval ::rtdbCtrl $image_ $opts \
		    -shorthelpwin        {$this} \
		    -rapid_frame_command {[code $this rapid_frame_command]} \
		    -feedback            {[code $this feedback]} \
		    -zoom_view_propagate 0 \
		    -shm_header          1 \
		    -shm_data            1
        }
    }

    # -- options --

    # verbose mode flag
    itk_option define -verbose verbose Verbose {0} {
	global ::rtdb
	set rtdb(VERBOSE) [cget -verbose]
    }

    # Rtd's window geometry
    itk_option define -rtd_geometry rtd_geometry Rtd_geometry {} {
	if {"[cget -rtd_geometry]" == ""} {
	    lassign [wm maxsize .] nx ny
	    wm geometry $w_ $nx\x$ny+0+0
	}
    }

    # -- protected vars -- 
}

# ___oOo___
