#***************************************************************************
#
# E.S.O. - VLT project
#
# "@(#) $Id: seqCon.tcl,v 2.97 2005/09/14 09:47:31 vltsccm Exp $" 
# 
# who      when       what
# -------- ---------- ------------------
# eallaert 07/07/1997 created for cmm
# eallaert 2000-06-08 check for HOSTNAME instead of HOST
# eallaert 2001-05-07 upgraded to tkcon 2.1 - use rcfile
# eallaert 2002-05-14 removed Itcl, Itk, BLT from static package list
# eallaert 2002-11-12 put code handling icon for console and seqWish windows
# eallaert 2005-07-22 read also own option file (XseqWish.console)
#
# seqCon.tcl --

#***************************************************************************
#   NAME 
#   seqCon - start up tkcon within the seqWish interpreter 
#  
#   SYNOPSIS 
#   source seqCon.tcl
#
#   DESCRIPTION
#   seqCon is a script which will invoke tkcon, and "customize" it to the
#   seqWish style.
#   The script will however return without action if the seqWish runstring
#   includes -noTkCon, or if this is a non-interactive shell.
#   
#   ENVIRONMENT 
#   The argv array is checked for the presence of the switch "-noTkCon",
#   in which case tkcon will not start up. 
# 
#   RETURN VALUES
#   
#   EXAMPLES 
#
#   SEE ALSO 
#   "Tcl and the Tk toolkit", John K. Ousterhout, ISBN 0-201-6337-X
#   tkcon documentation (http://www.cs.uoregon.edu/research/tcl/script/tkcon/)
#------------------------------------------------------------------------ 
#
# Use tkcon for console, but only if Tk is available and interactive.
# If the -noTkCon option was passed as a runstring argument, return immediately.
# (don't use the var "name": it will be unset by tkcon when starting up!)
if {[lsearch -exact $argv -noTkCon] >= 0 || !$tcl_interactive} {
    return
}
###wm withdraw .

# Avoid the tcl_unknown procs, as this apparently means that also
# tclPkgUnknown will execute. As at this point the auto_path var does
# not yet contain all entries, this would risk to yield an incomplete
# set of packages (and all their versions under different dirs).
# Instead, source directly the files needed; it is also quicker.
######lappend auto_path $seq_library;	# needed to find the procs seq_findFile & tkcon
source [file join $seq_library seqLib.tcl]

# Now set the proper X-resources.
# Reset MOTIF style, otherwise ESO conventions do not work
set tk_strictMotif 0
    
# Load resources from a single place, in the usual VLT-style search order.
# These are the resources that will be applied to tkCon itself.
if {![catch {seq_findFile app-defaults/XseqWish} rc]} {
    option readfile $rc widgetDefault
}
if {![catch {seq_findFile app-defaults/XseqWish.private} rc]} {
    option readfile $rc userDefault
}
if {![catch {seq_findFile app-defaults/XseqWish.console} rc]} {
    option readfile $rc userDefault
}
unset rc

# Indicate which resource file to use (append this to argv)
if {[file exists ~/.seqConRc]} {
    lappend argv -rcfile ~/.seqConRc
} else {
    lappend argv -rcfile $seq_library/../../config/.seqConRc
}
incr argc 2

# The next avoids the output of the standard Tk "%" prompt on the
# terminal from which seqWish is started (which is confusing).
set tcl_prompt1 {}

# Before tkcon.tcl is sourced in (via the "unknown" command, called when the
# first tkcon command is given), we have to set the defaults such that they
# are OK with us (without having to edit the tkcon.tcl source code). In
# particular, we do not want the console-window to appear before its icon
# is defined (as otherwise we'd have to do a "tkcon hide; update; tkcon show",
# which would make the window flash shortly on the screen, in order for
# the window manager to take the colour icon into account).
namespace eval ::tkcon {
    variable PRIV
    variable OPT
    set PRIV(showOnStartup) 0
    # Need to have the static package list in a specific order !!
    # Only the static packages (as declared via a Tcl_StaticPackage call
    # in tkXAppInit.c) initialized *before* tkcon is loaded,
    # must be included here. The packages loaded by tkXAppInit.c after
    # starting tkcon, and the ones loaded by the seqInit script (Itcl, 
    # Itk and BLT) should not be included here.
    set OPT(autoload) [list Tk]; # list of packages to load
    set OPT(exec) slave;	 #  Sets the named slave that tkcon operates in
}

# Now we can source the tkcon code.
source [file join $seq_library tkcon.tcl]
# Although the sourcing of the tkcon code ends with the execution of
# an init procedure, this latter proc has been modified (from version 2.4
# onwards) to skip some steps if tkcon is not the main script - like in
# our case. So we need to go through this phase explicitly.
eval tkcon::Init $argv

# Make sure that the small slave window is withdrawn from the very beginning;
# this "tkcon" command will result in source-ing in tkcon.tcl
tkcon slave slave wm withdraw .

# adjust the console title a little
set procName [file tail $argv0]
if {[info exists env(HOSTNAME)]} {
    set host $env(HOSTNAME)
} elseif {[info exists env(HOST)]} {
    set host $env(HOST)
} else {
    set host [info hostname]
}
tkcon title "${procName}@$host - Main console (TkCon)"

# tkcon seems to add an empty dir to auto_path at some point during its init-phase
while {[set idx [lsearch -exact $auto_path {}]] >= 0} {
    set auto_path [lreplace $auto_path $idx $idx]
} 

# Let's deal with the console window iconification.
if {![catch {seq_findFile bitmaps/${procName}Icon.gif} pm]} {
    image create photo seqConIcon -file $pm
    toplevel .seqConIcon
    pack [label .seqConIcon.l -image seqConIcon]
    wm iconwindow . .seqConIcon
    # WM needs "update" with window unmapped; this won't flash window,
    # due to PRIV(showOnStartup) setting (see higher).
    update
} elseif {![catch {seq_findFile bitmaps/${procName}Icon.xbm} bm]} {
    wm iconbitmap . @$bm
}
catch {wm iconname . ${procName}-console@$host}
catch {unset pm bm}

# To be allowed to run in background, we need to close stdin (x-term)
# As from Tcl 8.0, using channelIds, it needs to be closed in ALL interpreters
# before the proper clean-up actions can take place!
tkcon master close stdin
tkcon slave slave {close stdin}

# Next could be given earlier, e.g. just after the console window icon
# set-up, to make it clear that "something" is going on (satisfying the
# needs of the computer illiterates who need visual confirmation that
# "the system is alive" ;-).
# However, I prefer to pop up the window when it is ready, as otherwise
# you get an impression of "sluggishness". It will anyway not be very
# alive yet, as at this point there is still the seqInit.tcl to be
# sourced in the slave interpreter - unless there is an "update".
tkcon show
update
