#*************************************************************************
# E.S.O. - VLT project
#
# "@(#) $Id: uifPlot.tcl,v 3.49 2005/09/07 11:45:58 vltsccm Exp $"
#
# who       when        what
# --------  --------    ----------------------------------------------
# cerpello  17/09/96    created
# mcomin    03/12/96    Add plot configuration panel
#                       Add public method draw 
#           06/03/97    Porting to TK 4.2
# mathar    10/08/99    added selection of non-zero colX, non-one colY in input file
# mathar    10/08/99    added interactive selection of minX,maxX,minY,maxY
# mathar    10/08/99    skip lines starting with # in input file
# mathar    10/08/99    optional autoscaling of x and y axis
# mcomin    11/04/2000  Integration with official uif library.
#                       Add title, xlabe and y label options.
#                       Imptoved GUI for interactive plot configuration.
#
#************************************************************************
#   NAME
#     UifPlot - Widget to display curves
#   SYNOPSIS
#     UifPlot <widget-name> ?widget options?
#   DESCRIPTION
#     UifPlot is a widget that can be configured to display a static plot. 
#
#     The widget is formed by:
#     - A 'blt' graph widget : used to display the static plotting.
#
#   COMPONENTS
#
#   METHODS
#     Support the following methods:
#
#     plot
#        Makes the static plotting taking the data from a file.
#        The file is the one specified by the '-fileName' option 
#
#     draw <X> <Y>
#        Displays a curve described by a list of X and Y elements.
#
#     reset
#        Resets the plotted data points.
#
#     plotConfig
#        Displays a configuration panel where the user can specify :
#
#        - file containing data to be plotted.  Each row of the file 
#          represents a point and is made of a (X,Y) pair of reals.
#        - colour of the line to be drawn;
#        - symbol which associated to each plotted data point, such as :
#                 square, plus, circle and cross.
#
#     setPlotOptions <smoothOption> <symbol> <color> <file> <colX> <colY> 
#                    <minX> <maxX> <minY> <maxY>
#        Implements the programmatic interface of the 'plotConfig' method.
#
#   PUBLIC VARIABLES
#
#   WIDGET OPTIONS
#     Support the following options: 
#
#     background    <color>  Color of the border, if the class is visible,
#                            otherwise color of the background. 
#     colX          <val>    Column of X-data in input file.    
#     colY          <val>    Column of Y-data in input file. 
#     graphHeight   <val>    Graph height.
#     graphWidth    <val>    Graph width.
#     logScale      <on/off> Sets on/off logarithmic scale.    
#     minX          <val>    Minimum limit of the x-axis.   
#     maxX          <val>    Maximum limit of the x-axis.    
#     minY          <val>    Minimum limit of the y-axis.    
#     maxY          <val>    Maximum limit of the y-axis.
#     smooth        <val>    This option determines how the points are 
#                            connected. valid values are :
#                            - linear : points are connected by segments 
#                            - step : points are conncted by 'stairs-like'
#                              diagram.
#                            - natural and quadratic points are connected with
#                              a spline.
#     title         <string>
#     xlabel,ylabel <string> Add a label on X and Y axis.
#
#   CAUTIONS
#
#   EXAMPLES
#
#     Create an instance
#
#     UifPlot .x -minY 0 -maxY 500 -fileName ../test/plot1.dat \
#                   -graphWidth 200 -graphHeight 200
#
#     .x plot   ;# Plot the data contained in the file 
#
#     .x setPlotOptions circle green ../test/plot1.dat 0 1 Set new plot
#     .x plot                                           Plot the new data
#
#     Plot data specified in tcl/tk variables 
#
#     set xlist [list 10.0 20.0 30.0 40.0 50.0 60.0 ]
#     set ylist [list 34.0 45.0 56.0 27.0 50.0 12.0 ]
#     .x draw $xlist $ylist
#
#     Re-configure the widget                    
#       .x configure -minX -100 -maxX 150 -scale on
#
#    Other operations
#       set file [.x cget -fileName]     Get a widget's resource
#   SEE ALSO
#
#   BUGS     
#
#------------------------------------------------------------------------
#
itk::usual UifPlot { }

class UifPlot {
   inherit UifWidget
   
   constructor {args} {}

   itk_option define -background    background    Background    ""
   itk_option define -minX	    minX	  MinX          0
   itk_option define -maxX	    maxX 	  MaxX          100
   itk_option define -minY	    minY	  MinY	        -100
   itk_option define -maxY	    maxY	  MaxY          100
   itk_option define -graphWidth    graphWidth    GraphWidth    200
   itk_option define -graphHeight   graphHeight   GraphHeight   250
   itk_option define -fileName      fileName      FileName      ""
   itk_option define -logScale      logScale      LogScale      "off"
   itk_option define -smooth        smooth        Smooth        "linear"
   itk_option define -xlabel        xlabel        XLabel         ""
   itk_option define -ylabel        ylabel        YLabel         ""

   # added by mathar
   itk_option define -colX      colX      ColX      0
   itk_option define -colY      colY      ColY      1

   protected method xsort { x y }
   protected method makeConfigPanel  {w}

   public method draw  { x y }
   public method plot  {}
   public method reset {}
   public method plotConfig {}
   public method deleteTag  {tag} 
   public method createLine     {lineId xdata ydata color}
   public method setPlotOptions {smooth symbol color fileName colX colY minX maxX minY maxY}

   # This is used internally
   public method setConfiguration {w}
   public method fileBrowse {w}
   # added by mathar
   public method autoScale {w}

   protected variable X             ""
   protected variable Y             ""
   protected variable filePath      ""
   protected variable Colour        red
   protected variable Symbol        none 
   protected variable showSymbol    1
   protected variable numPlots      0
   # added by mathar
   protected variable colX    0
   protected variable colY    1
}

#---------------------------------------------------------------------------
# Constructor
#---------------------------------------------------------------------------

body UifPlot::constructor {args} {
                 
   itk_component add f {
      frame $itk_interior.f 
   }
   
   itk_component add plot {
     ::blt::graph  $itk_component(f).plot  -borderwidth 2 -relief groove
   } {
     keep -background -title
   }   
       
   pack $itk_component(f)
   pack $itk_component(plot) -side top -fill both

   eval itk_initialize $args
   
   $itk_component(plot) legend configure -hide yes  ;# No legend
} 

#---------------------------------------------------------------------------
#  OPTION : label for X and Y axis
#---------------------------------------------------------------------------
configbody UifPlot::xlabel {
    $itk_component(plot) xaxis configure  -title $itk_option(-xlabel)
}

configbody UifPlot::ylabel {
    $itk_component(plot) yaxis configure  -title $itk_option(-ylabel)
}

#---------------------------------------------------------------------------
#  OPTION : -minX
#---------------------------------------------------------------------------
configbody UifPlot::minX {
    $itk_component(plot) xaxis configure -min $itk_option(-minX)
}

#---------------------------------------------------------------------------
#  OPTION : -maxX
#---------------------------------------------------------------------------
configbody UifPlot::maxX {
    $itk_component(plot) xaxis configure -max $itk_option(-maxX)
}

#---------------------------------------------------------------------------
#  OPTION : -minY
#---------------------------------------------------------------------------
configbody UifPlot::minY {
    $itk_component(plot) yaxis configure -min $itk_option(-minY)
}

#---------------------------------------------------------------------------
#  OPTION : -maxY
#---------------------------------------------------------------------------
configbody UifPlot::maxY {
    $itk_component(plot) yaxis configure -max $itk_option(-maxY)
}

#---------------------------------------------------------------------------
#  OPTION : -graphWidth
#---------------------------------------------------------------------------
configbody UifPlot::graphWidth {
   if {$itk_option(-graphWidth) >= 100 && $itk_option(-graphWidth) <= 1200} {
      $itk_component(plot) configure -width $itk_option(-graphWidth)
   }
}

#---------------------------------------------------------------------------
#  OPTION : -graphHeight
#---------------------------------------------------------------------------
configbody UifPlot::graphHeight {
  if {$itk_option(-graphHeight) >= 100 && $itk_option(-graphHeight) <= 1100} {
     $itk_component(plot) configure -height $itk_option(-graphHeight)
  }
}

#---------------------------------------------------------------------------
#  OPTION : -logScale
#---------------------------------------------------------------------------
configbody UifPlot::logScale {
   $itk_component(plot) yaxis configure -logscale $itk_option(-logScale)
}

#---------------------------------------------------------------------------
#  OPTION : -fileName
#---------------------------------------------------------------------------
configbody UifPlot::fileName {

   if {$itk_option(-fileName) != ""} {

      if {[file exist $itk_option(-fileName)]} {
         set filePath $itk_option(-fileName)
      } else {
         set filePath [findVLTFile $itk_option(-fileName) n]
      }

   } else {
      set filePath ""
   }
}

#---------------------------------------------------------------------------
#  OPTION : -colX -colY
#---------------------------------------------------------------------------
configbody UifPlot::colX {
  if {$itk_option(-colX) >= 0} {
     set colX $itk_option(-colX)
  }
}
configbody UifPlot::colY {
  if {$itk_option(-colY) >= 0} {
     set colY $itk_option(-colY)
  }
}

#-----------------------------------------------------------------
# METHOD : xsort
#-----------------------------------------------------------------

body UifPlot::xsort {vectX vectY} {

   set sortedX [lsort -real $vectX] ; set X $sortedX
   set numPoints [llength $X]

   for {set i 0} {$i <= [expr {$numPoints -1}]} {incr i +1} {
       set sortedY([lsearch $sortedX [lindex $vectX $i]]) [lindex $vectY $i]
       set sortedX [lreplace $sortedX [lsearch $sortedX [lindex $vectX $i]]\
                   [lsearch $sortedX [lindex $vectX $i]] ""]
   }
   
   for {set i 0} {$i <= [expr {$numPoints -1}]} {incr i +1} {
       set Y [linsert $Y $i $sortedY($i)]
   }

   return $numPoints
}

#-----------------------------------------------------------------
# METHOD : plot
#-----------------------------------------------------------------

body UifPlot::plot {} {

   if {[cequal $filePath ""]} { return }

   set id [open $filePath r]

   while {[gets $id point] != -1} {
       # skip lines starting with '#' in ASCII file
       if { [string index [lindex $point 0] 0] != {#} } {
           lappend vectX [lindex $point $colX]
           lappend vectY [lindex $point $colY]
       }
   } 

   close $id
   draw $vectX $vectY
}

#-----------------------------------------------------------------
# METHOD : autoScale
# added by mathar
#-----------------------------------------------------------------

body UifPlot::autoScale {w} {
   global ::ranges

   if {[cequal $filePath ""]} { return }

   set id [open $filePath r]

   # Configure columns X and Y index
   configure -colX [$w.r3.colX get] -colY [$w.r4.colY get]

   set locminX 100
   set locmaxX 0

   while {[gets $id point] != -1} {
       if { [string index [lindex $point 0] 0] != {#} } {
           set newx [lindex $point $colX]
           set newy [lindex $point $colY]
	   if { $locminX > $locmaxX } {
              set locminX $newx
              set locmaxX $newx
              set locminY $newy
              set locmaxY $newy
	   } else {
	      if { $newx > $locmaxX }  { set locmaxX $newx }
	      if { $newx < $locminX }  { set locminX $newx }
	      if { $newy > $locmaxY }  { set locmaxY $newy }
	      if { $newy < $locminY }  { set locminY $newy }
           }
       }
   } 
   close $id
   if { $locminX == $locmaxX } {
       set locminX [ expr {$locmaxX - .5} ]
	set locmaxX [ expr {$locmaxX + .5} ]
   }
   if { $locminY == $locmaxY } {
	set locminY [ expr {$locmaxY - .5} ]
	set locmaxY [ expr {$locmaxY + .5} ]
   }

   configure -minX $locminX -maxX $locmaxX -minY $locminY -maxY $locmaxY

   set ranges(0) $locminX
   set ranges(1) $locmaxX
   set ranges(2) $locminY
   set ranges(3) $locmaxY
}

#-----------------------------------------------------------------
# METHOD : draw
#-----------------------------------------------------------------

body UifPlot::draw { X Y } {
   if {[cequal $Colour ""]} {set Colour black}

   $itk_component(plot) element create plot${numPlots} -symbol $Symbol\
                -xdata $X -ydata $Y -color $Colour \
                 -smooth $itk_option(-smooth) -pixel  0.15c
   incr numPlots
}


#----------------------------------------------------------------------------
# METHOD : reset
#----------------------------------------------------------------------------

body UifPlot::reset {} {
   foreach elem [ $itk_component(plot) element names] {
      $itk_component(plot) element delete $elem
   }
}

#----------------------------------------------------------------------------
# METHOD : setPlotOptions
#----------------------------------------------------------------------------

# added last two options mathar
body UifPlot::setPlotOptions {smooth symbol color file colX colY minX maxX minY maxY} {

   if {[set Symbol $symbol] == "none"} { 
      set showSymbol 0
   } else {
      set showSymbol 1
   }

   set Colour  $color

   # added these by mathar
   configure -fileName $file -colX $colX -colY $colY -minX $minX -maxX $maxX \
             -minY $minY -maxY $maxY -smooth $smooth

   set ranges(0) $minX
   set ranges(1) $maxX
   set ranges(2) $minY
   set ranges(3) $maxY
}

#----------------------------------------------------------------------------
# METHOD : fileBrowse
#----------------------------------------------------------------------------

body UifPlot::fileBrowse {w} {
   UifFileBrowser $this.fileBrowse
   lassign [$this.fileBrowse getFile] fname info

   if {![cequal $fname ""] } {
      $w delete 0 end
      $w insert 0 $fname ;# Extract just the file name
   }
}

#----------------------------------------------------------------------------
# METHOD : plotConfig
#----------------------------------------------------------------------------

body UifPlot::plotConfig {} {
   set c [namespace tail $this.plotConfig]

   if {![winfo exist $c]} {
     toplevel $c
     makeConfigPanel $c
     tkwait window $c ; # Wait until the user does something ...
   }
}

#---------------------------------------------------------------
# METHOD : createLine
#---------------------------------------------------------------

body UifPlot::createLine {lineId xdata ydata color} {

  $itk_component(plot) element create $lineId -symbol $Symbol\
                -xdata $xdata -ydata $ydata -color $color \
                -smooth $itk_option(-smooth) -pixel 0.15c
}

#---------------------------------------------------------------
# METHOD : deleteTag
#---------------------------------------------------------------

body UifPlot::deleteTag {tagList} {
    catch {eval $itk_component(plot) element delete $tagList}
}

#################################################################
# Methods used by plotConfig to create the configuration panel
#################################################################

body UifPlot::setConfiguration {w} {
   global ::plotSymbol ::ranges ::smoothOption

   setPlotOptions $smoothOption $plotSymbol [$w.r1.color get] [$w.r5.file get] \
                  [$w.r3.colX get] [$w.r4.colY get] $ranges(0) $ranges(1) \
                  $ranges(2) $ranges(3)
   $this configure -title  "[$w.r2.title get]"  -xlabel  "[$w.r3.xlabel get]" \
                   -ylabel "[$w.r4.ylabel get]"
}

body UifPlot::makeConfigPanel {w} {
   global ::plotSymbol ::ranges ::smoothOption

    if {[winfo exists $w.top]} { raise $w; return }

   wm resizable $w 0 0
   set top  [frame $w.top -borderwidth 2 -relief groove]
   set bot  [frame $w.bot]

   pack $top $bot -pady 5 -fill both -padx 3

   for {set i 1} {$i < 9} {incr i } {
       set row${i} [frame $top.r${i}]
       pack [set row${i}] -fill x -pady 3 
   }

   set cour14 "-Adobe-courier-medium-r-normal--14*"
   set helv14 "-Adobe-helvetica-bold-r-normal--14*"

   # ------------- row 1 ------------------
   UifEntry $row1.color -label "Color" -labelWidth 10 -width 10 -labelFont $cour14

   label $row1.labSymbol -text "Symbol" -font $cour14

   UifOptionButton $row1.valSymbol -length 5 -variable plotSymbol\
	   -itemList {none cross scross square plus splus circle triangle diamond}
 
   label $row1.labSmooth -text "Smooth" -font $cour14

   UifOptionButton $row1.smooth -length 8 -variable smoothOption \
                   -itemList {linear step natural quadratic}

   # ------------- row 2,3,4  ------------------

   UifEntry $row2.title  -label "Title"   -labelWidth 10 -width 40 -labelFont $cour14

   UifEntry $row3.xlabel -label "X Label" -labelWidth 10 -width 30 -labelFont $cour14
   UifEntry $row3.colX   -label "Column X" -width 3 -labelFont $cour14

   UifEntry $row4.ylabel -label "Y Label" -labelWidth 10 -width 30 -labelFont $cour14
   UifEntry $row4.colY   -label "Column Y" -width 3 -labelFont $cour14

   # ------------- row 5  ------------------
   UifEntry $row5.file -label "File " -labelWidth 10 -width 40 -labelFont $cour14

   button $row5.fileBrowse -text "Browse ..." -width 8 \
          -command "$this fileBrowse $row5.file"

   # added entries by mathar

   UifEntry $row7.valXMin -label "X Range " -width 10 -labelFont $cour14 -variable ranges(0)
   UifEntry $row7.valXMax -width 10 -labelFont $cour14 -variable ranges(1)

   UifEntry $row8.valYMin -label "Y Range " -width 10 -labelFont $cour14 -variable ranges(2)
   UifEntry $row8.valYMax -width 10 -labelFont $cour14 -variable ranges(3)

   button $row7.autosca -text "autoscale" -width 8 -command "$this autoScale $top"
  
   pack $row1.color $row1.labSymbol $row1.valSymbol $row1.labSmooth $row1.smooth -side left -padx 5

   pack $row2.title  -side left -padx 5 
   pack $row3.xlabel $row3.colX -side left -padx 5 
   pack $row4.ylabel $row4.colY -side left -padx 5 
   pack $row5.file $row5.fileBrowse -side left -padx 5
 
   # added by mathar

   pack $row7.valXMin $row7.valXMax $row7.autosca -side left -padx 4
   pack $row8.valYMin $row8.valYMax -side left -padx 4

   #
   # Add buttons 
   #
   button $bot.ok  -text "OK"  -width 8 -font $helv14 -command "$this setConfiguration $top"
   button $bot.apply  -text "Configure"  -width 12 -font $helv14 \
          -command "$this setConfiguration $top"

   button $bot.draw  -text "Configure and Draw" -width 18 -font $helv14 \
          -command "$this setConfiguration $top; $this plot"

   button $bot.cancel -text "Cancel" -width 8 -font $helv14 -command "destroy $w"

   pack $bot.ok $bot.apply $bot.draw $bot.cancel -side left -expand 1
         
   #
   # Initialize with current values
   #

   $row1.color  delete 0 end ; $row1.color  insert 0 $Colour
   $row5.file   delete 0 end ; $row5.file   insert 0 $itk_option(-fileName)
   $row2.title  delete 0 end ; $row2.title  insert 0 $itk_option(-title)
   $row3.xlabel delete 0 end ; $row3.xlabel insert 0 $itk_option(-xlabel)
   $row4.ylabel delete 0 end ; $row4.ylabel insert 0 $itk_option(-ylabel)

   set plotSymbol   $Symbol
   set smoothOption $itk_option(-smooth)

   set ranges(0) $itk_option(-minX)
   set ranges(1) $itk_option(-maxX)
   set ranges(2) $itk_option(-minY)
   set ranges(3) $itk_option(-maxY)

   $row3.colX delete 0 end ; $row3.colX insert 0 $itk_option(-colX)
   $row4.colY delete 0 end ; $row4.colY insert 0 $itk_option(-colY)
}

#
# ___oOo___

