#************************************************************************
# E.S.O. - VLT project
#
# "@(#) $Id: uifStripChart.tcl,v 3.49 2005/09/07 11:45:59 vltsccm Exp $"
#
# who       when        what
# --------  ----------  ----------------------------------------------
# cerpello  17/09/96    created
# mcomin    03/12/96    Update man page
#           28/05/98    Optimization
#           18/08/98    Add fix suggested by P.Santin (AOT)
# eallaert  1999-02-15  Ported to Tcl/Tk 8.0 and itcl 3.0

#************************************************************************
#   NAME
#     UifStripChart - Strip Chart
#   SYNOPSIS
#     UifStripChart <widget-name> ?widget options?
#   DESCRIPTION
#     UifStripChart is a widget that can be configured to display a  
#     strip chart.
#
#     The widget is formed by:
#
#     - A canvas where to plot the values versus time
#     - Two arrows that signal when the value goes beyond the lower/upper 
#       limit
#
#     Data can be read either from a variable or a data base point.
#     The plotting can run in polling or event driven.
#
#     Before starting a plotting, the option "-datasource" must be configured, 
#     otherwise an error occours.
#
#     'Polling' Based Updating
#
#     The procedure specified in the option "-updateProc" is recursively called
#     until methods 'stopPlot' or 'pausePlot' are invoked.
#
#     There is a default procedure called 'updateChartPoll' which is able to 
#     update the strip-chart from a value stored in a local tcl variable or
#     from the database.  
#     The procedure checks if the syntax of the 'datasource' option follows
#     the RTAP conventions
#     option and 
#
#     The user can specify it's own procedure which must follow the following
#     template :
#
#         proc userChartPoll { w type var rate} {
#             if {[winfo exists $w] == 0} { return }
# 
#             if {$type == ""} {  
#                set newValue [uplevel #0 set $var]
#             } else {  
#                set newValue <Read from Database>
#             }
#     
#             ... here the user can put some additional code
#             to process the value read from the datasource
#
#             $w drawValue $newValue
#             set plotState [$w getPlotState] 
#             if {$plotState != "PAUSE" && $plotState != "STOP"} {   
#                after $rate userChartPoll $w $type $var $rate  
#             }   
#         }
# 
#     'Event' Based Updating
#
#     The procedure specified in the option "-updateProc" is called
#     whenever an event from the database is generated : there is a default 
#     procedure called 'updateChartEvent'.
#
#     This mode works only with database attributes.
#
#     The user can specify it's own procedure which must follow the following
#     template :
#
#         proc updateChartEvent {w attr type data} {
#
#             if {![winfo exists $w]} { return }
#             if {$type != "SCALAR"}  { return }
#   
#             ... here the user can put some additional code
#             to process the value read from the datasource
#
#            set plotState [$w getPlotState] 
#            if {$plotState == "PLOT" || $plotState == "IDLE"} {
#                $w drawValue [lindex $data 2]
#            }
#         }
#   COMPONENTS
#
#   METHODS
#     Support the following methods:
#
#     plot
#        Starts the dynamic plotting.
#
#     stopPlot
#        Stops the dynamic plotting and resets plotted data points.
#
#     pausePlot
#        Stops temporarily the dynamic plotting.
#
#     continuePlot
#        Continues the dynamic plotting after a pause.
#
#     stripChartConfig
#	 Allows interactive configuration of the widget.  
#
#   PUBLIC VARIABLES
#
#   WIDGET OPTIONS
#     Support the following options: 
#           
#     colorUpperLim  <color>     Color of the upper limit.
#     colorLowerLim  <color>     Color of the lower limit.
#     valueUpperLim  <val>       Value of the upper limit.
#     valueLowerLim  <val>       Value of the lower limit.
#     datasource     <string>    Variable or database point containing the
#                                current value to be plotted.
#     dataType       <DB>/<TCL>  Defines the type of variable.
#                                - DB the value is read from the database
#                                - TCL the value is read from a global tcl
#                                  variable
#     samplingTime   <msec>      Sampling time of the plot.
#     startTime      <hh:mm:ss>  Start time of the plot.  
#     updateMode     <mode>	 Mode of plotting : "poll" or "event"
#
#     updateProc     <proc>      Name of the procedure used to generate the
#                                in the plotting for polling or event.
#                                Default values are :
#                                - updateChartPoll  when used in polling
#                                - updateChartEvent when used with events
#
#     And inherits the following options from uifPlot :
#
#     background     <color>       Color of the border, if the class is 
#                                  visible otherwise color of the background.
#     minX           <val>         Minimum limit of the x-axis.
#     maxX           <val>         Maximum limit of the x-axis.
#     minY           <val>         Minimum limit of the y-axis.
#     maxY           <val>         Maximum limit of the y-axis. 
#     title          <string>    
#     graphWidth     <val>         Graph width.
#     graphHeight    <val>         Graph height.
#
#   CAUTIONS
#     If the plotting is running event driven, the generated event by 
#     datasource  must be syncrhonized, otherwise an error occours in 
#     the layout of plotted data.
#   EXAMPLES
#
#     Create an instance
#
#     UifStripChart .x  -datasource ":uves:ins.HwState"  -dataType DB
#                       -minY 0 -maxY 1000 -graphWidth 200 -graphHeight 200\
#                       -valueUpperLim 950 -valueLowerLim 100\
#                       -colorUpperLim red -colorLowerLim blue
#
#    Re-configure the widget
#       .w configure -colorUpperLim Blue -minY -30
#
#    Other operations                           
#       set dbAttribute [.x cget -datasource]    Get a widget's resource
#   SEE ALSO
#     UifPlot
#   BUGS     
#
#------------------------------------------------------------------------
#
class UifStripChart {
   inherit UifWidget
   
   constructor {args} {}
   destructor {}
   
   itk_option define -colorUpperLim colorUpperLim ColorUpperLim "red"
   itk_option define -colorLowerLim colorLowerLim ColorLowerLim "blue"
   itk_option define -valueUpperLim valueUpperLim ValueUpperLim 1
   itk_option define -valueLowerLim valueLowerLim ValueLowerLim -1
   itk_option define -datasource    datasource    Datasource    "dummy"
   itk_option define -samplingTime  samplingTime  SamplingTime  "1000"
   itk_option define -startTime     startTime     StartTime     00:00:00
   itk_option define -updateMode    updateMode    UpdateMode    "poll"
   itk_option define -updateProc    updateProc    UpdateProc  "updateChartPoll"
   itk_option define -dataType      dataType      DataType     "DB"        
   
   public method plot          {}
   public method pausePlot     {}
   public method continuePlot  {}
   public method stopPlot      {}
   public method stripChartConfig  {}
   public method getPlotState  {} { return $plotState }
   public method drawValue     {value}

   protected method getStartTime {}
   protected method detachEvent  {}
   protected method getMinMax    {}
  
   protected variable plotState	    "IDLE"    

   protected variable valX	    0
   protected variable sampleTime    0  
   protected variable sampleTimeMin 0
   protected variable oldValue      0
   protected variable wthis         ""
   protected variable eventId       "" 
   protected variable range	    0
   protected variable counter       0
   protected variable firstCount    0
   protected variable initialized   0
   protected variable paused        0   
}

#---------------------------------------------------------------------------
# Constructor
#---------------------------------------------------------------------------

body UifStripChart::constructor {args} {
    global tcl_version

    set wthis [namespace tail $this]

   itk_component add f1 {
      frame $itk_interior.f1 -borderwidth 2 -relief groove
   }
         
   itk_component add plot {
      UifPlot $itk_component(f1).plot -xlabel "Time (minutes)"
   } {     
     keep -background -minX -maxX -title
     keep -minY -maxY -graphWidth -graphHeight
   }
      
   itk_component add f2 {
      frame $itk_component(f1).f2  
   }
          
   itk_component add initTime { 
      label $itk_component(f2).initTime  -text "Start Time     00:00:00" \
            -font "-Adobe-helvetica-bold-r-normal--12*" 
   } {
     keep -cursor
   }
   
   # Frame containing 'limit arrow'
   
   itk_component add f3 {
      frame $itk_component(f1).f3  
   } 
           
   itk_component add cvArrow {
      canvas $itk_component(f3).cvArrow -width 30 \
      			-height [expr $itk_option(-graphHeight) -30]
   } {
     keep -cursor
   }
 
   pack $itk_component(f1)
   pack $itk_component(f3) -side right  -fill y 
   pack $itk_component(f2) -side bottom -fill x
   pack $itk_component(plot) -side left  
   
   pack $itk_component(initTime) -side left -padx 20
   pack $itk_component(cvArrow)

   eval itk_initialize
   set initialized 1
   eval configure $args  ;# User's configuration
} 

#---------------------------------------------------------------------------
# Destructor
#---------------------------------------------------------------------------

body UifStripChart::destructor {} {
   detachEvent
}

#--------------------------------------------------------------------------
#  OPTION : -valueUpperLim
#--------------------------------------------------------------------------
configbody UifStripChart::valueUpperLim {

   if {$initialized} {
      $itk_component(plot) deleteTag upperLimit
 
      $itk_component(plot) createLine upperLimit [getMinMax] \
           "$itk_option(-valueUpperLim) $itk_option(-valueUpperLim)"\
           $itk_option(-colorUpperLim)
   }
}

#---------------------------------------------------------------------------
#  OPTION : -valueLowerLim
#--------------------------------------------------------------------------
configbody UifStripChart::valueLowerLim {

   if {$initialized} {
       $itk_component(plot) deleteTag lowerLimit

       $itk_component(plot) createLine lowerLimit [getMinMax]\
            "$itk_option(-valueLowerLim) $itk_option(-valueLowerLim)"\
            $itk_option(-colorLowerLim)
   }
}

#---------------------------------------------------------------------------
#  OPTION : -colorUpperLim
#---------------------------------------------------------------------------
configbody UifStripChart::colorUpperLim {

   if {$initialized} {  
      $itk_component(plot) deleteTag upperLimit

      $itk_component(plot) createLine upperLimit [getMinMax]\
           "$itk_option(-valueUpperLim) $itk_option(-valueUpperLim)" \
           $itk_option(-colorUpperLim)
   }
}

#---------------------------------------------------------------------------
#  OPTION : -colorLowerLim
#---------------------------------------------------------------------------
configbody UifStripChart::colorLowerLim {

   if {$initialized} {
      $itk_component(plot) deleteTag lowerLimit

      $itk_component(plot) createLine lowerLimit [getMinMax]\
           "$itk_option(-valueLowerLim) $itk_option(-valueLowerLim)"\
           $itk_option(-colorLowerLim)
   }
}

#---------------------------------------------------------------------------
#  OPTION : -samplingTime
#---------------------------------------------------------------------------
configbody UifStripChart::samplingTime {
   
   if {$initialized && $sampleTime != $itk_option(-samplingTime)} {
     
     set oldTime $sampleTime
     if { [set sampleTime $itk_option(-samplingTime)] <= 0} {
       set sampleTime 1000
     }
     
     set sampleTimeMin [expr {$sampleTime /(1000*60.0)} ]
     set range [expr {[$itk_component(plot) cget -graphWidth] * $sampleTimeMin} ]

#     puts "DBG> UifStripChart::samplingTime range = $range"
#     puts "     maxX = $itk_option(-maxX) "

     if {[expr {$itk_option(-maxX) - $range} ] < 0} {        
         configure -minX 0 -maxX $range    
     } else {
         configure -minX $valX -maxX [expr {$itk_option(-minX) + $range} ]
     }        
   }
}

#---------------------------------------------------------------------------
#  METHOD : getStartTime
#---------------------------------------------------------------------------

body UifStripChart::getStartTime {} {
   
   if {[uplevel #0 info exists $itk_option(-startTime)] } {
   
      # Check if startTime is an internal variable
      set myTime [uplevel #0 set $itk_option(-startTime)]
      
   } elseif { [string first "." $itk_option(-startTime)] != -1} {
   
      # Check if startTime is a DB attribute
   
      if {[catch {seq_dbReadSymbolic $itk_option(-startTime) } myTime] } {
         seq_logData "$this : ERROR READING $itk_option(-startTime)"
         set myTime 0
      }
   } else {
      set myTime [clock format [clock seconds] -format %T]"
   }
   
   puts "getStartTime  $myTime"
   $itk_component(initTime) configure -text "Start Time    $myTime" 
}

#-----------------------------------------------------------------
# METHOD : plot
#-----------------------------------------------------------------

body UifStripChart::plot {} {

   set counter    0
   set firstCount 0
   set valX       0
   
   if {![cequal $plotState STOP] && ![cequal $plotState IDLE]} { return }

   getStartTime

   # Fix suggested by P.Santin
   set x1  [$itk_component(plot) cget -minX]
   set x2  [$itk_component(plot) cget -maxX]
   set range [expr {$x2 - $x1} ]

   if {[cequal $plotState STOP]} {
       $itk_component(plot) reset
       detachEvent

       configure -minX 0 -maxX $range \
                 -valueUpperLim $itk_option(-valueUpperLim) \
       		 -valueLowerLim $itk_option(-valueLowerLim)   
   }

   set plotState PLOT

   set var    $itk_option(-datasource)
   set myProc $itk_option(-updateProc)
   set type   $itk_option(-dataType)

   if {[cequal $sampleTime ""]} { set sampleTime 1000 } ;# in case of mistake

   if { [cequal $itk_option(-updateMode) poll] } { 
       eval $myProc $wthis $type $var $sampleTime
   } elseif { [cequal $itk_option(-updateMode) event] && [cequal $type DB]} {  
       if {[catch {seq_evtAttach $var w "$myProc $wthis %A %T %D"} eventId] } {
          seq_logData "$this : Cannot Attach Event to $var"
          set eventId ""
          return
       }
   }                           
}

#-----------------------------------------------------------------
# METHOD : drawValue
#-----------------------------------------------------------------

body UifStripChart::drawValue {value} {
                 
   set i2 [expr {$valX + $sampleTimeMin} ]

   if {$i2 > [$itk_component(plot) cget -maxX]} {

      # Re-scale the X axis continuing the drawing from the middle

      set min [expr {$i2 - $range/2} ]
      set max [expr {$i2 + $range/2} ]
      configure -minX $min -maxX $max

      # Delete all not-displayed points

      set tmp [expr {($counter + $firstCount)/2} ]
      
#      puts "counter = $counter   firstCount = $firstCount  tmp = $tmp"
      for {set i $firstCount} {$i < $tmp} { incr i} {
	  $itk_component(plot) deleteTag $i
      }
      
      
      puts "scroll X  time [clock format [clock seconds] -format %T] X min/max $min/$max"
      set firstCount $tmp

      # Draw the two lines for upper/lower limits

      $itk_component(plot) deleteTag "upperLimit lowerLimit"

      $itk_component(plot) createLine upperLimit "$min $max"\
              "$itk_option(-valueUpperLim) $itk_option(-valueUpperLim)" \
              $itk_option(-colorUpperLim)

      $itk_component(plot) createLine lowerLimit "$min $max"\
               "$itk_option(-valueLowerLim) $itk_option(-valueLowerLim)"\
               $itk_option(-colorLowerLim)
   }

   # Ckeck if the current value exceeds limits

   if { $value >= $itk_option(-valueUpperLim) } {

       if {$oldValue < $itk_option(-valueUpperLim) } {
          $itk_component(cvArrow) delete limitFlag
          $itk_component(cvArrow) create polygon 0 30 20 30 10 15 \
                        -fill $itk_option(-colorUpperLim) -tag limitFlag
       } 

   } elseif { $value <= $itk_option(-valueLowerLim) } {

       if { $oldValue > $itk_option(-valueLowerLim) } {
          $itk_component(cvArrow) delete limitFlag      
          $itk_component(cvArrow) create polygon 0 190 20 190 10 205 \
                        -fill $itk_option(-colorLowerLim) -tag limitFlag
       }
              
   } else {

      if { ![lempty [$itk_component(cvArrow) gettags limitFlag]]} {
         $itk_component(cvArrow) delete limitFlag
      }
   }

   if {$paused} {
      $itk_component(plot) createLine $counter "$valX $i2" "$oldValue $value" Grey90
   } else {
      $itk_component(plot) createLine $counter "$valX $i2" "$oldValue $value" black
   }

   incr counter;
   set oldValue $value         
   set valX $i2
   update
}

#----------------------------------------------------------------------------
# METHOD : getMinMax
#----------------------------------------------------------------------------

body UifStripChart::getMinMax {} {
   set x1  [$itk_component(plot) cget -minX]
   set x2  [$itk_component(plot) cget -maxX]
   return [list $x1 $x2]
}

#----------------------------------------------------------------------------
# METHOD : detachEvent
#----------------------------------------------------------------------------

body UifStripChart::detachEvent {} {
   if {$eventId != ""} { catch {seq_evtDetach $eventId} }
}

#----------------------------------------------------------------------------
# METHOD : continuePlot
#----------------------------------------------------------------------------

body UifStripChart::continuePlot {} {
   set paused 0
#   set plotState PLOT
#   $this plot
}

#----------------------------------------------------------------------------
# METHOD : pausePlot
#----------------------------------------------------------------------------

body UifStripChart::pausePlot {} {
   set paused 1
#   set plotState PAUSE ; detachEvent
}

#----------------------------------------------------------------------------
# METHOD : stopPlot
#----------------------------------------------------------------------------

body UifStripChart::stopPlot {} {
   set plotState STOP ; detachEvent
}

#----------------------------------------------------------------------------
# METHOD : stripChartConfig
#----------------------------------------------------------------------------

body UifStripChart::stripChartConfig {} {

   if {![winfo exist .wdgConfigure]} {   
     set c ".wdgConfigure"
     toplevel $c

     set cw        $c.cw
     uifChartConfig  $cw $this -topLevel $c 
     pack $cw -fill both -expand 1

     wm minsize $c 300 200

     $cw viewList        
     tkwait window $c ; # Wait until the user do something ...
   }
}


#####################################################################
#                  DEFINE EXTERNAL PROCEDURE
#####################################################################

proc updateChartPoll { w type var rate} {
     
#  puts "DBG> updateChartPoll $w $type $var $rate"

  if {![winfo exists $w]}  { return }
  if {$rate == 0 || [cequal $var ""] }  { return }

  if {![cequal $type DB] && [uplevel #0 info exists $var] == 0 } {
     uplevel #0 set $var 0
  }

  if {[cequal $type TCL]} {  
     set newValue [uplevel #0 set $var]
  } else {  
     if {[catch {seq_dbReadSymbolic $var } newValue] } { return }
  }

  $w drawValue $newValue
  set plotState [$w getPlotState] 

  if {![cequal $plotState PAUSE] && ![cequal $plotState STOP]} {   
     after $rate updateChartPoll $w $type $var $rate  
  }   
}


proc updateChartEvent {w attr type data} {

   if {![winfo exists $w]}     { return }
   if {![cequal $type SCALAR]} { return }

#   puts "DBG> updateChartEvent $w $attr $type $data"
   
   set plotState [$w getPlotState] 
   if {[cequal $plotState PLOT] || [cequal $plotState IDLE]} {
      $w drawValue [lindex $data 2]
   }
}

#
# ___oOo___



